/*
 * @(#)SaslServerFactory.java	1.7 01/01/30
 *
 * Copyright 1999-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 */

package com.sun.security.sasl.preview;

import java.util.Hashtable;
import javax.security.auth.callback.CallbackHandler;

/**
 * An interface for creating instances of <tt>SaslServer</tt>.
 * It is not normally accessed directly by a server, which will use the
 * <tt>Sasl</tt> static methods
 * instead. However, a particular environment may provide and install a
 * new or different <tt>SaslServerFactory</tt>.
 *
 * @see SaslServer
 * @see Sasl
 *
 * @author Rosanna Lee
 * @author Rob Weltman
 */
public abstract interface SaslServerFactory {
    /**
     * Creates a <tt>SaslServer</tt> using the parameters supplied.
     * It returns null
     * if no <tt>SaslServer</tt> can be created using the parameters supplied.
     * Throws <tt>SaslException</tt> if it cannot create a <tt>SaslServer</tt>
     * because of an error.
     *
     * @param mechanism The non-null
     * IANA-registered name of a SASL mechanism. (e.g. "GSSAPI", "CRAM-MD5").
     * @param protocol The non-null string name of the protocol for which
     * the authentication is being performed (e.g., "ldap").
     * @param serverName The non-null string name of the server to which
     * we are creating an authenticated connection.
     * @param props The possibly null set of properties used to select the SASL
     * mechanism and to configure the authentication exchange of the selected
     * mechanism. See the <tt>Sasl</tt> class for a list of standard properties. 
     * Other, possibly mechanism-specific, properties can be included.
     * Properties not relevant to the selected mechanism are ignored.
     *
     * @param cbh The possibly null callback handler to used by the SASL
     * mechanisms to get further information from the application/library
     * to complete the authentication. For example, a SASL mechanism might
     * require the authentication ID, password and realm from the caller.
     * The authentication ID is requested by using a <tt>NameCallback</tt>.
     * The password is requested by using a <tt>PasswordCallback</tt>.
     * The realm is requested by using a <tt>RealmChoiceCallback</tt> if there is a list
     * of realms to choose from, and by using a <tt>RealmCallback</tt> if
     * the realm must be entered. 
     *
     *@return A possibly null <tt>SaslServer</tt> created using the parameters
     * supplied. If null, this factory cannot produce a <tt>SaslServer</tt>
     * using the parameters supplied.
     *@exception SaslException If cannot create a <tt>SaslServer</tt> because
     * of an error.
     */
    public abstract SaslServer createSaslServer(
	String mechanism,
	String protocol,
	String serverName,
	Hashtable props,
	CallbackHandler cbh) throws SaslException;

    /**
     * Returns an array of names of mechanisms that match the specified
     * mechanism selection policies.
     * @param props The possibly null set of properties used to specify the
     * security policy of the SASL mechanism. For example, if <tt>props</tt>
     * contains the <tt>Sasl.POLICY_NOPLAINTEXT</tt> property with the value
     * <tt>"true"</tt>, then the factory must not return any SASL mechanisms
     * that are susceptible to simple plain passive attacks.
     * See the <tt>Sasl</tt> class for a complete list of policy properties.
     * Non-policy related properties, if present in <tt>props</tt>, are ignored.
     * @return A non-null array containing a IANA-registered SASL mechanism names.
     */
    public abstract String[] getMechanismNames(Hashtable props);
}
