/*
 * @(#)InputStreamReader.java	1.36  01/05/08
 *
 * Copyright 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 */

package java.io;

import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.BufferOverflowException;
import java.nio.BufferUnderflowException;
import java.nio.channels.FileChannel;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.IllegalCharsetNameException;
import java.nio.charset.MalformedInputException;
import java.nio.charset.UnsupportedCharsetException;
import sun.io.ByteToCharConverter;
import sun.io.Converters;
import sun.io.ConversionBufferFullException;
import sun.nio.cs.HistoricallyNamedCharset;


/**
 * An InputStreamReader is a bridge from byte streams to character streams: It
 * reads bytes and decodes them into characters using a specified {@link
 * java.nio.charset.Charset </code>charset<code>}.  The charset that it uses
 * may be specified by name or may be given explicitly, or the platform's
 * default charset may be accepted.
 *
 * <p> Each invocation of one of an InputStreamReader's read() methods may
 * cause one or more bytes to be read from the underlying byte-input stream.
 * To enable the efficient conversion of bytes to characters, more bytes may
 * be read ahead from the underlying stream than are necessary to satisfy the
 * current read operation.
 *
 * <p> For top efficiency, consider wrapping an InputStreamReader within a
 * BufferedReader.  For example:
 *
 * <pre>
 * BufferedReader in
 *   = new BufferedReader(new InputStreamReader(System.in));
 * </pre>
 *
 * @see BufferedReader
 * @see InputStream
 * @see java.nio.charset.Charset
 *
 * @version     1.36, 01/05/08
 * @author      Mark Reinhold
 * @since       JDK1.1
 */

public class InputStreamReader extends Reader {

    private static final int defaultByteBufferSize = 8192;

    // Encapsulates either a ByteToCharConverter or a CharsetDecoder,
    // depending upon which type we decide to use during construction
    //
    private abstract static class Filler {
	abstract int fill(char[] cb, int off, int end) throws IOException;
	abstract String encodingName();
	abstract boolean ready();
    }

    private Filler filler;		// The filler we're using

    private InputStream in;

    // In order to handle surrogates properly we must never try to produce
    // fewer than two characters at a time.  If we're only asked to return one
    // character then the other is saved here to be returned later.
    //
    private boolean haveLeftoverChar = false;
    private char leftoverChar;

    // Setup the filler
    //
    private void setup(String charsetName)
         throws UnsupportedEncodingException
    {
        Charset cs = null;
        try {
            if (Charset.isSupported(charsetName))
		filler = new CharsetFiller(Charset.forName(charsetName));
	    else
		filler = new ConverterFiller(charsetName);
        } catch (IllegalCharsetNameException x) {
		filler = new ConverterFiller(charsetName);
	} catch (UnsupportedCharsetException x) {
	    throw new UnsupportedEncodingException(charsetName);
	}
    }

    /**
     * Create an InputStreamReader that uses the default charset.
     *
     * @param  in   An InputStream
     */
    public InputStreamReader(InputStream in) {
	super(in);
	this.in = in;
        try {
	    setup(Converters.getDefaultEncodingName());
        } catch (UnsupportedEncodingException e) {
	    // The default encoding should always be available
	    throw new Error(e);		// ## assert false
	}
    }

    /**
     * Create an InputStreamReader that uses the named charset.
     *
     * @param  in
     *         An InputStream
     *
     * @param  charsetName
     *         The name of a supported
     *         {@link java.nio.charset.Charset </code>charset<code>}
     *
     * @exception  UnsupportedEncodingException
     *             If the named charset is not supported
     */
    public InputStreamReader(InputStream in, String charsetName)
        throws UnsupportedEncodingException
    {
	super(in);
        this.in = in;
	setup(charsetName);
    }

    /**
     * Create an InputStreamReader that uses the given charset. </p>
     *
     * @param  in       An InputStream
     * @param  charset  A charset
     *
     * @since 1.4
     * @spec JSR-51
     */
    public InputStreamReader(InputStream in, Charset cs) {
        super(in);
	this.in = in;
	filler = new CharsetFiller(cs);
    }

    /**
     * Return the name of the character encoding being used by this stream.
     *
     * <p> If the encoding has an historical name then that name is returned;
     * otherwise the encoding's canonical name is returned.
     *
     * <p> If this instance was created with the {@link
     * #InputStreamReader(InputStream, String)} constructor then the returned
     * name, being unique for the encoding, may differ from the name passed to
     * the constructor.  This method may return <code>null</code> if the stream
     * has been closed. </p>
     *
     * @return The historical name of this encoding, or possibly
     *         <code>null</code> if the stream has been closed
     *
     * @see java.nio.charset.Charset
     *
     * @revised 1.4
     * @spec JSR-51
     */
    public String getEncoding() {
        synchronized (lock) {
	    if (filler != null)
	        return filler.encodingName();
	    else
		return null; //Stream closed.
	}
    }


    private class ConverterFiller
	extends Filler
    {
	private ByteToCharConverter btc;
	private byte[] bb;		/* Input buffer */
	private int nBytes = 0;		/* -1 implies EOF has been reached */
	private int nextByte = 0;

	private ConverterFiller(String charsetName)
	    throws UnsupportedEncodingException
	{
	    this.btc = sun.io.ByteToCharConverter.getConverter(charsetName);
	    bb = new byte[defaultByteBufferSize];
	}

	private void malfunction() {
	    throw new InternalError("Converter malfunction (" +
				    btc.getCharacterEncoding() +
				    ") -- please submit a bug report via " +
				    System.getProperty("java.vendor.url.bug"));
	}

	private int convertInto(char cbuf[], int off, int end)
	    throws IOException
	{
	    int nc = 0;
	    if (nextByte < nBytes) {
		try {
		    nc = btc.convert(bb, nextByte, nBytes,
				     cbuf, off, end);
		    nextByte = nBytes;
		    if (btc.nextByteIndex() != nextByte)
			malfunction();
		}
		catch (ConversionBufferFullException x) {
		    nextByte = btc.nextByteIndex();
		    nc = btc.nextCharIndex() - off;
		}
	    }
	    return nc;
	}

	private int flushInto(char cbuf[], int off, int end)
	    throws IOException
	{
	    int nc = 0;
	    try {
		nc = btc.flush(cbuf, off, end);
	    }
	    catch (ConversionBufferFullException x) {
		nc = btc.nextCharIndex() - off;
	    }
	    return nc;
	}

	int fill(char cbuf[], int off, int end) throws IOException {
	    int nc = 0;

	    if (nextByte < nBytes)
		nc = convertInto(cbuf, off, end);

	    while (off + nc < end) {

		if (nBytes != -1) {
		    if ((nc > 0) && !inReady())
			break;      /* Block at most once */
		    nBytes = in.read(bb);
		}

		if (nBytes == -1) {
		    nBytes = 0; /* Allow file to grow */
		    nc += flushInto(cbuf, off + nc, end);
		    if (nc == 0)
			return -1;
		    else
			break;
		} else {
		    nextByte = 0;
		    nc += convertInto(cbuf, off + nc, end);
		}

	    }
	    return nc;
	}

	String encodingName() {
	    if (btc != null)
	        return btc.getCharacterEncoding();
	    else
	        return null;
	}

	boolean ready() {
            return (nextByte < nBytes);
	}

    }



    // In the early stages of the build we haven't yet built the NIO native
    // code, so guard against that by catching the first UnsatisfiedLinkError
    // and setting this flag so that later attempts fail quickly.
    //
    private static volatile boolean channelsAvailable = true;

    private static FileChannel getChannel(FileInputStream in) {
	if (!channelsAvailable)
	    return null;
	try {
	    return in.getChannel();
	} catch (UnsatisfiedLinkError x) {
	    channelsAvailable = false;
	    return null;
	}
    }

    private class CharsetFiller
	extends Filler
    {
	private CharsetDecoder decoder;
	private Charset cs;
	private ByteBuffer bb;

	// If (in instanceof FileInputStream)
	// then this is set to in.getChannel()
	private FileChannel fc = null;

	// Only used when fc == null
	private byte[] bytes = null;

	CharsetFiller(Charset cs) {
	    this.cs = cs;
	    this.decoder = cs.newDecoder();
	    if (in instanceof FileInputStream) {
		fc = getChannel((FileInputStream)in);
		if (fc != null)
		    bb = ByteBuffer.allocateDirect(defaultByteBufferSize);
	    }
	    if (fc == null) {
		bytes = new byte[defaultByteBufferSize];
		bb = ByteBuffer.wrap(bytes);
	    }
	    bb.flip();			// So that bb is initially empty
	}

	private int readBytes() throws IOException {
	    bb.compact();
	    if (fc != null) {
		// Read from the file channel
		int n = fc.read(bb);
		if (n < 0) {
		    bb.flip();
		    return n;
		}
	    } else {
		// Read from the input stream, and then update the buffer
		int n = in.read(bytes, bb.position(), bb.remaining());
		if (n < 0) {
		    bb.flip();
		    return n;
		}
		if (!(n <= bb.remaining())) throw new Error(); // ## assert
		bb.position(bb.position() + n);
	    }
	    bb.flip();
	    if (bb.remaining() == 0) throw new Error();	// ## assert
	    return bb.remaining();
	}

	int fill(char[] cbuf, int off, int end) throws IOException {

	    // In order to handle surrogate pairs, this method requires that
	    // the invoker attempt to read at least two characters.  Saving the
	    // extra character, if any, at a higher level is easier than trying
	    // to deal with it here.
	    if (!(end - off > 1)) throw new Error(); // ## assert

	    CharBuffer cb = CharBuffer.wrap(cbuf, off, end - off);
	    if (cb.position() != 0)
		// Ensure that cb[0] == cbuf[off]
		cb = cb.slice();

	    boolean eof = false;
	    for (;;) {
		try {
		    decoder.decode(bb, cb, eof);
		} catch (BufferUnderflowException x) {
		    if (eof)
			break;
		    if ((cb.position() > 0) && !inReady())
			break;		// Block at most once
		    int n = readBytes();
		    if (n < 0) {
			eof = true;
			if ((cb.position() == 0) && (!bb.hasRemaining()))
			    break;
		    }
		} catch (BufferOverflowException x) {
		    if (!(cb.position() > 0)) throw new Error(); // ## assert
		    break;
		}
	    }

	    if (cb.position() == 0) {
		if (eof)
		    return -1;
		throw new Error();	// ## assert
	    }
	    return cb.position();
	}

	String encodingName() {
	    return ((cs instanceof HistoricallyNamedCharset)
		    ? ((HistoricallyNamedCharset)cs).historicalName()
		    : cs.name());
	}

	boolean ready() {
	    return bb.hasRemaining();
	}

    }


    /**
     * Tell whether the underlying byte stream is ready to be read.  Return
     * false for those streams that do not support available(), such as the
     * Win32 console stream.
     */
    private boolean inReady() {
        try {
            return in.available() > 0;
        } catch (IOException x) {
            return false;
        }
    }

    /** Check to make sure that the stream has not been closed */
    private void ensureOpen() throws IOException {
        if (in == null)
            throw new IOException("Stream closed");
    }

    /**
     * Read a single character.
     *
     * @return The character read, or -1 if the end of the stream has been
     *         reached
     *
     * @exception  IOException  If an I/O error occurs
     */
    public int read() throws IOException {
        return read0();
    }

    private int read0() throws IOException {
	synchronized (lock) {

	    // Return the leftover char, if there is one
	    if (haveLeftoverChar) {
		haveLeftoverChar = false;
		return leftoverChar;
	    }

	    // Convert more bytes
	    char cb[] = new char[2];
	    int n = read(cb, 0, 2);
	    switch (n) {
	    case -1:
		return -1;
	    case 2:
		leftoverChar = cb[1];
		haveLeftoverChar = true;
		// FALL THROUGH
	    case 1:
		return cb[0];
	    default:
		throw new Error();		// ## assert false
	    }
	}
    }

    /**
     * Read characters into a portion of an array.
     *
     * @param      cbuf  Destination buffer
     * @param      off   Offset at which to start storing characters
     * @param      len   Maximum number of characters to read
     *
     * @return     The number of characters read, or -1 if the end of the stream
     *             has been reached
     *
     * @exception  IOException  If an I/O error occurs
     */
    public int read(char cbuf[], int offset, int length) throws IOException {
	int off = offset;
	int len = length;
        synchronized (lock) {
            ensureOpen();
            if ((off < 0) || (off > cbuf.length) || (len < 0) ||
                ((off + len) > cbuf.length) || ((off + len) < 0)) {
                throw new IndexOutOfBoundsException();
            }
	    if (len == 0)
                return 0;

	    int n = 0;

	    if (haveLeftoverChar) {
		// Copy the leftover char into the buffer
		cbuf[off] = leftoverChar;
		off++; len--;
		haveLeftoverChar = false;
		n = 1;
		if ((len == 0) || (!filler.ready() && !inReady()))
		    // Return now if this is all we can produce w/o blocking
		    return n;
	    }

	    if (len == 1) {
		// Treat single-character array reads just like read()
		int c = read0();
		if (c == -1)
		    return (n == 0) ? -1 : n;
		cbuf[off] = (char)c;
		return n + 1;
	    }

	    return n + filler.fill(cbuf, off, off + len);
        }
    }

    /**
     * Tell whether this stream is ready to be read.  An InputStreamReader is
     * ready if its input buffer is not empty, or if bytes are available to be
     * read from the underlying byte stream.
     *
     * @exception  IOException  If an I/O error occurs
     */
    public boolean ready() throws IOException {
        synchronized (lock) {
            ensureOpen();
	    return haveLeftoverChar || filler.ready() || inReady();
        }
    }

    /**
     * Close the stream.
     *
     * @exception  IOException  If an I/O error occurs
     */
    public void close() throws IOException {
        synchronized (lock) {
            if (in == null)
                return;
            in.close();
            in = null;
	    filler = null;
	}
    }

}
