/*
 * @(#)OutputStreamWriter.java	1.37	01/05/08
 *
 * Copyright 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package java.io;

import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.BufferOverflowException;
import java.nio.BufferUnderflowException;
import java.nio.channels.FileChannel;
import java.nio.charset.Charset;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.IllegalCharsetNameException;
import java.nio.charset.UnsupportedCharsetException;
import sun.io.CharToByteConverter;
import sun.io.Converters;
import sun.io.ConversionBufferFullException;
import sun.nio.cs.HistoricallyNamedCharset;


/**
 * An OutputStreamWriter is a bridge from character streams to byte streams:
 * Characters written to it are encoded into bytes using a specified {@link
 * java.nio.charset.Charset </code>charset<code>}.  The charset that it uses
 * may be specified by name or may be given explicitly, or the platform's
 * default charset may be accepted.
 *
 * <p> Each invocation of a write() method causes the encoding converter to be
 * invoked on the given character(s).  The resulting bytes are accumulated in a
 * buffer before being written to the underlying output stream.  The size of
 * this buffer may be specified, but by default it is large enough for most
 * purposes.  Note that the characters passed to the write() methods are not
 * buffered.
 *
 * <p> For top efficiency, consider wrapping an OutputStreamWriter within a
 * BufferedWriter so as to avoid frequent converter invocations.  For example:
 *
 * <pre>
 * Writer out
 *   = new BufferedWriter(new OutputStreamWriter(System.out));
 * </pre>
 *
 * <p> A <i>surrogate pair</i> is a character represented by a sequence of two
 * <tt>char</tt> values: A <i>high</i> surrogate in the range '&#92;uD800' to
 * '&#92;uDBFF' followed by a <i>low</i> surrogate in the range '&#92;uDC00' to
 * '&#92;uDFFF'.  If the character represented by a surrogate pair cannot be
 * encoded by a given charset then a charset-dependent <i>substitution
 * sequence</i> is written to the output stream.
 *
 * <p> A <i>malformed surrogate element</i> is a high surrogate that is not
 * followed by a low surrogate or a low surrogate that is not preceeded by a
 * high surrogate.  It is illegal to attempt to write a character stream
 * containing malformed surrogate elements.  The behavior of an instance of
 * this class when a malformed surrogate element is written is not specified.
 *
 * @see BufferedWriter
 * @see OutputStream
 * @see java.nio.charset.Charset
 *
 * @version 	1.37, 01/05/08
 * @author	Mark Reinhold
 * @since	JDK1.1
 */

public class OutputStreamWriter extends Writer {

    private OutputStream out;
    private static final int defaultByteBufferSize = 8192;

    private abstract static class ByteOutputter {
	abstract void fillBytes(char[] cb, int off, int len)
	    throws IOException;
	abstract String encodingName();
	abstract void flushConvBuffer() throws IOException;
    }
    private ByteOutputter byteOutputter;   // the byte outputter we are using

    // Setup the byte outputter
    //
    private void setup(String charsetName)
         throws UnsupportedEncodingException
    {
        try {
            if (Charset.isSupported(charsetName)) {
		byteOutputter
		    = new CharsetByteOutputter(Charset.forName(charsetName));
	    } else {
		byteOutputter = new ConverterByteOutputter(charsetName);
	    }
        } catch (IllegalCharsetNameException x) {
	    byteOutputter = new ConverterByteOutputter(charsetName);
	} catch (UnsupportedCharsetException x) {
	    throw new UnsupportedEncodingException(charsetName);
	}
    }

    /**
     * Create an OutputStreamWriter that uses the named charset.
     *
     * @param  out
     *         An OutputStream
     *
     * @param  charsetName
     *         The name of a supported
     *         {@link java.nio.charset.Charset </code>charset<code>}
     *
     * @exception  UnsupportedEncodingException
     *             If the named encoding is not supported
     */
    public OutputStreamWriter(OutputStream out, String charsetName)
	throws UnsupportedEncodingException
    {
	super(out);
	this.out = out;
	setup(charsetName);
    }

    /**
     * Create an OutputStreamWriter that uses the default character encoding.
     *
     * @param  out  An OutputStream
     */
    public OutputStreamWriter(OutputStream out) {
	super(out);
	this.out = out;
	try {
	    setup(Converters.getDefaultEncodingName());
	} catch (UnsupportedEncodingException e) {
	    throw new Error(e);
        } 
    }

    /**
     * Create an OutputStreamWriter that uses the given charset. </p>
     *
     * @param  in
     *         An InputStream
     *
     * @param  charset
     *         A charset
     *
     * @since 1.4
     * @spec JSR-51
     */
    public OutputStreamWriter(OutputStream out, Charset charset) {
	super(out);
	this.out = out;
	byteOutputter = new CharsetByteOutputter(charset); 
    }

    /**
     * Return the name of the character encoding being used by this stream.
     *
     * <p> If the encoding has an historical name then that name is returned;
     * otherwise the encoding's canonical name is returned.
     *
     * <p> If this instance was created with the {@link
     * #OutputStreamWriter(OutputStream, String)} constructor then the returned
     * name, being unique for the encoding, may differ from the name passed to
     * the constructor.  This method may return <tt>null</tt> if the stream has
     * been closed. </p>
     *
     * @return The historical name of this encoding, or possibly
     *         <code>null</code> if the stream has been closed
     *
     * @see java.nio.charset.Charset
     *
     * @revised 1.4
     * @spec JSR-51
     */
    public String getEncoding() {
	synchronized (lock) {
	    if (byteOutputter != null)
	        return byteOutputter.encodingName();
	    else
	        return null;
	}
    }


    /** Check to make sure that the stream has not been closed */
    private void ensureOpen() throws IOException {
	if (out == null)
	    throw new IOException("Stream closed");
    }

    private class ConverterByteOutputter extends ByteOutputter {

	private CharToByteConverter ctb;
	private byte[] bb;

	/* nextByte is where the next byte will be written into bb */
	private int nextByte = 0;

	/* nByte is the buffer size = defaultByteBufferSize in this class */
	private int nBytes = 0;


	private ConverterByteOutputter() {
	    this.ctb = CharToByteConverter.getDefault();
	    bb = new byte[defaultByteBufferSize];
	    nBytes = defaultByteBufferSize;
        }

	private ConverterByteOutputter(String enc)
		throws UnsupportedEncodingException
	{
	    this.ctb = CharToByteConverter.getConverter(enc);
	    bb = new byte[defaultByteBufferSize];
	    nBytes = defaultByteBufferSize;
        }

	void fillBytes(char[] cbuf, int off, int len ) throws IOException {
	    int ci = off, end = off + len;
	    boolean bufferFlushed = false; 

	    while (ci < end) {
		boolean bufferFull = false;
		try {
		    nextByte += ctb.convertAny(cbuf, ci, end,
					       bb, nextByte, nBytes);
		    ci = end;
		}
		catch (ConversionBufferFullException x) {
		    int nci = ctb.nextCharIndex();
		    if ((nci == ci) && bufferFlushed) {
			/* If the buffer has been flushed and it 
			   still does not hold even one character */
			throw new 
			    CharConversionException("Output buffer too small");
		    }
		    ci = nci;
		    bufferFull = true;
		    nextByte = ctb.nextByteIndex();
		} 
		if ((nextByte >= nBytes) || bufferFull) {
		    out.write(bb, 0, nextByte);
		    nextByte = 0;
		    bufferFlushed = true;
		}
	    }
	}

	/**
	 * Flush the output buffer to the underlying byte stream, without flushing
	 * the byte stream itself.  This method is non-private only so that it may
	 * be invoked by PrintStream.
	 */
	void flushConvBuffer() throws IOException {
	     synchronized (lock) {
		ensureOpen();
		for (;;) {
		    try {
			nextByte += ctb.flushAny(bb, nextByte, nBytes);
		    }
		    catch (ConversionBufferFullException x) {
			nextByte = ctb.nextByteIndex();
		    }
		    if (nextByte == 0)
			break;
		    if (nextByte > 0) {
			out.write(bb, 0, nextByte);
			nextByte = 0;
		    }
		}
	    }
	}

	String encodingName() {
	    return ctb.getCharacterEncoding();
	}
    }

    private class CharsetByteOutputter extends ByteOutputter {

	private Charset cs = null;
	private CharsetEncoder encoder = null;
	private ByteBuffer bb;

	// If (in instanceof FileOutputStream)
	// then this is set to in.getChannel()
	private FileChannel fc = null;

	// Only used when fc == null
	private byte[] bytes = null;

	CharsetByteOutputter(Charset cs) {
	    this.cs = cs;
	    this.encoder = cs.newEncoder();

	    if (out instanceof FileOutputStream) {
                fc = ((FileOutputStream)out).getChannel();
		if (fc != null)
                    bb = ByteBuffer.allocateDirect(defaultByteBufferSize);
	    }
            if (fc == null) {
		bytes = new byte[defaultByteBufferSize];
		bb = ByteBuffer.wrap(bytes);
	    }
	}

	private void writeBytes() throws IOException {
	    bb.flip();
	    int n = bb.remaining();
	    if (n > 0) {
		if (fc != null) {
		    if (fc.write(bb) != n) throw new Error(); // ## assert
		} else {
		    out.write(bytes, 0, n);
		}
	    }
	    bb.clear();
        }

	void fillBytes(char cbuf[], int off, int len) throws IOException {
	    CharBuffer cb = CharBuffer.wrap(cbuf, off, len);

	    while (cb.hasRemaining()) {
		try {
		    encoder.encode(cb, bb, false);
		} catch (BufferOverflowException x) {
		    if (!(bb.position() > 0)) throw new Error(); // ## assert
		    writeBytes();
		} catch (BufferUnderflowException x) {
		    throw new Error(x);	// ## assert
		}
	    }

	}

	void flushConvBuffer() throws IOException {
	    try {
		encoder.flush(bb);
	    } catch (BufferOverflowException x) {
		// In case we were already near the end of the byte buffer
		writeBytes();
		encoder.flush(bb);
	    }
	    if (bb.position() > 0)
		writeBytes();
	}

	String encodingName() {
	    return ((cs instanceof HistoricallyNamedCharset)
		    ? ((HistoricallyNamedCharset)cs).historicalName()
		    : cs.name());
	}

    }

    /**
     * Flush the output buffer to the underlying byte stream, without flushing
     * the byte stream itself.  This method is non-private only so that it may
     * be invoked by PrintStream.
     */
    void flushBuffer() throws IOException {
        synchronized (lock) {
	    if (byteOutputter != null)
	       byteOutputter.flushConvBuffer();
	    else
		throw new IOException("Stream closed");
	}
    }

    /**
     * Write a single character.
     *
     * @exception  IOException  If an I/O error occurs
     */
    public void write(int c) throws IOException {
	char cbuf[] = new char[1];
	cbuf[0] = (char) c;
	write(cbuf, 0, 1);
    }

    /**
     * Write a portion of an array of characters.
     *
     * @param  cbuf  Buffer of characters
     * @param  off   Offset from which to start writing characters
     * @param  len   Number of characters to write
     *
     * @exception  IOException  If an I/O error occurs
     */
    public void write(char cbuf[], int off, int len) throws IOException {
	synchronized (lock) {
	    ensureOpen();
            if ((off < 0) || (off > cbuf.length) || (len < 0) ||
                ((off + len) > cbuf.length) || ((off + len) < 0)) {
                throw new IndexOutOfBoundsException();
            } else if (len == 0) {
                return;
            }
	    byteOutputter.fillBytes(cbuf, off, len);
	}
    }

    /**
     * Write a portion of a string.
     *
     * @param  str  A String
     * @param  off  Offset from which to start writing characters
     * @param  len  Number of characters to write
     *
     * @exception  IOException  If an I/O error occurs
     */
    public void write(String str, int off, int len) throws IOException {
	/* Check the len before creating a char buffer */
	if (len < 0)
	    throw new IndexOutOfBoundsException();

	char cbuf[] = new char[len];
	str.getChars(off, off + len, cbuf, 0);
	write(cbuf, 0, len);
    }


    /**
     * Flush the stream.
     *
     * @exception  IOException  If an I/O error occurs
     */
    public void flush() throws IOException {
	synchronized (lock) {
	    flushBuffer();
	    out.flush();
	}
    }

    /**
     * Close the stream.
     *
     * @exception  IOException  If an I/O error occurs
     */
    public void close() throws IOException {
	synchronized (lock) {
	    if (out == null)
		return;
	    flush();
	    out.close();
	    out = null;
	    byteOutputter = null;
	}
    }

}
