/*
 * @(#)InheritableThreadLocal.java	1.13 00/10/27
 *
 * Copyright 1998-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package java.lang;
import java.util.*;

/**
 * This class extends <tt>ThreadLocal</tt> to provide inheritance of values
 * from parent thread to child thread: when a child thread is created, the
 * child receives initial values for all inheritable thread-local variables
 * for which the parent has values.  Normally the child's values will be
 * identical to the parent's; however, the child's value can be made an
 * arbitrary function of the parent's by overriding the <tt>childValue</tt>
 * method in this class.
 * 
 * <p>Inheritable thread-local variables are used in preference to
 * ordinary thread-local variables when the per-thread-attribute being
 * maintained in the variable (e.g., User ID, Transaction ID) must be
 * automatically transmitted to any child threads that are created.
 *
 * @author  Josh Bloch
 * @version 1.13, 10/27/00
 * @see ThreadLocal
 * @since 1.2
 */

public class InheritableThreadLocal extends ThreadLocal {
    /**
     * The expected maximum number of inheritable thread-local variables
     * that a thread will access in its lifetime.  This value is used
     * as a hint for setting the the initial size of the IdentityHashMap
     * used to map InheritableThreadLocal to value.  (One such map exists per
     * thread.) If this value turns out to be low, a very slight cost may be
     * incurred in growing the map.  If it's set too high, space may be
     * wasted.
     */
    private static final int EXPECTED_MAX = 8;

    /**
     * Creates an inheritable thread-local variable.
     */
    public InheritableThreadLocal() {
    }

    /**
     * Returns the value in the current thread's copy of this inheritable
     * thread-local variable.  Creates and initializes the copy if this is
     * the first time the thread has called this method.
     *
     * @return the current thread's value of this thread-local
     */
    public Object get() {
	Thread ourThread = Thread.currentThread();
        Map map = ourThread.inheritableThreadLocals;
	Object value = map.get(this);
        if (value==null && !map.containsKey(this)) {
            if (map == Collections.EMPTY_MAP)
                map = ourThread.inheritableThreadLocals
                    = new IdentityHashMap(EXPECTED_MAX);
            value = initialValue();
            map.put(this, value);
        }
	return value;
    }

    /**
     * Sets the current thread's copy of this thread-local variable
     * to the specified value.  This is only used to change the value from
     * the one assigned by the <tt>initialValue</tt> method, and many
     * applications will have no need for this functionality.
     *
     * @param value the value to be stored in the current threads' copy of
     *	      this thread-local.
     */
    public void set(Object value) {
	Thread ourThread = Thread.currentThread();
        Map map = ourThread.inheritableThreadLocals;
        if (map == Collections.EMPTY_MAP)
            map = ourThread.inheritableThreadLocals
                = new IdentityHashMap(EXPECTED_MAX);
        map.put(this, value);
    }

    /**
     * Computes the child's initial value for this inheritable thread-local
     * variable as a function of the parent's value at the time the child
     * thread is created.  This method is called from within the parent
     * thread before the child is started.
     * <p>
     * This method merely returns its input argument, and should be overridden
     * if a different behavior is desired.
     *
     * @param parentValue the parent thread's value
     * @return the child thread's initial value
     */
    protected Object childValue(Object parentValue) {
        return parentValue;
    }

    /**
     * Passes the thread-local values represented by the specified list of
     * Entries on to the specified child thread.  (The child's value is
     * computed from the parent's as per the childValue method.)  This
     * method is invoked (only) at thread creation time, by Thread.init.
     */
    static void bequeath(Thread parent, Thread child) {
        Map parentMap = parent.inheritableThreadLocals;
        if (parentMap == Collections.EMPTY_MAP)
            return;
        Map childMap = child.inheritableThreadLocals
                     = new IdentityHashMap(EXPECTED_MAX);

        for (Iterator i=parentMap.entrySet().iterator(); i.hasNext(); ) {
            Map.Entry e = (Map.Entry) i.next();
            InheritableThreadLocal itl = (InheritableThreadLocal) e.getKey();
            childMap.put(itl, itl.childValue(e.getValue()));
        }
    }
}
