/*
 * @(#)ThreadLocal.java	1.17 00/10/27
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package java.lang;
import java.util.*;

/**
 * This class provides thread-local variables.  These variables differ from
 * their normal counterparts in that each thread that accesses one (via its
 * <tt>get</tt> or <tt>set</tt> method) has its own, independently initialized
 * copy of the variable.  <tt>ThreadLocal</tt> instances are typically private
 * static fields in classes that wish to associate state with a thread (e.g.,
 * a user ID or Transaction ID).
 *
 * <p>For example, in the class below, the private static <tt>ThreadLocal</tt>
 * instance (<tt>serialNum</tt>) maintains a "serial number" for each thread
 * that invokes the class's static <tt>SerialNum.get()</tt> method, which
 * returns the current thread's serial number.  (A thread's serial number is
 * assigned the first time it invokes <tt>SerialNum.get()</tt>, and remains
 * unchanged on subsequent calls.)
 * <pre>
 * public class SerialNum {
 *     // The next serial number to be assigned
 *     private static int nextSerialNum = 0;
 * 
 *     private static ThreadLocal serialNum = new ThreadLocal() {
 *         protected synchronized Object initialValue() {
 *             return new Integer(nextSerialNum++);
 *         }
 *     };
 * 
 *     public static int get() {
 *         return ((Integer) (serialNum.get())).intValue();
 *     }
 * }
 * </pre>
 * 
 * <p>Each thread holds an implicit reference to its copy of a thread-local
 * variable as long as the thread is alive and the <tt>ThreadLocal</tt>
 * instance is accessible; after a thread goes away, all of its copies of
 * thread-local instances are subject to garbage collection (unless other
 * references to these copies exist).
 *
 * @author  Josh Bloch
 * @version 1.17, 10/27/00
 * @since   1.2
 */

public class ThreadLocal {
    /**
     * The expected maximum number of thread-local variables that
     * a thread will access in its lifetime.  This value is used
     * as a hint for setting the the initial size of the IdentityHashMap
     * used to map ThreadLocal to value.  (One such map exists per thread.)
     * If this value turns out to be low, a very slight cost may be
     * incurred in growing the map.  If it's set too high, space may be
     * wasted.
     */
    private static final int EXPECTED_MAX = 8;

    /**
     * Creates a thread-local variable.
     */
    public ThreadLocal() {
    }

    /**
     * Returns the current thread's initial value for this thread-local
     * variable.  This method will be called once per accessing thread for each
     * thread-local, the first time each thread accesses the variable with the
     * {@link #get()} or {@link #set(Object)} method.  If the programmer
     * desires thread-local variables to be initialized to some value other
     * than <tt>null</tt>, <tt>ThreadLocal</tt> must be subclassed, and this
     * method overridden.  Typically, an anonymous inner class will be used.
     * Typical implementations of <tt>initialValue</tt> will call an
     * appropriate constructor and return the newly constructed object.
     *
     * @return the initial value for this thread-local
     */
    protected Object initialValue() {
	return null;
    }

    /**
     * Returns the value in the current thread's copy of this thread-local
     * variable.  Creates and initializes the copy if this is the first time
     * the thread has called this method.
     *
     * @return the current thread's value of this thread-local
     */
    public Object get() {
	Thread ourThread = Thread.currentThread();
        Map map = ourThread.threadLocals;
	Object value = map.get(this);
        if (value==null && !map.containsKey(this)) {
            if (map == Collections.EMPTY_MAP)
                map = ourThread.threadLocals
                    = new IdentityHashMap(EXPECTED_MAX);
            value = initialValue();
            map.put(this, value);
        }
	return value;
    }

    /**
     * Sets the current thread's copy of this thread-local variable
     * to the specified value.  This is only used to change the value from
     * the one assigned by the <tt>initialValue</tt> method, and many
     * applications will have no need for this functionality.
     *
     * @param value the value to be stored in the current threads' copy of
     *	      this thread-local.
     */
    public void set(Object value) {
	Thread ourThread = Thread.currentThread();
        Map map = ourThread.threadLocals;
        if (map == Collections.EMPTY_MAP)
            map = ourThread.threadLocals = new IdentityHashMap(EXPECTED_MAX);
        map.put(this, value);
    }
}
