/*
 * @(#)Inet4Address.java	1.70 00/02/02
 *
 * Copyright 1995-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package java.net;

import java.security.AccessController;
import java.io.ObjectStreamException;
import sun.security.action.*;

/**
 * Historically the IPv4 TTL field in the IP header has doubled as a
 * multicast scope field: a TTL of 0 means node-local, 1 means
 * link-local, up through 32 means site-local, up through 64 means
 * region-local, up through 128 means continent-local, and up through
 * 255 are global. However, the administrative scoping is preferred. 
 */

public final
class Inet4Address extends InetAddress {
    final static int INADDRSZ = 4;
    /*
     * Perform initializations.
     */
    static {
        init();
    }
   
    Inet4Address() {
	super();
	hostName = null;
	address = 0;
	family = IPv4;
    }

    Inet4Address(String hostName, byte addr[]) {
	this.hostName = hostName;
	this.family = IPv4;
	if (addr != null) {
	    if (addr.length == INADDRSZ) {
		address  = addr[3] & 0xFF;
		address |= ((addr[2] << 8) & 0xFF00);
		address |= ((addr[1] << 16) & 0xFF0000);
		address |= ((addr[0] << 24) & 0xFF000000);
	    } 
	} 
    }
    Inet4Address(String hostName, int address) {
	this.hostName = hostName;
	this.family = IPv4;
	this.address = address;
    }

    private Object writeReplace() throws ObjectStreamException {
	// will replace the to be serialized 'this' object
	InetAddress inet = new InetAddress();
	inet.hostName = this.hostName;
	inet.family = this.family;
	inet.address = this.address;
	return inet;
    }

    /**
     * Utility routine to check if the InetAddress is an
     * IP multicast address. IP multicast address is a Class D
     * address i.e first four bits of the address are 1110.
     * @return a <code>boolean</code> indicating if the InetAddress is 
     * an IP multicast address
     * @since   JDK1.1
     */
    public boolean isMulticastAddress() {
	return ((address & 0xf0000000) == 0xe0000000);
    }

    /**
     * Utility routine to check if the InetAddress in a wildcard address.
     * @return a <code>boolean</code> indicating if the Inetaddress is
     *         a wildcard address.
     * @since 1.4
     */    
    public boolean isAnyLocalAddress() {
	return address == 0;
    }

    /**
     * Utility routine to check if the InetAddress is a loopback address. 
     *
     * @return a <code>boolean</code> indicating if the InetAddress is 
     * a loopback address; or false otherwise.
     * @since 1.4
     */
    private static final int loopback = 2130706433; /* 127.0.0.1 */
    public boolean isLoopbackAddress() {
 	/* 127.x.x.x */
 	byte[] byteAddr = getAddress();
 	return byteAddr[0] == 127;
    }

    /**
     * Utility routine to check if the InetAddress is an link local address. 
     *
     * @return a <code>boolean</code> indicating if the InetAddress is 
     * a link local address; or false if address is not a link local unicast address.
     * @since 1.4
     */
    public boolean isLinkLocalAddress() {
	// link-local unicast in IPv4 (169.254.0.0/16)
	return (((address >>> 24) & 0xFF) == 169) 
	    && (((address >>> 16) & 0xFF) == 254);
    }

    /**
     * Utility routine to check if the InetAddress is a site local address. 
     *
     * @return a <code>boolean</code> indicating if the InetAddress is 
     * a site local address; or false if address is not a site local unicast address.
     * @since 1.4
     */
    public boolean isSiteLocalAddress() {
	// refer to RFC 1918
	// 10/8 prefix
	// 172.16/12 prefix
	// 192.168/16 prefix
	return (((address >>> 24) & 0xFF) == 10)
	    || ((((address >>> 24) & 0xFF) == 172) 
		&& (((address >>> 16) & 0xFF) >= 16))
	    || ((((address >>> 24) & 0xFF) == 192) 
		&& (((address >>> 16) & 0xFF) == 168));
    }

    /**
     * Utility routine to check if the multicast address has global scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of global scope, false if it is not 
     *         of global scope or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCGlobal() {
	// 224.0.1.0 to 238.255.255.255
 	byte[] byteAddr = getAddress();
 	return ((byteAddr[0] & 0xff) >= 224 && (byteAddr[0] & 0xff) <= 238 ) &&
 	    !((byteAddr[0] & 0xff) == 224 && byteAddr[1] == 0 &&
 	      byteAddr[2] == 0);
    }

    /**
     * Utility routine to check if the multicast address has node scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of node-local scope, false if it is not 
     *         of node-local scope or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCNodeLocal() {
	// unless ttl == 0
	return false;
    }

    /**
     * Utility routine to check if the multicast address has link scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of link-local scope, false if it is not 
     *         of link-local scope or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCLinkLocal() {
	// 224.0.0/24 prefix and ttl == 1
	return (((address >>> 24) & 0xFF) == 224) 
	    && (((address >>> 16) & 0xFF) == 0)
	    && (((address >>> 8) & 0xFF) == 0);
    }

    /**
     * Utility routine to check if the multicast address has site scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of site-local scope, false if it is not 
     *         of site-local scope or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCSiteLocal() {
	// 239.255/16 prefix and ttl < 32
	return (((address >>> 24) & 0xFF) == 239) 
	    && (((address >>> 16) & 0xFF) == 255);
    }

    /**
     * Utility routine to check if the multicast address has organization scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of organization-local scope, 
     *         false if it is not of organization-local scope 
     *         or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCOrgLocal() {
	// 239.192 - 239.195
	return (((address >>> 24) & 0xFF) == 239) 
	    && (((address >>> 16) & 0xFF) >= 192)
	    && (((address >>> 16) & 0xFF) <= 195);
    }  

    /**
     * Returns the raw IP address of this <code>InetAddress</code>
     * object. The result is in network byte order: the highest order
     * byte of the address is in <code>getAddress()[0]</code>.
     *
     * @return  the raw IP address of this object.
     */
    public byte[] getAddress() {
	byte[] addr = new byte[INADDRSZ];

	addr[0] = (byte) ((address >>> 24) & 0xFF);
	addr[1] = (byte) ((address >>> 16) & 0xFF);
	addr[2] = (byte) ((address >>> 8) & 0xFF);
	addr[3] = (byte) (address & 0xFF);
	return addr;
    }

    /**
     * Returns the IP address string in standard text presentation  form.
     *
     * @return  the raw IP address in a string format.
     * @since   JDK1.0.2
     */
    public String getHostAddress() {
	return numericToTextFormat(getAddress());
    }

    /**
     * Returns a hashcode for this IP address.
     *
     * @return  a hash code value for this IP address.
     */
    public int hashCode() {
	return address;
    }
    
    /**
     * Compares this object against the specified object.
     * The result is <code>true</code> if and only if the argument is
     * not <code>null</code> and it represents the same IP address as
     * this object.
     * <p>
     * Two instances of <code>InetAddress</code> represent the same IP
     * address if the length of the byte arrays returned by
     * <code>getAddress</code> is the same for both, and each of the
     * array components is the same for the byte arrays.
     *
     * @param   obj   the object to compare against.
     * @return  <code>true</code> if the objects are the same;
     *          <code>false</code> otherwise.
     * @see     java.net.InetAddress#getAddress()
     */
    public boolean equals(Object obj) { 
	return (obj != null) && (obj instanceof Inet4Address) &&
	    (((InetAddress)obj).address == address);	
    }

    static void initStack() {
	/*
	 * Property "impl.prefix" will be prepended to the classname
	 * of the implementation object we instantiate, to which we
	 * delegate the real work (like native methods).  This
	 * property can vary across implementations of the java.
	 * classes.  The default is an empty String "".
	 */
	String prefix = (String)AccessController.doPrivileged(
		      new GetPropertyAction("impl.prefix", ""));
	InetAddressImpl impl4 = null;
	try {
	    impl4 = (InetAddressImpl)(Class.forName("java.net." + prefix +
						   "Inet4AddressImpl")
				     .newInstance());
	} catch (ClassNotFoundException e) {
	    System.err.println("Class not found: java.net." + prefix +
			       "Inet4AddressImpl:\ncheck impl.prefix property " +
			       "in your properties file.");
	} catch (InstantiationException e) {
	    System.err.println("Could not instantiate: java.net." + prefix +
			       "Inet4AddressImpl:\ncheck impl.prefix property " +
			       "in your properties file.");
	} catch (IllegalAccessException e) {
	    System.err.println("Cannot access class: java.net." + prefix +
			       "Inet4AddressImpl:\ncheck impl.prefix property " +
			       "in your properties file.");
	}

	if (impl4 == null) {
	    try {
		impl4 = (InetAddressImpl)(Class.forName("java.net.Inet4AddressImpl")
					 .newInstance());
	    } catch (Exception e) {
		throw new Error("System property impl.prefix incorrect");
	    }
	}

	if (impl4 != null) {
	    impl = impl4;
	}
    }

    static void initAddress () {
	unknownAddress = new Inet4Address();
	anyLocalAddress = new Inet4Address(); // {0x00,0x00,0x00,0x00}
	anyLocalAddress.hostName = "0.0.0.0";
	byte[] loopback = {0x7f,0x00,0x00,0x01};
	loopbackAddress = new Inet4Address("localhost", loopback);

	/* find the local host name */
	try {
	    localHost = new Inet4Address();
	    localHost.hostName = impl.getLocalHostName();
	    /* we explicitly leave the address of the local host
	     * uninitialized.  A DNS lookup in this, the static
	     * initializer, will cause a machine disconnected
	     * from the network to hang - it'll be trying to query
	     * a DNS server that isn't there.
	     *
	     * Instead, we just get the hostname of the local host.
	     * The native code for this just calls gethostname()
	     * which should be pretty innocuous - it shouldn't try
	     * to contact a DNS server.  If any application
	     * calls InetAddress.getLocalHost(), we initialize
	     * the local host's address there if not already initialized.
	     *
	     * Note that for this to work it is also essential that
	     * the localHost InetAddress is _NOT_ put into the address cache
	     * here in the static initializer (which happens if we call
	     * getByName() from the static initializer).  It _IS_ OK
	     * to put it in the addressCache after initialization.
	     *
	     * The unitialized state of the localHost's address is -1,
	     * or IP address 255.255.255.255 which we know cannot be
	     * a legal host address.
	     */
	    localHost.address = -1;
	} catch (Exception ex) { /* this shouldn't happen */
	    localHost = unknownAddress;
	}

	/* cache the name/address pair "0.0.0.0"/0.0.0.0 */
	unknown_array = new Inet4Address[1];
	unknown_array[0] = anyLocalAddress;
    }

    // Utilities
    /* 
     * Converts IPv4 binary address into a string suitable for presentation.
     *
     * @param src a byte array representing an IPv4 numeric address
     * @return a String representing the IPv4 address in standard format
     * @since 1.4
     */
    
    static String numericToTextFormat(byte[] src)
    {
	return (src[0] & 0xff) + "." + (src[1] & 0xff) + "." + (src[2] & 0xff) + "." + (src[3] & 0xff);
    }

    /*
     * Converts IPv4 address in its standard text presentation form 
     * into its numeric binary form.
     * 
     * @param src a String representing an IPv4 address in standard format
     * @return a byte array representing the IPv4 numeric address
     * @since 1.4
     */
    static byte[] textToNumericFormat(String src)
    {
	if (src.length() == 0) {
	    return null;
	}
	
	int octets;
	char ch;
	byte[] dst = new byte[INADDRSZ];
        char[] srcb = src.toCharArray();
	boolean saw_digit = false;

	octets = 0;
	int i = 0;
	int cur = 0;
	while (i < srcb.length) {
	    ch = srcb[i++];
	    if (Character.isDigit(ch)) {
		int sum =  dst[cur]*10 + (Character.digit(ch, 10) & 0xff);
		
		if (sum > 255)
		    return null;

		dst[cur] = (byte)(sum & 0xff);
		if (! saw_digit) {
		    if (++octets > INADDRSZ)
			return null;
		    saw_digit = true;
		}
	    } else if (ch == '.' && saw_digit) {
		if (octets == INADDRSZ)
		    return null;
		cur++;
		dst[cur] = 0;
		saw_digit = false;
	    } else
		return null;
	}
	if (octets < INADDRSZ)
	    return null;
	return dst;
    }
    
    /**
     * Perform class load-time initializations.
     */
    private static native void init();
}
class Inet4AddressImpl extends InetAddressImpl {
    native String getLocalHostName() throws UnknownHostException;
    native byte[][]
        lookupAllHostAddr(String hostname) throws UnknownHostException;
    native String getHostByAddr(byte[] addr) throws UnknownHostException;
}
