/*
 * @(#)Inet6Address.java	1.70 00/02/02
 *
 * Copyright 1995-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package java.net;

import java.security.AccessController;
import java.io.ObjectStreamException;
import sun.security.action.*;


public final
class Inet6Address extends InetAddress {
    final static int INADDRSZ = 16;

    /*
     * @serial
     */
    byte[] ipaddress;

    /*
     * Perform initializations.
     */
    static {
        init();
    }

    Inet6Address() {
	super();
	hostName = null;
	ipaddress = new byte[INADDRSZ];
	family = IPv6;
    }

    Inet6Address(String hostName, byte addr[]) {
	this.hostName = hostName;
	if (addr.length == INADDRSZ) { // normal IPv6 address
	    family = IPv6;
	    ipaddress = (byte[])addr.clone();
	} 
    }
    
    /**
     * Utility routine to check if the InetAddress is an
     * IP multicast address. IP multicast address is a Class D
     * address i.e first four bits of the address are 1110.
     * @return a <code>boolean</code> indicating if the InetAddress is 
     * an IP multicast address
     * @since   JDK1.1
     */
    public boolean isMulticastAddress() {
	return ((ipaddress[0] & 0xff) == 0xff);
    }

    /**
     * Utility routine to check if the InetAddress in a wildcard address.
     * @return a <code>boolean</code> indicating if the Inetaddress is
     *         a wildcard address.
     * @since 1.4
     */        
    public boolean isAnyLocalAddress() {
	byte test = 0x00;
	for (int i = 0; i < INADDRSZ; i++) {
	    test |= ipaddress[i];
	}
	return (test == 0x00);
    }

    /**
     * Utility routine to check if the InetAddress is a loopback address. 
     *
     * @return a <code>boolean</code> indicating if the InetAddress is 
     * a loopback address; or false otherwise.
     * @since 1.4
     */
    public boolean isLoopbackAddress() {
	byte test = 0x00;
	for (int i = 0; i < 15; i++) {
	    test |= ipaddress[i];
	}
	return (test == 0x00) && (ipaddress[15] == 0x01);
    }

    /**
     * Utility routine to check if the InetAddress is an link local address. 
     *
     * @return a <code>boolean</code> indicating if the InetAddress is 
     * a link local address; or false if address is not a link local unicast address.
     * @since 1.4
     */
    public boolean isLinkLocalAddress() {
	return (ipaddress[0] & 0xff) == 0xfe 
	    && (ipaddress[1] & 0xff) == 0x80
	    && (ipaddress[2] & 0xff) == 0x00
	    && (ipaddress[3] & 0xff) == 0x00
	    && (ipaddress[4] & 0xff) == 0x00
	    && (ipaddress[5] & 0xff) == 0x00
	    && (ipaddress[6] & 0xff) == 0x00
	    && (ipaddress[7] & 0xff) == 0x00;
    }

    /**
     * Utility routine to check if the InetAddress is a site local address. 
     *
     * @return a <code>boolean</code> indicating if the InetAddress is 
     * a site local address; or false if address is not a site local unicast address.
     * @since 1.4
     */
    public boolean isSiteLocalAddress() {
	return (ipaddress[0] & 0xff) == 0xfe 
	    && (ipaddress[1] & 0xff) == 0xc0
	    && (ipaddress[2] & 0xff) == 0x00
	    && (ipaddress[3] & 0xff) == 0x00
	    && (ipaddress[4] & 0xff) == 0x00
	    && (ipaddress[5] & 0xff) == 0x00;
    }

    /**
     * Utility routine to check if the multicast address has global scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of global scope, false if it is not 
     *         of global scope or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCGlobal() {
	return (ipaddress[0] & 0xff) == 0xff && (ipaddress[1] & 0x0f) == 14;
    }

    /**
     * Utility routine to check if the multicast address has node scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of node-local scope, false if it is not 
     *         of node-local scope or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCNodeLocal() {
	return (ipaddress[0] & 0xff) == 0xff && (ipaddress[1] & 0x0f) == 1;
    }

    /**
     * Utility routine to check if the multicast address has link scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of link-local scope, false if it is not 
     *         of link-local scope or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCLinkLocal() {
	return (ipaddress[0] & 0xff) == 0xff && (ipaddress[1] & 0x0f) == 2;
    }

    /**
     * Utility routine to check if the multicast address has site scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of site-local scope, false if it is not 
     *         of site-local scope or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCSiteLocal() {
	return (ipaddress[0] & 0xff) == 0xff && (ipaddress[1] & 0x0f) == 5;
    }

    /**
     * Utility routine to check if the multicast address has organization scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of organization-local scope, 
     *         false if it is not of organization-local scope 
     *         or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCOrgLocal() {
	return (ipaddress[0] & 0xff) == 0xff && (ipaddress[1] & 0x0f) == 8;
    }  

    /**
     * Returns the raw IP address of this <code>InetAddress</code>
     * object. The result is in network byte order: the highest order
     * byte of the address is in <code>getAddress()[0]</code>.
     *
     * @return  the raw IP address of this object.
     */
    public byte[] getAddress() {
	return (byte[])ipaddress.clone();
    }

    /**
     * Returns the IP address string in standard text presentation  form.
     *
     * @return  the raw IP address in a string format.
     * @since   JDK1.0.2
     */
    public String getHostAddress() {
	return "["+numericToTextFormat(ipaddress)+"]";
    }

    /**
     * Returns a hashcode for this IP address.
     *
     * @return  a hash code value for this IP address.
     */
    public int hashCode() {
	if (ipaddress != null) {
	    int hash = 0;
	    for (int i = 0; i < ipaddress.length; i++) {
		hash += ipaddress[i];
	    }
	    return hash;

	} else {
	    return 0;
	}
    }
    
    /**
     * Compares this object against the specified object.
     * The result is <code>true</code> if and only if the argument is
     * not <code>null</code> and it represents the same IP address as
     * this object.
     * <p>
     * Two instances of <code>InetAddress</code> represent the same IP
     * address if the length of the byte arrays returned by
     * <code>getAddress</code> is the same for both, and each of the
     * array components is the same for the byte arrays.
     *
     * @param   obj   the object to compare against.
     * @return  <code>true</code> if the objects are the same;
     *          <code>false</code> otherwise.
     * @see     java.net.InetAddress#getAddress()
     */
    public boolean equals(Object obj) {
	if (obj == null || 
	    !(obj instanceof Inet6Address))
	    return false;

	Inet6Address inetAddr = (Inet6Address)obj;

	for (int i = 0; i < INADDRSZ; i++) {
	    if (ipaddress[i] != inetAddr.ipaddress[i])
		return false;
	}
	
	return true;
    }

    /**
     * Utility routine to check if the InetAddress is an
     * IPv4 mapped IPv6 address. 
     *
     * @return a <code>boolean</code> indicating if the InetAddress is 
     * an IPv4 mapped IPv6 address; or false if address is IPv4 address.
     */
    static boolean isIPv4MappedAddress(byte[] addr) {
	if (addr.length < INADDRSZ) {
	    return false;
	}
	if ((addr[0] == 0x00) && (addr[1] == 0x00) && 
	    (addr[2] == 0x00) && (addr[3] == 0x00) && 
	    (addr[4] == 0x00) && (addr[5] == 0x00) && 
	    (addr[6] == 0x00) && (addr[7] == 0x00) && 
	    (addr[8] == 0x00) && (addr[9] == 0x00) && 
	    (addr[10] == (byte)0xff) && 
	    (addr[11] == (byte)0xff))  {   
	    return true;
	}
	return false;
    }

    static byte[] convertFromIPv4MappedAddress(byte[] addr) {
	if (isIPv4MappedAddress(addr)) {
	    byte[] newAddr = new byte[Inet4Address.INADDRSZ];
	    System.arraycopy(addr, 12, newAddr, 0, Inet4Address.INADDRSZ);
	    return newAddr;
	}
	return null;
    }

    /**
     * Utility routine to check if the InetAddress is an
     * IPv4 compatible IPv6 address. 
     *
     * @return a <code>boolean</code> indicating if the InetAddress is 
     * an IPv4 compatible IPv6 address; or false if address is IPv4 address.
     * @since 1.4
     */
    public boolean isIPv4CompatibleAddress() {
	if ((ipaddress[0] == 0x00) && (ipaddress[1] == 0x00) && 
	    (ipaddress[2] == 0x00) && (ipaddress[3] == 0x00) && 
	    (ipaddress[4] == 0x00) && (ipaddress[5] == 0x00) && 
	    (ipaddress[6] == 0x00) && (ipaddress[7] == 0x00) && 
	    (ipaddress[8] == 0x00) && (ipaddress[9] == 0x00) && 
	    (ipaddress[10] == 0x00) && (ipaddress[11] == 0x00))  {   
	    return true;
	}
	return false;
    }

    static void initStack() {
	/*
	 * Property "impl.prefix" will be prepended to the classname
	 * of the implementation object we instantiate, to which we
	 * delegate the real work (like native methods).  This
	 * property can vary across implementations of the java.
	 * classes.  The default is an empty String "".
	 */
	String prefix = (String)AccessController.doPrivileged(
		      new GetPropertyAction("impl.prefix", ""));

	InetAddressImpl impl6 = null;
	try {
	    impl6 = (InetAddressImpl)(Class.forName("java.net." + prefix +
						   "Inet6AddressImpl")
				     .newInstance());
	} catch (ClassNotFoundException e) {
	    System.err.println("Class not found: java.net." + prefix +
			       "Inet6AddressImpl:\ncheck impl.prefix property " +
			       "in your properties file.");
	} catch (InstantiationException e) {
	    System.err.println("Could not instantiate: java.net." + prefix +
			       "Inet6AddressImpl:\ncheck impl.prefix property " +
			       "in your properties file.");
	} catch (IllegalAccessException e) {
	    System.err.println("Cannot access class: java.net." + prefix +
			       "Inet6AddressImpl:\ncheck impl.prefix property " +
			       "in your properties file.");
	}

	if (impl6 == null) {
	    try {
		impl6 = (InetAddressImpl)(Class.forName("java.net.Inet6AddressImpl")
					 .newInstance());
	    } catch (Exception e) {
		throw new Error("System property impl.prefix incorrect");
	    }
	}
	if (impl6 != null) {
	    impl = impl6;
	}
    }

    static void initAddress () {
	/* find the local host name */
	try {
	    localHost = new Inet6Address();
	    localHost.hostName = impl.getLocalHostName();
	    /* we explicitly leave the address of the local host
	     * uninitialized.  A DNS lookup in this, the static
	     * initializer, will cause a machine disconnected
	     * from the network to hang - it'll be trying to query
	     * a DNS server that isn't there.
	     *
	     * Instead, we just get the hostname of the local host.
	     * The native code for this just calls gethostname()
	     * which should be pretty innocuous - it shouldn't try
	     * to contact a DNS server.  If any application
	     * calls InetAddress.getLocalHost(), we initialize
	     * the local host's address there if not already initialized.
	     *
	     * Note that for this to work it is also essential that
	     * the localHost InetAddress is _NOT_ put into the address cache
	     * here in the static initializer (which happens if we call
	     * getByName() from the static initializer).  It _IS_ OK
	     * to put it in the addressCache after initialization.
	     *
	     * The unitialized state of the localHost's address is -1,
	     * or IP address 255.255.255.255 which we know cannot be
	     * a legal host address.
	     */
	    localHost.address = -1;
	} catch (Exception ex) { /* this shouldn't happen */
	    localHost = unknownAddress;
	}

	/* cache the name/address pair "::"/:: */
	anyLocalAddress = new Inet6Address();
	anyLocalAddress.hostName = "::";
	byte[] loopback = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
			   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01};
	loopbackAddress = new Inet6Address("localhost", loopback);

	unknownAddress = new Inet6Address();
	
	unknown_array = new InetAddress[1];
	unknown_array[0] = anyLocalAddress;
    }
    
    // Utilities
    private final static int INT16SZ = 2;
    /*
     * Convert IPv6 binary address into presentation (printable) format.
     * Credit:
     *  Converted from C code from Solaris 8 (inet_ntop)
     *
     * @param src a byte array representing the IPv6 numeric address
     * @return a String representing an IPv6 address in standard format
     * @since 1.4
     */
    static String numericToTextFormat(byte[] src)
    {
	class Pos {
	    int base, len;
	}

	Pos best = new Pos(), cur = new Pos();

	String tmp = "ffff:ffff:ffff:ffff:ffff:ffff:255.255.255.255";
	int i;

	/*
	 * Preprocess:
	 *	Copy the input (bytewise) array into a wordwise array.
	 *	Find the longest run of 0x00's in src[] for :: shorthanding.
	 */

	best.base = -1;
	cur.base = -1;

	for (i = 0; i < (INADDRSZ / INT16SZ); i++) {
	    if (src[i<<1] == 0 && src[(i<<1)+1] == 0) {
		if (cur.base == -1) {
		    cur.base = i; 
		    cur.len = 1;
		}
		else {
		    cur.len++;
		}
	    } else {
		if (cur.base != -1) {
		    if (best.base == -1 || cur.len > best.len) {
			best.base = cur.base;
			best.len = cur.len;
		    }
		    cur.base = -1;
		}
	    }
	}
	
	if (cur.base != -1) {
	    if (best.base == -1 || cur.len > best.len) {
		best = cur;
	    }
	}

	if (best.base != -1 && best.len < 2) {
	    best.base = -1;
	}

	/*
	 * Format the result.
	 */
	StringBuffer sb = new StringBuffer(tmp.length());

	for (i = 0; i < (INADDRSZ / INT16SZ); i++) {
	    /* Are we inside the best run of 0x00's? */
	    if (best.base != -1 && i >= best.base &&
		i < (best.base + best.len)) {
		if (i == best.base)
		    sb.append(":");
		continue;
	    }
	    /* Are we following an initial run of 0x00s or any real hex? */
	    if (i != 0)
		sb.append(":");

	    /* Is this address an encapsulated IPv4? */
	    if (i == 6 && best.base == 0 &&
		(best.len == 6 || 
		 (best.len == 5 
		  && ((src[10] & 0xff) == 0xff) 
		  && ((src[11] & 0xff) == 0xff)))) {
		byte[] dst = new byte[Inet4Address.INADDRSZ];
		dst[0] = src[12];
		dst[1] = src[13];
		dst[2] = src[14];
		dst[3] = src[15];
		sb.append(Inet4Address.numericToTextFormat(dst));
		break;
	    }
	
	    sb.append(""+Integer.toHexString(((src[i<<1]<<8) & 0xff00) | (src[(i<<1)+1] & 0xff)));
	}
	/* Was it a trailing run of 0x00's? */
	if (best.base != -1 && (best.base + best.len) == (INADDRSZ / INT16SZ))
	    sb.append(":");

	return sb.toString();
    }

    /* 
     * Convert IPv6 presentation level address to network order binary form.
     * credit:
     *  Converted from C code from Solaris 8 (inet_pton)
     *
     * @param src a String representing an IPv6 address in standard format
     * @return a byte array representing the IPv6 numeric address
     * @since 1.4
     */
    static byte[] textToNumericFormat(String src)
    {
	if (src.length() == 0) {
	    return null;
	}

	int colonp;
	char ch;
	boolean saw_xdigit;
	int val;
	char[] srcb = src.toCharArray();
	byte[] dst = new byte[INADDRSZ];

	colonp = -1;
	int i = 0, j = 0;
	/* Leading :: requires some special handling. */
	if (srcb[i] == ':')
	    if (srcb[++i] != ':')
		return null;
	int curtok = i;
	saw_xdigit = false;
	val = 0;
	while (i < srcb.length) {
	    ch = srcb[i++];
	    int chval = Character.digit(ch, 16);
	    if (chval != -1) {
		val <<= 4;
		val |= chval;
		if (val > 0xffff)
		    return null;
		saw_xdigit = true;
		continue;
	    }
	    if (ch == ':') {
		curtok = i;
		if (!saw_xdigit) {
		    if (colonp != -1)
			return null;
		    colonp = j;
		    continue;
		} else if (i == srcb.length) {
		    return null;
		}
		if (j + INT16SZ > INADDRSZ)
		    return null;
		dst[j++] = (byte) ((val >> 8) & 0xff);
		dst[j++] = (byte) (val & 0xff);
		saw_xdigit = false;
		val = 0;
		continue;
	    }
	    if (ch == '.' && ((j + Inet4Address.INADDRSZ) <= INADDRSZ)) {
		byte[] v4addr = Inet4Address.textToNumericFormat(src.substring(curtok));
		if (v4addr == null) {
		    return null;
		}
		for (int k = 0; k < Inet4Address.INADDRSZ; k++) {
		    dst[j++] = v4addr[k];
		}
		saw_xdigit = false;
		break;	/* '\0' was seen by inet_pton4(). */
	    }
	    return null;
	}
	if (saw_xdigit) {
	    if (j + INT16SZ > INADDRSZ)
		return null;
	    dst[j++] = (byte) ((val >> 8) & 0xff);
	    dst[j++] = (byte) (val & 0xff);
	}

	if (colonp != -1) {
	    int n = j - colonp;
	    
	    if (j == INADDRSZ)
		return null;
	    for (i = 1; i <= n; i++) {
		dst[INADDRSZ - i] = dst[colonp + n - i];
		dst[colonp + n - i] = 0;
	    }
	    j = INADDRSZ;
	}
	if (j != INADDRSZ)
	    return null;
	byte[] newdst = convertFromIPv4MappedAddress(dst);
	if (newdst != null) {
	    return newdst;
	} else {
	    return dst;
	}
    }

    /**
     * Perform class load-time initializations.
     */
    private static native void init();

}
class Inet6AddressImpl extends InetAddressImpl {
    native String getLocalHostName() throws UnknownHostException;
    native byte[][]
        lookupAllHostAddr(String hostname) throws UnknownHostException;
    native String getHostByAddr(byte[] addr) throws UnknownHostException;
}
