/*
 * @(#)InetAddress.java	1.71 00/06/28
 *
 * Copyright 1995-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package java.net;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Random;
import java.util.Iterator;
import java.util.Vector;
import java.security.AccessController;
import java.io.ObjectStreamException;
import sun.security.action.*;
import sun.net.InetAddressCachePolicy;
import sun.misc.Service;
import sun.net.spi.nameservice.*;

/**
 * This class represents an Internet Protocol (IP) address.
 * <p>
 * Applications should use the methods <code>getLocalHost</code>,
 * <code>getByName</code>, or <code>getAllByName</code> to
 * create a new <code>InetAddress</code> instance.
 *
 * @author  Chris Warth
 * @version 1.71, 06/28/00
 * @see     java.net.InetAddress#getAllByName(java.lang.String)
 * @see     java.net.InetAddress#getByName(java.lang.String)
 * @see     java.net.InetAddress#getLocalHost()
 * @since   JDK1.0
 */
public
class InetAddress implements java.io.Serializable {
    /** 
     * Specify the address family: Internet Protocol, Version 4
     * @since 1.4
     */
    static final int IPv4 = 1;

    /** 
     * Specify the address family: Internet Protocol, Version 6
     * @since 1.4
     */
    static final int IPv6 = 2;

    /* Used to specify address family preference */
    static transient boolean preferIPv6Address = false;

    /**
     * @serial
     */
    String hostName;

    /*
     * Currently we only deal effectively with 32-bit addresses.
     * However this field can be expanded to be a byte array
     * or a 64-bit quantity without too much effort.
     *
     * @serial
     */
    int address;

    /**
     * @serial
     */
    int family;

    /* Used to store the name service provider */
    private static NameService nameService = null; 

    /* Used to store the best available hostname */
    private transient String canonicalHostName = null;

    /** use serialVersionUID from JDK 1.0.2 for interoperability */
    private static final long serialVersionUID = 3286316764910316507L;

    /*
     * Load net library into runtime, and perform initializations.
     */
    static {
	preferIPv6Address = 
	    ((Boolean)java.security.AccessController.doPrivileged(
		 new GetBooleanAction("java.net.preferIPv6Addresses"))).booleanValue();
	AccessController.doPrivileged(new LoadLibraryAction("net"));
        init();
    }

    /**
     * Constructor for the Socket.accept() method.
     * This creates an empty InetAddress, which is filled in by
     * the accept() method.  This InetAddress, however, is not
     * put in the address cache, since it is not created by name.
     */
    InetAddress() {
    }

    private Object readResolve() throws ObjectStreamException {
	// will replace the deserialized 'this' object
	return new Inet4Address(this.hostName, this.address); 
    }

    /**
     * Utility routine to check if the InetAddress is an
     * IP multicast address. IP multicast address is a Class D
     * address i.e first four bits of the address are 1110.
     * @return a <code>boolean</code> indicating if the InetAddress is 
     * an IP multicast address
     * @since   JDK1.1
     */
    public boolean isMulticastAddress() {
	return false;
    }

    /**
     * Utility routine to check if the InetAddress in a wildcard address.
     * @return a <code>boolean</code> indicating if the Inetaddress is
     *         a wildcard address.
     * @since 1.4
     */    
    public boolean isAnyLocalAddress() {
	return false;
    }

    /**
     * Utility routine to check if the InetAddress is a loopback address. 
     *
     * @return a <code>boolean</code> indicating if the InetAddress is 
     * a loopback address; or false otherwise.
     * @since 1.4
     */
    public boolean isLoopbackAddress() {
	return false;
    }

    /**
     * Utility routine to check if the InetAddress is an link local address. 
     *
     * @return a <code>boolean</code> indicating if the InetAddress is 
     * a link local address; or false if address is not a link local unicast address.
     * @since 1.4
     */
    public boolean isLinkLocalAddress() {
	return false;
    }

    /**
     * Utility routine to check if the InetAddress is a site local address. 
     *
     * @return a <code>boolean</code> indicating if the InetAddress is 
     * a site local address; or false if address is not a site local unicast address.
     * @since 1.4
     */
    public boolean isSiteLocalAddress() {
	return false;
    }

    /**
     * Utility routine to check if the multicast address has global scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of global scope, false if it is not 
     *         of global scope or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCGlobal() {
	return false;
    }

    /**
     * Utility routine to check if the multicast address has node scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of node-local scope, false if it is not 
     *         of node-local scope or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCNodeLocal() {
	return false;
    }

    /**
     * Utility routine to check if the multicast address has link scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of link-local scope, false if it is not 
     *         of link-local scope or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCLinkLocal() {
	return false;
    }

    /**
     * Utility routine to check if the multicast address has site scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of site-local scope, false if it is not 
     *         of site-local scope or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCSiteLocal() {
	return false;
    }

    /**
     * Utility routine to check if the multicast address has organization scope
     *
     * @return a <code>boolean</code> indicating if the address has 
     *         is a multicast address of organization-local scope, 
     *         false if it is not of organization-local scope 
     *         or it is not a multicast address
     * @since 1.4
     */
    public boolean isMCOrgLocal() {
	return false;
    }

    /**
     * Gets the host name for this IP address.
     *
     * <p>If there is a security manager, its
     * <code>checkConnect</code> method is first called
     * with the hostname and <code>-1</code> 
     * as its arguments to see if the operation is allowed.
     *
     * @return the host name for this IP address.
     * 
     * @exception  SecurityException  if a security manager exists and its  
     *  <code>checkConnect</code> method doesn't allow the operation .
     * 
     * @see SecurityManager#checkConnect
     */
    public String getHostName() {
	return getHostName(true);
    }

    /**
     * Returns the hostname for this address.
     * If the host is equal to null, then this address refers to any
     * of the local machine's available network addresses.
     * this is package private so SocketPermission can make calls into
     * here without a security check.
     *
     * <p>If there is a security manager, this method first
     * calls its <code>checkConnect</code> method
     * with the hostname and <code>-1</code> 
     * as its arguments to see if the calling code is allowed to know
     * the hostname for this IP address, i.e., to connect to the host.
     * 
     * @return  the host name for this IP address.
     * 
     * @param check make security check if true
     * 
     * @exception  SecurityException  if a security manager exists and its  
     *  <code>checkConnect</code> method doesn't allow the connection.
     * 
     * @see SecurityManager#checkConnect
     */
    String getHostName(boolean check) {
	if (hostName == null) {
	    hostName = InetAddress.getHostFromNameService(this, check);
	}
	return hostName;
    }

    /**
     * Gets the fully qualified domain name for this IP address.
     * Best effort method, meaning we may not be able to return 
     * the FQDN depending on the underlying system configuration.
     *
     * @exception  SecurityException  if a security manager exists and its  
     *  <code>checkConnect</code> method doesn't allow the operation .
     *
     * @see SecurityManager#checkConnect
     *
     * @since 1.4
     */
    public String getCanonicalHostName() {
	if (canonicalHostName == null) {
	    canonicalHostName = 
		InetAddress.getHostFromNameService(this, true);
	}
	return canonicalHostName;
    }

    /**
     * Returns the hostname for this address.
     * If the host is equal to null, then this address refers to any
     * of the local machine's available network addresses.
     * this is package private so SocketPermission can make calls into
     * here without a security check.
     *
     * <p>If there is a security manager, this method first
     * calls its <code>checkConnect</code> method
     * with the hostname and <code>-1</code> 
     * as its arguments to see if the calling code is allowed to know
     * the hostname for this IP address, i.e., to connect to the host.
     * 
     * @return  the host name for this IP address.
     * 
     * @param check make security check if true
     * 
     * @exception  SecurityException  if a security manager exists and its  
     *  <code>checkConnect</code> method doesn't allow the connection.
     * 
     * @see SecurityManager#checkConnect
     */
    private static String getHostFromNameService(InetAddress addr, boolean check) {
	String host;
	try {
	    // first lookup the hostname
	    host = nameService.getHostByAddr(addr.getAddress());

	    /* check to see if calling code is allowed to know
	     * the hostname for this IP address, ie, connect to the host
	     */
	    if (check) {
		SecurityManager sec = System.getSecurityManager();
		if (sec != null) {
		    sec.checkConnect(host, -1);
		}
	    }

	    /* now get all the IP addresses for this hostname,
	     * and make sure one of them matches the original IP
	     * address. We do this to try and prevent spoofing.
	     */
	    
	    InetAddress[] arr = InetAddress.getAllByName0(host, check);
	    boolean ok = false;

	    if(arr != null) {
		for(int i = 0; !ok && i < arr.length; i++) {
		    ok = addr.equals(arr[i]);
		}
	    }

	    //XXX: if it looks a spoof just return the address?
	    if (!ok) {
		host = addr.getHostAddress();
		return host;
	    }

	} catch (SecurityException e) {
	    host = addr.getHostAddress();
	} catch (UnknownHostException e) {
	    host = addr.getHostAddress();
	}
	return host;
    }

    /**
     * Returns the raw IP address of this <code>InetAddress</code>
     * object. The result is in network byte order: the highest order
     * byte of the address is in <code>getAddress()[0]</code>.
     *
     * @return  the raw IP address of this object.
     */
    public byte[] getAddress() {
	return null;
    }

    /**
     * Returns the IP address string in standard text presentation  form.
     *
     * @return  the raw IP address in a string format.
     * @since   JDK1.0.2
     */
    public String getHostAddress() {
	return null;
     }
    
    /**
     * Returns a hashcode for this IP address.
     *
     * @return  a hash code value for this IP address.
     */
    public int hashCode() {
	return -1;
    }
   
    /**
     * Compares this object against the specified object.
     * The result is <code>true</code> if and only if the argument is
     * not <code>null</code> and it represents the same IP address as
     * this object.
     * <p>
     * Two instances of <code>InetAddress</code> represent the same IP
     * address if the length of the byte arrays returned by
     * <code>getAddress</code> is the same for both, and each of the
     * array components is the same for the byte arrays.
     *
     * @param   obj   the object to compare against.
     * @return  <code>true</code> if the objects are the same;
     *          <code>false</code> otherwise.
     * @see     java.net.InetAddress#getAddress()
     */
    public boolean equals(Object obj) {
	return false;
    }

    /**
     * Converts this IP address to a <code>String</code>. The 
     * string returned is of the form: hostname / literal IP 
     * address.
     *
     * @return  a string representation of this IP address.
     */
    public String toString() {
	return ((hostName != null) ? hostName : "") 
	    + "/" + getHostAddress();
    }

    /*
     * Cached addresses - our own litle nis, not!
     */
    private static LinkedHashMap	    addressCache = new LinkedHashMap();
    static InetAddress	    unknownAddress;
    static InetAddress      localHost;
    static InetAddress[]    unknown_array; // put THIS in cache
    static InetAddress	    anyLocalAddress;
    static InetAddress      loopbackAddress;

    static InetAddressImpl  impl;

    private static HashMap          lookupTable = new HashMap();

    static final class CacheEntry {

	CacheEntry(String hostname, Object address, long expiration) {
	    this.hostname = hostname;
	    this.address = address;
	    this.expiration = expiration;
	}

	String hostname;
	Object address;
	long expiration;
    }

    private static void cacheAddress(String hostname, Object address,
				     boolean success) {

	int policy = (success ? 
		      InetAddressCachePolicy.get() :
		      InetAddressCachePolicy.getNegative());

	// if the cache policy is to cache nothing, just return

	if (policy == 0) {
	    return;
	}

	long expiration = -1;
	if (policy != InetAddressCachePolicy.FOREVER) {
	    expiration = System.currentTimeMillis() + (policy * 1000);
	}
	cacheAddress(hostname, address, expiration);
    }

    private static void cacheAddress(String hostname, Object address, long expiration) {
        hostname = hostname.toLowerCase();
	synchronized (addressCache) {
	    CacheEntry entry = (CacheEntry)addressCache.get(hostname);
	    if (entry == null) {
		entry = new CacheEntry(hostname, address, expiration);
		addressCache.put(hostname, entry);
	    } else {
		if (!entry.address.equals(address)) {
		    entry.address = address;
		    entry.expiration = expiration;
		    // remove the old entry from the cache 
		    // and add a new item
		    addressCache.remove(hostname);
		    addressCache.put(hostname, entry);
		}
	    }
	}
    }

    private static Object getCachedAddress(String hostname) {
        hostname = hostname.toLowerCase();
	if ((InetAddressCachePolicy.get() == 0) && 
	    (InetAddressCachePolicy.getNegative() == 0)) {
	    return null;
	}
	synchronized (addressCache) {
	    // purge expired entries from cache
	    Iterator itr = addressCache.keySet().iterator();
	    // pass the first entry, which is for unknown address
	    itr.next();
	    Vector expired = new Vector();
	    while (itr.hasNext()) {
		Object key = itr.next();
		CacheEntry value = (CacheEntry)addressCache.get(key);
		if (value != null && value.expiration < System.currentTimeMillis() &&
		    value.expiration >= 0) {
		    expired.add(key);
		} else {
		    break;
		}
	    }

	    for (int i = 0; i < expired.size(); i++) {
		Object key = expired.elementAt(i);
		addressCache.remove(key);
	    }

	    CacheEntry entry = (CacheEntry)addressCache.get(hostname);
	    if (entry == null) {
		return null;
	    }
	    return entry.address;
	}
    }

    static {
	/*
	 * Property "impl.prefix" will be prepended to the classname
	 * of the implementation object we instantiate, to which we
	 * delegate the real work (like native methods).  This
	 * property can vary across implementations of the java.
	 * classes.  The default is an empty String "".
	 */
	String prefix = (String)AccessController.doPrivileged(
		      new GetPropertyAction("impl.prefix", ""));

	try {
	    impl = null;
	    impl = (InetAddressImpl)(Class.forName("java.net." + prefix +
						   "InetAddressImpl")
				     .newInstance());
	} catch (ClassNotFoundException e) {
	    System.err.println("Class not found: java.net." + prefix +
			       "InetAddressImpl:\ncheck impl.prefix property " +
			       "in your properties file.");
	} catch (InstantiationException e) {
	    System.err.println("Could not instantiate: java.net." + prefix +
			       "InetAddressImpl:\ncheck impl.prefix property " +
			       "in your properties file.");
	} catch (IllegalAccessException e) {
	    System.err.println("Cannot access class: java.net." + prefix +
			       "InetAddressImpl:\ncheck impl.prefix property " +
			       "in your properties file.");
	}

	if (impl == null) {
	    try {
		impl = (InetAddressImpl)(Class.forName("java.net.InetAddressImpl")
					 .newInstance());
	    } catch (Exception e) {
		throw new Error("System property impl.prefix incorrect");
	    }
	}
	
	if (impl.isIPv6Supported()) {
	    Inet6Address.initStack();
	} else {
	    Inet4Address.initStack();
	}

	if (preferIPv6Address) {
	    Inet6Address.initAddress();
	} else {
	    Inet4Address.initAddress();
	}
	
	cacheAddress(anyLocalAddress.getHostName(), unknown_array, InetAddressCachePolicy.FOREVER);

	// get name service if provided and requested
	String provider = null;;
	String propPrefix = "sun.net.spi.nameservice.provider.";
	int n = 1;
	    while (nameService == null) {
		provider 
		    = (String)AccessController.doPrivileged(
			new GetPropertyAction(propPrefix+n, "default"));
		n++;
		if (provider.equals("default")) {
		    // initialize the default name service
		    nameService = new NameService() {
			public byte[][] lookupAllHostAddr(String host) 
			    throws UnknownHostException {
			    return impl.lookupAllHostAddr(host);
			}
			public String getHostByAddr(byte[] addr) 
			    throws UnknownHostException {
			    return impl.getHostByAddr(addr);
			}
		    };
		    break;
		}
		Iterator itr
		    = Service.providers(NameServiceDescriptor.class);
		while (itr.hasNext()) {
		    NameServiceDescriptor nsd 
			= (NameServiceDescriptor)itr.next();
		    if (provider.
			equalsIgnoreCase(nsd.getType()+","
			       +nsd.getProviderName())) {
			try {
			    nameService = nsd.createNameService();
			    break;
			} catch (Exception e) {
			    e.printStackTrace();
			    System.err.println("Cannot create name service:"
					       +provider);
			}
		    }
		}
	    }
    }
    
    /**
     * Create an InetAddress based on the provided host name and IP address
     * No name service is checked for the validity of the address. 
     *
     * If addr specifies an IPv4 address an instance of Inet4Address 
     * will be returned; otherwise, an instance of Inet6Address 
     * will be returned.
     *
     * IPv4 address byte array must be 4 bytes long and IPv6 byte array 
     * must be 16 bytes long
     *
     * @param host the specified host
     * @param addr the raw IP address in network byte order
     * @return  an InetAddress object created from the raw IP address.
     * @exception  UnknownHostException  if IP address is of illegal length
     * @since 1.4
     */
    public static InetAddress getByAddr(String host, byte[] addr) 
	throws UnknownHostException {
	if (host.charAt(0) == '[') {
	    if (host.charAt(host.length()-1) == ']') {
		host = host.substring(1, host.length() -1);
	    } else {
		throw new UnknownHostException(host);
	    }
	}
	if (addr != null) {
	    if (addr.length == Inet4Address.INADDRSZ) {
		return new Inet4Address(host, addr);
	    } else if (addr.length == Inet6Address.INADDRSZ) {
		byte[] newAddr 
		    = Inet6Address.convertFromIPv4MappedAddress(addr);
		if (newAddr != null) {
		    return new Inet4Address(host, newAddr);
		} else {
		    return new Inet6Address(host, addr);
		}
	    } 
	} 
	throw new UnknownHostException("addr is of illegal length");
    }

    /**
     * Determines the IP address of a host, given the host's name. The
     * host name can either be a machine name, such as
     * "<code>java.sun.com</code>", or a string representing its IP
     * address, such as "<code>206.26.48.100</code>".
     *
     * @param      host   the specified host, or <code>null</code> for the
     *                    local host.
     * @return     an IP address for the given host name.
     * @exception  UnknownHostException  if no IP address for the
     *               <code>host</code> could be found.
     * @exception  SecurityException if a security manager exists
     *             and its checkConnect method doesn't allow the operation 
     */
    public static InetAddress getByName(String host)
	throws UnknownHostException {
	return InetAddress.getAllByName(host)[0];
    }

    /**
     * Determines all the IP addresses of a host, given the host's name.
     * The host name can either be a machine name, such as
     * "<code>java.sun.com</code>", or a string representing
     * its IP address, such as "<code>206.26.48.100</code>".
     *
     * <p>If there is a security manager and <code>host</code> is not 
     * null and <code>host.length() </code> is not equal to zero, the
     * security manager's
     * <code>checkConnect</code> method is called
     * with the hostname and <code>-1</code> 
     * as its arguments to see if the operation is allowed.
     *
     * @param      host   the name of the host.
     * @return     an array of all the IP addresses for a given host name.
     * 
     * @exception  UnknownHostException  if no IP address for the
     *               <code>host</code> could be found.
     * @exception  SecurityException  if a security manager exists and its  
     *               <code>checkConnect</code> method doesn't allow the operation.
     * 
     * @see SecurityManager#checkConnect
     */
    public static InetAddress[] getAllByName(String host)
	throws UnknownHostException {

	if (host == null || host.length() == 0) {
	    InetAddress[] ret = new InetAddress[1];
	    ret[0] = loopbackAddress;
	    return ret;
	}
	
	if (host.charAt(0) == '[') {
	    if (host.charAt(host.length()-1) == ']') {
		host = host.substring(1, host.length() -1);
	    } else {
		throw new UnknownHostException(host);
	    }
	}

	// if host is an IP address, we won't do further lookup
	if (Character.digit(host.charAt(0), 16) != -1 
	    || (host.charAt(0) == ':')) {
	    byte[] addr = null;
	    // see if it is IPv4 address
	    addr = Inet4Address.textToNumericFormat(host);
	    if (addr == null) {
		// see if it is IPv6 address
		addr = Inet6Address.textToNumericFormat(host);
	    }
	    InetAddress[] ret = new InetAddress[1];
	    if(addr != null) {
		if (addr.length == Inet4Address.INADDRSZ) {
		    ret[0] = new Inet4Address(null, addr);
		} else {
		    ret[0] = new Inet6Address(null, addr);
		}
		return ret;
	    }
	}
	return getAllByName0(host);
    }

    private static InetAddress[] getAllByName0 (String host)
	throws UnknownHostException
    {
	return getAllByName0(host, true);
    }

    /**
     * package private so SocketPermission can call it
     */
    static InetAddress[] getAllByName0 (String host, boolean check)
	throws UnknownHostException  {
	/* If it gets here it is presumed to be a hostname */
	/* Cache.get can return: null, unknownAddress, or InetAddress[] */
        Object obj = null;
	Object objcopy = null;

	/* make sure the connection to the host is allowed, before we
	 * give out a hostname
	 */
	if (check) {
	    SecurityManager security = System.getSecurityManager();
	    if (security != null) {
		security.checkConnect(host, -1);
	    }
	}

	obj = getCachedAddress(host);

	/* If no entry in cache, then do the host lookup */
	if (obj == null) {
	    obj = getAddressFromNameService(host);
	}
        if (obj == unknown_array)
            throw new UnknownHostException(host);

	/* Make a copy of the InetAddress array */
	objcopy = ((InetAddress [])obj).clone();

	return (InetAddress [])objcopy;
    }

    private static Object getAddressFromNameService(String host) {
	Object obj = null;
	boolean success = false;

	// Check whether the host is in the lookupTable.
	// 1) If the host isn't in the lookupTable when
	//    checkLookupTable() is called, checkLookupTable()
	//    would add the host in the lookupTable and
	//    return null. So we will do the lookup.
	// 2) If the host is in the lookupTable when
	//    checkLookupTable() is called, the current thread
	//    would be blocked until the host is removed
	//    from the lookupTable. Then this thread
	//    should try to look up the addressCache.
	//     i) if it found the address in the
	//        addressCache, checkLookupTable()  would
	//        return the address.
	//     ii) if it didn't find the address in the
	//         addressCache for any reason,
	//         it should add the host in the
	//         lookupTable and return null so the
	//         following code would do  a lookup itself.
	if ((obj = checkLookupTable(host)) == null) {
	    // This is the first thread which looks up the address 
	    // this host or the cache entry for this host has been
	    // expired so this thread should do the lookup.
	    try {
		/*
		 * Do not put the call to lookup() inside the
		 * constructor.  if you do you will still be
		 * allocating space when the lookup fails.
		 */
		byte[][] byte_array;
		byte_array = nameService.lookupAllHostAddr(host);
		InetAddress[] addr_array =
		    new InetAddress[byte_array.length];

		for (int i = 0; i < byte_array.length; i++) {
		    byte addr[] = byte_array[i];
		    if (addr.length == Inet4Address.INADDRSZ) {
			addr_array[i] = new Inet4Address(host, addr);
		    } else {
			addr_array[i] = new Inet6Address(host, addr);
		    }
		}
		obj = addr_array;
		success = true;
	    } catch (UnknownHostException e) {
		obj  = unknown_array;
		success = false;
	    } finally {
		// Cache the address.
		cacheAddress(host, obj, success);
		// Delete the host from the lookupTable, and
		// notify all threads waiting for the monitor
		// for lookupTable.
		updateLookupTable(host);
	    }
	}

	return obj;
    }
	
		
    private static Object checkLookupTable(String host) {
	// make sure obj  is null.
	Object obj = null;
	
	synchronized (lookupTable) {
	    // If the host isn't in the lookupTable, add it in the
	    // lookuptable and return null. The caller should do
	    // the lookup.
	    if (lookupTable.containsKey(host) == false) {
		lookupTable.put(host, null);
		return obj;
	    }

	    // If the host is in the lookupTable, it means that another
	    // thread is trying to look up the address of this host.
	    // This thread should wait.
	    while (lookupTable.containsKey(host)) {
		try {
		    lookupTable.wait();
		} catch (InterruptedException e) {
		}
	    }
	}

	// The other thread has finished looking up the address of
	// the host. This thread should retry to get the address
	// from the addressCache. If it doesn't get the address from
	// the cache,  it will try to look up the address itself.
	obj = getCachedAddress(host);
	if (obj == null) {
	    synchronized (lookupTable) {
		lookupTable.put(host, null);
	    }
	}
	 
	return obj;
    }

    private static void updateLookupTable(String host) {
	synchronized (lookupTable) {
	    lookupTable.remove(host);
	    lookupTable.notifyAll();
	}
    }

    /**
     * Returns an <code>InetAddress</code> object given the raw IP address . 
     * The argument is in network byte order: the highest order
     * byte of the address is in <code>getAddress()[0]</code>.
     *
     * This method doesn't block, i.e. no reverse NS lookup is performed.
     *
     * IPv4 address byte array must be 4 bytes long and IPv6 byte array 
     * must be 16 bytes long
     *
     * @param addr the raw IP address in network byte order
     * @return  an InetAddress object created from the raw IP address.
     * @since 1.4
     */
    public static InetAddress getByAddr(byte[] addr) 
	throws UnknownHostException {
	if (addr == null || 
	    !(addr.length == Inet4Address.INADDRSZ
	      || addr.length == Inet6Address.INADDRSZ)) {
	    throw new UnknownHostException("addr is of illegal length");
	}
	if (Inet6Address.isIPv4MappedAddress(addr)) {
	    return getByAddr(addr, 12, addr.length - 12);
	} else {
	    return getByAddr(addr, 0, addr.length);
	}
    }

    static InetAddress getByAddr(byte[] addr, int offset, int len) {
	byte[] newAddr = new byte[len];
	System.arraycopy(addr, offset, newAddr, 0, len);
	if (len == Inet4Address.INADDRSZ) {
	    return new Inet4Address(null, newAddr);
	} else {
	    return new Inet6Address(null, newAddr);
	}
    }

    /**
     * Returns the local host.
     *
     * <p>If there is a security manager, its
     * <code>checkConnect</code> method is called
     * with the local host name and <code>-1</code> 
     * as its arguments to see if the operation is allowed.
     *
     * @return     the IP address of the local host.
     * 
     * @exception  UnknownHostException  if no IP address for the
     *               <code>host</code> could be found.
     * @exception  SecurityException  if a security manager exists and its  
     *               <code>checkConnect</code> method doesn't allow the operation.
     * 
     * @see SecurityManager#checkConnect
     */
    public synchronized static InetAddress getLocalHost() throws UnknownHostException {
        /* make sure the connection to the host is allowed: if yes,
	 * return the "real" localHost; if not, return loopback "127.0.0.1"
	 */
	SecurityManager security = System.getSecurityManager();
	try {

	    /* If the localhost's address is not initialized yet, initialize
	     * it.  It is no longer initialized in the static initializer
	     * (see comment there).
	     */

	    if (localHost.address == -1) {
		localHost = getAllByName(localHost.hostName)[0];
		/* This puts it in the address cache as well */
	    }

	    if (security != null)
		security.checkConnect(localHost.getHostName(), -1);
	} catch (java.lang.SecurityException e) {
	    return loopbackAddress;
	}
	return localHost;
    }

    /**
     * Perform class load-time initializations.
     */
    private static native void init();
}

class InetAddressImpl {
    native String getLocalHostName() throws UnknownHostException;
    native byte[][]
        lookupAllHostAddr(String hostname) throws UnknownHostException;
    native String getHostByAddr(byte[] addr) throws UnknownHostException;
    native boolean isIPv6Supported();
}
