/*
 * @(#)Channel.java	1.10 01/05/08
 *
 * Copyright 2000 by Sun Microsystems, Inc.  All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 */

package java.nio.channels;

import java.io.IOException;


/**
 * A nexus for I/O operations.
 *
 * <p> A channel represents an open connection to an entity such as a hardware
 * device, a file, a network socket, or a program component that is capable of
 * performing one or more distinct I/O operations, for example reading or
 * writing.
 *
 * <p> A channel is either open or closed.  A channel is open upon creation,
 * and once closed it remains closed.  Once a channel is closed, any attempt to
 * invoke an I/O operation upon it will cause a {@link ClosedChannelException}
 * to be thrown.  Whether or not a channel is open may be tested by invoking
 * its {@link #isOpen isOpen} method.
 *
 * <p> Channels are <i>asynchronously closeable:</i> If a thread is blocked in
 * an I/O operation on a channel then another thread may invoke the channel's
 * {@link #close close} method.  This will cause the blocked thread to receive
 * an {@link AsynchronousCloseException}.
 *
 * <p> Channels are also <i>interruptible:</i> If a thread is blocked in an I/O
 * operation on a channel then another thread may invoke the blocked thread's
 * {@link Thread#interrupt() interrupt} method.  This will cause the channel to
 * be closed, the blocked thread to receive a {@link
 * ClosedByInterruptException}, and the blocked thread's interrupt status to be
 * set.
 *
 * <p> If a thread's interrupt status is already set and it invokes a blocking
 * I/O operation upon a channel then the channel will be closed and the thread
 * will immediately receive a {@link ClosedByInterruptException}; its interrupt
 * status will remain set.
 *
 * <p> Channels are, in general, intended to be safe for multithreaded access
 * as described in the specifications of the interfaces and classes that extend
 * and implement this interface.
 *
 *
 * <a name="dn">
 * <h4>Design note: Interrupts</h4>
 *
 * <p> That channels are asynchronously closeable and interruptible arises from
 * the practical need to be able to break threads out of blocking I/O
 * operations.
 *
 * <p> The {@link java.io} package specifies an {@link
 * java.io.InterruptedIOException} for this purpose, the idea being that
 * invoking a thread's {@link Thread#interrupt() interrupt} method will cause
 * the thread to break out of a blocking I/O operation and receive this
 * exception.  Unfortunately this approach has turned out to be impossible to
 * implement both efficiently and reliably on most popular operating systems,
 * and in fact it is well known not to work uniformly across different
 * implementations of the Java platform.
 *
 * <p> A technique that does work, however, is to close the file descriptor or
 * handle that is the subject of the operation in which the target thread is
 * blocked.  Despite the fact that this closes the corresponding stream or
 * channel, this behavior is often acceptable because the most common reason
 * for interrupting a thread is to ask it to shut down.  It is also difficult
 * in practice to write systems that deal gracefully at a high level with
 * partly-completed I/O operations; it is often simply easier to discard a
 * stream or channel that's in an unknown state and recover in some other way,
 * <i>e.g.</i>, by opening a new connection.  Hence this technique has come to
 * be known as a portable workaround for the lack of true interruptible I/O in
 * the {@link java.io} package.
 *
 * <p> The approach taken here, then, is to codify the close-on-interrupt
 * technique in the {@link java.nio} package.  The closing-on-interrupt of a
 * channel is treated as a special case of asynchronously closing a channel.
 * Each condition is given a distinct exception because application code may
 * want to treat the interrupting case specially. </p>
 *
 *
 * @author Mark Reinhold
 * @version 1.10, 01/05/08
 * @since 1.4
 */

public interface Channel {

    /**
     * Tells whether or not this channel is open.  </p>
     *
     * @return <tt>true</tt> if, and only if, this channel is open
     */
    public boolean isOpen();

    /**
     * Closes this channel.
     *
     * <p> Any thread currently blocked in an I/O operation upon this channel
     * will receive an {@link AsynchronousCloseException}.
     *
     * <p> After a channel is closed, any further attempt to invoke I/O
     * operations upon it will cause a {@link ClosedChannelException} to be
     * thrown.
     *
     * <p> If this channel is already closed then invoking this method has no
     * effect.
     *
     * <p> This method may be invoked at any time.  If some other thread has
     * already invoked it, however, then another invocation will block until
     * the first invocation is complete, after which it will return without
     * effect. </p>
     *
     * @throws  IOException  If an I/O error occurs
     */
    public void close() throws IOException;

}
