/*
 * @(#)ServerSocketChannel.java	1.14 01/05/08
 *
 * Copyright 2000 by Sun Microsystems, Inc.  All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 */

package java.nio.channels;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.nio.channels.spi.*;


/**
 * A selectable channel for stream-oriented listening sockets.
 *
 * <p> Server-socket channels are created by invoking the {@link #open open}
 * method of this class, which also creates an associated {@link
 * java.net.ServerSocket} object.  The socket encapsulates addressing and other
 * protocol-specific information, including socket options, but delegates all
 * of its I/O operations to the channel.  It is not possible to create a
 * channel for an arbitrary, pre-existing server socket, nor is it possible to
 * specify the {@link java.net.SocketImpl} object to be used by a socket
 * channel's socket.
 *
 * <p> Server-socket channels support only one I/O operation, namely that of
 * accepting a new connection.  They are safe for use by multiple concurrent
 * threads.
 *
 * <p> Server-socket channels support timeouts on accept operations.  If a
 * server-socket channel is in blocking mode and a timeout period has been set
 * by invoking its socket's {@link java.net.Socket#setSoTimeout(int)
 * setSoTimeout} method, then a {@link TimeoutException} is thrown if the
 * {@link #accept accept} method is invoked but fails to complete before the
 * timeout period expires.  </p>
 *
 * @author Mark Reinhold
 * @version 1.14, 01/05/08
 * @since 1.4
 *
 * @see java.net.ServerSocket#getChannel()
 */

public abstract class ServerSocketChannel
    extends AbstractSelectableChannel
{

    /**
     * Initializes a new instance of this class.
     */
    protected ServerSocketChannel() { }

    /**
     * Opens a server-socket channel.
     *
     * <p> The new channel is created by invoking the {@link
     * java.nio.channels.spi.SelectorProvider#openServerSocketChannel
     * openServerSocketChannel} method of the system-wide default {@link
     * java.nio.channels.spi.SelectorProvider} object.
     *
     * <p> The new channel's socket is initially unbound; it must be bound to a
     * specific address via one of its socket's {@link
     * java.net.ServerSocket#bind(SocketAddress) bind} methods before
     * connections can be accepted.  </p>
     *
     * @return  A new socket channel
     *
     * @throws  IOException
     *          If an I/O error occurs
     */
    public static ServerSocketChannel open() throws IOException {
	return SelectorProvider.provider().openServerSocketChannel();
    }

    /**
     * Returns an operation set identifying this channel's supported
     * operations.
     *
     * <p> Server-socket channels only support the accepting of new
     * connections, so this method returns {@link SelectionKey#OP_ACCEPT}.
     * </p>
     *
     * @return  The valid-operation set
     */
    public final int validOps() {
	return SelectionKey.OP_ACCEPT;
    }


    // -- ServerSocket-specific operations --

    /**
     * Retrieves the server socket associated with this channel.  </p>
     *
     * @return  The server socket
     */
    public abstract ServerSocket socket();

    /**
     * Accepts a connection made to this channel's socket.
     *
     * <p> If this channel is in non-blocking mode then this method will
     * immediately return <tt>null</tt> if there are no pending connections.
     * Otherwise it will block indefinitely, waiting for a new connection or
     * for the timeout period, if there is one, to expire.
     *
     * <p> This method performs exactly the same security checks as the {@link
     * java.net.ServerSocket#accept accept} method of the {@link
     * java.net.ServerSocket} class.  That is, if a security manager has been
     * installed then for each new connection this method verifies that the
     * address and port number of the connection's remote endpoint are
     * permitted by the security manager's {@link
     * java.lang.SecurityManager#checkAccept checkAccept} method.  </p>
     *
     * @return  The socket for the new connection,
     *          or <tt>null</tt> if this channel is in non-blocking mode
     *          and no connection is available to be accepted
     *
     * @throws  ClosedChannelException
     *          If this channel is closed
     *
     * @throws  AsynchronousCloseException
     *          If another thread closes this channel
     *          while the accept operation is in progress
     *
     * @throws  ClosedByInterruptException
     *          If another thread interrupts the current thread
     *          while the accept operation is in progress, thereby
     *          closing the channel and setting the current thread's
     *          interrupt status
     *
     * @throws  TimeoutException
     *          If this channel is in blocking mode, its socket's timeout value 
     *          is nonzero, and the timeout period expires before any
     *          connections become available
     *
     * @throws  SecurityException
     *          If a security manager has been installed
     *          and it does not permit access to the remote endpoint
     *          of the new connection
     *
     * @throws  IOException
     *          If some other I/O error occurs
     */
    public abstract Socket accept() throws IOException;

}
