/*
 * @(#)SelectorProvider.java	1.9 01/05/08
 *
 * Copyright 2000 by Sun Microsystems, Inc.  All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 */

package java.nio.channels.spi;

import java.io.FileDescriptor;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.nio.channels.*;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Iterator;
import sun.misc.Service;
import sun.misc.ServiceConfigurationError;
import sun.security.action.GetPropertyAction;


/**
 * Service-provider class for selectors and selectable channels.
 *
 * <p> A selector provider is a concrete subclass of this class that has a
 * zero-argument constructor and implements the abstract methods specified
 * below.  A given invocation of the Java virtual machine maintains a single
 * provider instance, which is returned by the {@link #provider provider}
 * method.  The first invocation of that method will locate the provider object
 * as specified below.
 *
 * <p> All of the methods in this class are safe for use by multiple concurrent
 * threads.  </p>
 *
 *
 * @author Mark Reinhold
 * @version 1.9, 01/05/08
 * @since 1.4
 */

public abstract class SelectorProvider {

    private static final Object lock = new Object();
    private static SelectorProvider provider = null;

    /**
     * Initializes a new instance of this class.  </p>
     */
    protected SelectorProvider() { }

    /**
     * Sets the system-wide selector provider.  </p>
     *
     * @param  provider
     *         The provider to be used; must not be <tt>null</tt>
     *
     * @throws  IllegalStateException
     *          If a provider for this invocation of the Java virtual machine
     *          has already been determined
     *
     * @throws  SecurityException
     *          If a security manager has been installed and it denies
     *          {@link RuntimePermission}<tt>("selectorProvider")</tt>
     */
    public static void setProvider(SelectorProvider provider) {
	if (provider == null)
	    throw new IllegalArgumentException();
	synchronized (lock) {
	    if (SelectorProvider.provider != null)
		throw new IllegalStateException();
	    SecurityManager sm = System.getSecurityManager();
	    if (sm != null)
		sm.checkPermission(new RuntimePermission("selectorProvider"));
	    SelectorProvider.provider = provider;
	}
    }

    private static boolean loadProviderFromProperty() {
	PrivilegedAction pa
	    = new GetPropertyAction("java.nio.channels.spi.SelectorProvider");
	String cn = (String) AccessController.doPrivileged(pa);
	if (cn == null)
	    return false;
	try {
	    Class c = Class.forName(cn); 	// ## Need privs?
	    provider = (SelectorProvider)c.newInstance();
	    return true;
	} catch (ClassNotFoundException x) {
	    throw new ServiceConfigurationError(x);
	} catch (IllegalAccessException x) {
	    throw new ServiceConfigurationError(x);
	} catch (InstantiationException x) {
	    throw new ServiceConfigurationError(x);
	}
    }

    private static boolean loadProviderAsService() {
	Iterator i = Service.providers(SelectorProvider.class, null);
	if (!i.hasNext())
	    return false;
	provider = (SelectorProvider)i.next();
	return true;
    }

    /**
     * Returns the selector provider for this invocation of the Java virtual
     * machine.
     *
     * <p> The first invocation of this method locates the system-wide provider
     * object as follows:  </p>
     *
     * <ol>
     *
     *   <li><p> If the {@link #setProvider setProvider} method has been
     *   succesfully invoked then the supplied provider is returned.  </p></li>
     *
     *   <li><p> If the system property <tt>java.nio.spi.SelectorProvider</tt>
     *   is defined then it is taken to be the fully-qualified name of a
     *   concrete provider class.  The class is loaded and instantiated; if
     *   this process fails then an unspecified error is thrown.  </p></li>
     *
     *   <li><p> If a provider class has been installed in a jar file that is
     *   visible to the system class loader, and that jar file contains a
     *   provider-configuration file named
     *   <tt>java.nio.spi.SelectorProvider</tt> in the resource directory
     *   <tt>META-INF/services</tt>, then the first class name specified in
     *   that file is taken.  The class is loaded and instantiated; if this
     *   process fails then an unspecified error is thrown.  </p></li>
     *
     *   <li><p> Finally, if no provider has been specified by any of the above
     *   means then the system-default provider class is instantiated and the
     *   result is returned.  </p></li>
     *
     * </ol>
     *
     * <p> Subsequent invocations of this method return the provider that was
     * returned by the first invocation.  </p>
     *
     * @return  The system-wide selector provider
     */
    public static SelectorProvider provider() {
	synchronized (lock) {
	    if (provider != null)
		return provider;
	    if (loadProviderFromProperty())
		return provider;
	    if (loadProviderAsService())
		return provider;
	    provider = new sun.nio.ch.SelectorProviderImpl();
	    return provider;
	}
    }

    /**
     * Opens a datagram channel.  </p>
     *
     * @param  port
     *         The local port to which the channel is to be bound,
     *         or zero if an ephemeral port is to be used; must be
     *         a non-negative integer
     *
     * @return  The new channel
     */
    public abstract DatagramChannel openDatagramChannel(int port)
	throws IOException;

    /**
     * Opens a pipe.  </p>
     *
     * @return  The new pipe
     */
    public abstract Pipe openPipe()
	throws IOException;

    /**
     * Opens a selector.  </p>
     *
     * @return  The new selector
     */
    public abstract AbstractSelector openSelector()
	throws IOException;

    /**
     * Opens a server-socket channel.  </p>
     *
     * @return  The new channel
     */
    public abstract ServerSocketChannel openServerSocketChannel()
	throws IOException;

    /**
     * Opens a socket channel. </p>
     *
     * @return  The new channel
     */
    public abstract SocketChannel openSocketChannel()
	throws IOException;

}
