/*
 * @(#)Charset-X-Coder.java	1.23 01/05/08
 *
 * Copyright 2000 by Sun Microsystems, Inc.  All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 */

// -- This file was mechanically generated: Do not edit! -- //

package java.nio.charset;

import java.nio.Buffer;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.BufferOverflowException;
import java.nio.BufferUnderflowException;


/**
 * An engine that can transform a sequence of bytes in a specific charset into a sequence of
 * sixteen-bit Unicode characters.
 *
 * <a name="steps">
 *
 * <p> The input byte sequence is provided in a byte buffer or a series
 * of such buffers.  The output character sequence is written to a character buffer
 * or a series of such buffers.  A decoder should always be used by making
 * the following sequence of method invocations, hereinafter referred to as a
 * <i>decoding operation</i>:
 *
 * <ol>
 *
 *   <li><p> Reset the decoder via the {@link #reset reset} method, unless it
 *   has not been used before; </p></li>
 *
 *   <li><p> Invoke the {@link #decode decode} method zero or more times, as
 *   long as additional input may be available, passing <tt>false</tt> for the
 *   <tt>endOfInput</tt> argument and filling the input buffer and flushing the
 *   output buffer between invocations; </p></li>
 *
 *   <li><p> Invoke the {@link #decode decode} method one final time, passing
 *   <tt>true</tt> for the <tt>endOfInput</tt> argument; and then </p></li>
 *
 *   <li><p> Invoke the {@link #flush flush} method so that the decoder can
 *   flush any internal state to the output buffer. </p></li>
 *
 * </ol>
 *
 * Each invocation of the {@link #decode decode} method will decode as many
 * bytes as possible from the input buffer, writing the resulting characters
 * to the output buffer.  The {@link #decode decode} method will throw an
 * exception when more input is required, when there is not enough room in the
 * output buffer, or when a decoding error occurs.  When it throws an
 * exception it is guaranteed that the positions of the input and output
 * buffers have their original values.  Hence the invoker can fill the input
 * buffer, flush the output buffer, or attempt to recover from a decoding
 * error, as appropriate, and try again.
 *
 * <p> There are two general types of decoding errors.  If the input byte
 * sequence is not legal for this charset then the input is considered <i>malformed</i>.  If
 * the input byte sequence is legal but cannot be mapped to a valid
 * Unicode character then an <i>unmappable character</i> has been encountered.
 *
 * <p> How decoding errors are handled depends upon whether the
 * <i>substitution</i> is defined.  The substitution is a string that
 *





 * is initially set to <tt>"&#92;uFFFD"</tt>;

 *
 * its value may be changed via the {@link #substitute(java.lang.String)
 * substitute} method.  If the substitution is defined then it is copied to the
 * output buffer each time a decoding error occurs.  If the substitution is
 * not defined then an appropriate exception, either {@link
 * MalformedInputException} or {@link UnmappableCharacterException}, is thrown
 * when such an error occurs.
 *
 * <p> This class is designed to handle all of the bookkeeping required by the
 * decoding process.  A decoder for a specific charset, which is a concrete
 * subclass of this class, need only implement the abstract {@link #decodeStep
 * decodeStep} method, which encapsulates a basic decoding step.  A subclass
 * that maintains internal state should, additionally, override the {@link
 * #flush flush} and {@link #reset reset} methods.
 *
 * <p> Instances of this class are not safe for use by multiple concurrent
 * threads.  </p>
 *
 *
 * @version 1.23, 01/05/08
 * @author Mark Reinhold
 * @since 1.4
 *
 * @see ByteBuffer
 * @see CharBuffer
 * @see Charset
 * @see CharsetEncoder
 */

public abstract class CharsetDecoder {

    private String subst;
    private float averageCharsPerByte;
    private float maxCharsPerByte;

    /**
     * Initializes a new decoder.  The new decoder will have the given
     * chars-per-byte and substitution values. </p>
     *
     * @param  averageCharsPerByte
     *         A positive float value indicating the expected number of
     *         characters that will be produced for each input byte
     *
     * @param  maxCharsPerByte
     *         A positive float value indicating the maximum number of
     *         characters that will be produced for each input byte
     *
     * @param  substitution
     *         The initial substitution; may not be <tt>null</tt>
     */
    private
    CharsetDecoder(float averageCharsPerByte,
		   float maxCharsPerByte,
		   String substitution)
    {
	if (substitution == null)
	    throw new IllegalArgumentException("Null substitution");
	if (averageCharsPerByte <= 0.0f)
	    throw new IllegalArgumentException("Non-positive "
					       + "averageCharsPerByte");
	if (maxCharsPerByte <= 0.0f)
	    throw new IllegalArgumentException("Non-positive "
					       + "maxCharsPerByte");
	subst = substitution;
	this.averageCharsPerByte = averageCharsPerByte;
	this.maxCharsPerByte = maxCharsPerByte;
    }

    /**
     * Initializes a new decoder.  The new decoder will have the given
     * chars-per-byte values and its substitution will be the
     * string <tt>"&#92;uFFFD"</tt>. </p>
     *
     * @param  averageCharsPerByte
     *         A positive float value indicating the expected number of
     *         characters that will be produced for each input byte
     *
     * @param  maxCharsPerByte
     *         A positive float value indicating the maximum number of
     *         characters that will be produced for each input byte
     */
    protected CharsetDecoder(float averageCharsPerByte,
			     float maxCharsPerByte)
    {
	this(averageCharsPerByte, maxCharsPerByte,
	     "\uFFFD");
    }

    /**
     * Returns this decoder's substitution, or <tt>null</tt> if the
     * substitution is not defined. </p>
     *
     * @return  The current substitution of this decoder;
     *          may be <tt>null</tt>
     */
    public final String substitution() {
	return subst;
    }

    /**
     * Sets this decoder's substitution.  To disable substitution,
     * set the substitution to <tt>null</tt>. </p>
     *
     * @param  newSubstitution
     *         The new substitution, or <tt>null</tt> if
     *         substitution is not to be performed
     */
    public final void substitute(String newSubstitution) {
	subst = newSubstitution;
    }

    /**
     * Returns the average number of characters that will be produced for each
     * byte of input.  This value may be used to estimate the size of the
     * output buffer required for a given input sequence. </p>
     *
     * @return  The average number of characters produced
     *          per byte of input
     */
    public final float averageCharsPerByte() {
	return averageCharsPerByte;
    }

    /**
     * Returns the maximum number of characters that will be produced for each
     * byte of input.  This value may be used to compute the worst-case size
     * of the output buffer required for a given input sequence. </p>
     *
     * @return  The maximum number of characters produced per byte of input
     */
    public final float maxCharsPerByte() {
	return maxCharsPerByte;
    }

    /**
     * Decodes as many bytes as possible from the given input buffer,
     * writing the results to the given output buffer.
     *
     * <p> The buffers are read from, and written to, starting at their current
     * positions.  At most {@link Buffer#remaining in.remaining()} bytes
     * will be read, and at most {@link Buffer#remaining out.remaining()}
     * characters will be written.  The buffers' positions will be advanced to
     * reflect the bytes read and the characters written, but their marks and
     * limits will not be modified.
     *
     * <p> This method either completes successfully, having read at least one
     * byte from the input buffer and written zero or more characters to the
     * output buffer, or it throws one of the exceptions listed below.  If it
     * throws an exception then no bytes will have been consumed, no
     * characters will have been produced, and the buffers' positions will have
     * been restored to their initial values.
     *
     * <p> The <tt>endOfInput</tt> parameter advises this method as to whether
     * the invoker can provide further input beyond that contained in the given
     * input buffer.  If there is a possibility of providing additional input
     * then the invoker should pass <tt>false</tt> for this parameter; if there
     * is no possibility of providing further input then the invoker should
     * pass <tt>true</tt>.  It is not erroneous, and in fact it is quite
     * common, to pass <tt>false</tt> in one invocation and later discover that
     * no further input was actually available.  It is critical, however, that
     * the final invocation of this method in a sequence of invocations always
     * pass <tt>true</tt> so that a {@link MalformedInputException} rather than
     * a {@link BufferUnderflowException} can be thrown if there is
     * insufficient input to complete the <a href="#steps">decoding
     * operation</a>.
     *
     * <p> If the substitution is defined then it will be copied to the output
     * buffer each time a decoding error occurs; if the substitution is not
     * defined then an appropriate exception will be thrown.
     *
     * <p> This method works by repeatedly invoking the {@link #decodeStep
     * decodeStep} method.  If that method throws an exception before producing
     * any output then the loop is terminated, the buffers' positions are
     * reset, and the exception is rethrown.  If that method throws an
     * exception after some output has been produced then the loop is
     * terminated, the buffers are reset to their states prior to the most
     * recent step, and the method returns normally. </p>
     *
     * @param  in
     *         The input byte buffer
     *
     * @param  out
     *         The output character buffer
     *
     * @param  endOfInput
     *         <tt>true</tt> if, and only if, this is the final invocation of
     *         this method in the current decoding operation
     *
     * @return  <tt>false</tt> if, and only if, the invocation terminated
     *          after decoding one or more bytes because the next
     *          decoding step would have thrown a character-coding exception
     *
     * @throws  BufferUnderflowException
     *          If more input bytes are required.  Any bytes remaining in
     *          the given input buffer should be preserved so that they are the
     *          first bytes read by the next invocation of this method.
     *
     * @throws  BufferOverflowException
     *          If there is insufficient room in the output buffer
     *
     * @throws  IllegalStateException
     *          If the previous step of the current decoding operation was an
     *          invocation neither of the {@link #reset reset} method nor of
     *          this method with a value of <tt>false</tt> for the
     *          <tt>endOfInput</tt> parameter
     *
     * @throws  MalformedInputException
     *          If the byte sequence starting at the input buffer's current
     *          position is not legal for this charset and the substitution is not defined
     *
     * @throws  UnmappableCharacterException
     *          If the byte sequence starting at the input buffer's current
     *          position cannot be mapped to an equivalent character sequence and 
     *          the substitution is not defined
     */
    public final boolean decode(ByteBuffer in, CharBuffer out,
				boolean endOfInput)
	throws CharacterCodingException
    {
	if (!in.hasRemaining())
	    throw new BufferUnderflowException();
	int inLimit = in.limit();
	int outStart = out.position();
	int inMark = 0;
	int outMark = 0;
	try {
	    try {
		for (;;) {
		    inMark = in.position();
		    outMark = out.position();
		    if (inMark >= inLimit)
			// Avoid creating a new exception object
			throw bufferUnderflowException;
		    try {
			decodeStep(in, out);
		    } catch (CharacterCodingException x) {
			if (subst == null)
			    throw x;
			// Insert the substitution and skip over
			// the troublesome input
			out.put(subst);
			int n = ((x instanceof MalformedInputException)
				 ? ((MalformedInputException)x).getInputLength()
				 : 1);
			n = Math.min(in.remaining(), n);
			if (n == 0)
			    return true;
			in.position(in.position() + n);
		    }
		}
	    } catch (BufferUnderflowException x) {
		if (!handle(in, out, outStart, inMark, outMark)) {
		    if (!endOfInput)
			throw x;
		    if (subst != null)
			out.put(subst);
		    else
			throw new MalformedInputException(in.remaining());
		}
		return true;
	    } catch (CharacterCodingException x) {
		if (!handle(in, out, outStart, inMark, outMark))
		    throw x;
		return false;
	    }
	} catch (BufferOverflowException x) {
	    if (!handle(in, out, outStart, inMark, outMark))
		throw x;
	    return true;
	}
    }

    // Cached exception used above in order to avoid creating a new exception
    // object in every invocation of the decode method.  Some clients, e.g.,
    // java.io.PrintStream in auto-flushing mode, code many very small input
    // sequences.
    //
    private static BufferUnderflowException bufferUnderflowException
	= new BufferUnderflowException();

    // Utility method for handling exceptions
    //
    private static boolean handle(Buffer in, Buffer out,
				  int outStart, int inMark, int outMark)
    {
	in.position(inMark);
	out.position(outMark);
	return (out.position() > outStart);
    }

    /**
     * Flushes this decoder.
     *
     * <p> Some decoders maintain internal state and may need to write some
     * final characters to the output buffer once the overall input sequence has
     * been read.
     *
     * <p> Any additional output is written to the output buffer beginning at
     * its current position.  At most {@link Buffer#remaining out.remaining()}
     * characters will be written.  The buffer's position will be advanced
     * appropriately, but its mark and limit will not be modified.
     *
     * <p> This method either completes successfully, having written zero or
     * more characters to the output buffer, or it throws a {@link
     * BufferOverflowException}.  If it throws an exception then no characters
     * will have been written to the output buffer.  If this happens then this
     * method must be invoked again, with an output buffer that has more room,
     * in order to complete the current <a href="#steps">decoding
     * operation</a>.
     *
     * <p> The default implementation of this method does nothing.  This method
     * should be overridden by decoders that may need to write final characters
     * to the output buffer once the input sequence has been read. </p>
     *
     * @param  out
     *         The output character buffer
     *
     * @throws  BufferOverflowException
     *          If there is insufficient room in the output buffer for all of
     *          the characters that must be written
     *
     * @throws  IllegalStateException
     *          If the previous step of the current decoding operation was an
     *          invocation neither of the {@link #reset reset} method nor of
     *          the three-argument {@link
     *          #decode(ByteBuffer,CharBuffer,boolean) decode} method
     *          with a value of <tt>true</tt> for the <tt>endOfInput</tt>
     *          parameter
     */
    public void flush(CharBuffer out) { }

    /**
     * Resets this decoder, clearing any internal state.
     *
     * <p> The default implementation of this method does nothing.  This method
     * should be overridden by decoders that maintain internal state. </p>
     */
    public void reset() { }

    /**
     * Decodes one or more bytes into one or more characters.
     *
     * <p> This method performs a single decoding step; it is intended to be
     * invoked by the {@link #decode decode} method, which handles the
     * bookkeeping involved in the decoding process.  This method does not
     * attempt to decode the entire input buffer; rather, it decodes as few
     * bytes as possible.
     *
     * <p> Upon invocation this method assumes that there is at least one
     * byte remaining in the input buffer.  This method may advance the
     * position of each buffer, but it does not change their marks or limits.
     *
     * <p> An invocation of method either completes successfully, having
     * consumed one or more bytes from the input buffer and written zero or
     * more characters to the output buffer, or it throws one of the exceptions
     * listed below.
     *
     * <p> If an invocation of this method throws an exception because a
     * decoding error occurs then the invoking {@link #decode decode} method
     * will check whether the substitution is defined.  If so then it will
     * insert the substitution and continue the <a href="#steps">decoding
     * operation</a>.  Otherwise, and for all other types of exceptions, the
     * {@link #decode decode} method will reset the buffers' positions to the
     * values they had prior to the failing invocation and then propagate the
     * exception.
     *
     * <p> An implementation of this method may perform arbitrary lookahead
     * simply by reading forward in the input buffer.  If there is insufficient
     * input then a {@link BufferUnderflowException} will be thrown and the
     * {@link #decode decode} method will re-invoke this method with additional
     * input if any is available. </p>
     *
     * @param  in
     *         The input byte buffer; it must have at least one remaining
     *         byte, and its mark must be equal to its position
     *
     * @param  out
     *         The output character buffer; its mark must be equal to its
     *         position
     *
     * @throws  BufferUnderflowException
     *          If more input bytes are required.  Any bytes remaining in
     *          the given input buffer should be preserved so that they are the
     *          first bytes read by the next invocation of this method.
     *
     * @throws  BufferOverflowException
     *          If there is insufficient room in the output buffer
     *
     * @throws  MalformedInputException
     *          If the byte sequence starting at the input buffer's current
     *          position is not legal for this charset and the substitution is not defined
     *
     * @throws  UnmappableCharacterException
     *          If the byte sequence starting at the input buffer's current
     *          position cannot be mapped to an equivalent character sequence and 
     *          the substitution is not defined
     */
    protected abstract void decodeStep(ByteBuffer in, CharBuffer out)
	throws CharacterCodingException;

    /**
     * Convenience method that decodes the entire content of a single input
     * byte buffer into a newly-allocated character buffer.
     *
     * <p> This method implements an entire <a href="#steps">decoding
     * operation</a>; that is, it resets this decoder, then it decodes the
     * bytes in the given byte buffer, and finally it flushes this
     * decoder.  This method should therefore not be invoked if a decoding
     * operation is already in progress.  </p>
     *
     * @param  in
     *         The input byte buffer
     *
     * @return A newly-allocated character buffer containing the result of the
     *         decoding operation.  The buffer's position will be zero and its
     *         limit will follow the last character written.
     *
     * @throws  IllegalStateException
     *          If a decoding operation is already in progress
     *
     * @throws  MalformedInputException
     *          If the byte sequence starting at the input buffer's current
     *          position is not legal for this charset and the substitution is not defined
     *
     * @throws  UnmappableCharacterException
     *          If the byte sequence starting at the input buffer's current
     *          position cannot be mapped to an equivalent character sequence and 
     *          the substitution is not defined
     */
    public final CharBuffer decode(ByteBuffer in)
	throws CharacterCodingException
    {
	int n = (int)(in.remaining() * averageCharsPerByte());
	CharBuffer out = CharBuffer.allocate(n);
	for (;;) {
	    try {
		if (in.hasRemaining())
		    decode(in, out, true);
		else {
		    flush(out);
		    break;
		}
	    } catch (BufferOverflowException x) {
		n *= 2;
		CharBuffer o = CharBuffer.allocate(n);
		out.flip();
		o.put(out);
		out = o;
	    }
	}
	out.flip();
	return out;
    }




    /**
     * Retrieves the charset that was detected by this decoder.
     *
     * <p> If this decoder implements an auto-detecting charset then this
     * method returns the actual charset once it has been detected.  If not
     * enough input bytes have yet been read to determine the actual charset,
     * or if this decoder does not implement an auto-detecting charset, then
     * this method returns <tt>null</tt>.
     *
     * <p> The default implementation of this method always returns
     * <tt>null</tt>; it should be overridden by auto-detecting decoders to
     * return the appropriate value.  </p>
     *
     * @return  The charset detected by this auto-detecting decoder,
     *          or <tt>null</tt> if the charset is not yet known or if this
     *          is not an auto-detecting decoder
     */
    public Charset detectedCharset() {
	return null;
    }


















































































}
