/*
 * @(#)Charset-X-Coder.java	1.23 01/05/08
 *
 * Copyright 2000 by Sun Microsystems, Inc.  All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 */

// -- This file was mechanically generated: Do not edit! -- //

package java.nio.charset;

import java.nio.Buffer;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.BufferOverflowException;
import java.nio.BufferUnderflowException;


/**
 * An engine that can transform a sequence of sixteen-bit Unicode characters into a sequence of
 * bytes in a specific charset.
 *
 * <a name="steps">
 *
 * <p> The input character sequence is provided in a character buffer or a series
 * of such buffers.  The output byte sequence is written to a byte buffer
 * or a series of such buffers.  An encoder should always be used by making
 * the following sequence of method invocations, hereinafter referred to as an
 * <i>encoding operation</i>:
 *
 * <ol>
 *
 *   <li><p> Reset the encoder via the {@link #reset reset} method, unless it
 *   has not been used before; </p></li>
 *
 *   <li><p> Invoke the {@link #encode encode} method zero or more times, as
 *   long as additional input may be available, passing <tt>false</tt> for the
 *   <tt>endOfInput</tt> argument and filling the input buffer and flushing the
 *   output buffer between invocations; </p></li>
 *
 *   <li><p> Invoke the {@link #encode encode} method one final time, passing
 *   <tt>true</tt> for the <tt>endOfInput</tt> argument; and then </p></li>
 *
 *   <li><p> Invoke the {@link #flush flush} method so that the encoder can
 *   flush any internal state to the output buffer. </p></li>
 *
 * </ol>
 *
 * Each invocation of the {@link #encode encode} method will encode as many
 * characters as possible from the input buffer, writing the resulting bytes
 * to the output buffer.  The {@link #encode encode} method will throw an
 * exception when more input is required, when there is not enough room in the
 * output buffer, or when an encoding error occurs.  When it throws an
 * exception it is guaranteed that the positions of the input and output
 * buffers have their original values.  Hence the invoker can fill the input
 * buffer, flush the output buffer, or attempt to recover from an encoding
 * error, as appropriate, and try again.
 *
 * <p> There are two general types of encoding errors.  If the input character
 * sequence is not a legal sixteen-bit Unicode sequence then the input is considered <i>malformed</i>.  If
 * the input character sequence is legal but cannot be mapped to a valid
 * byte sequence in the given charset then an <i>unmappable character</i> has been encountered.
 *
 * <p> How encoding errors are handled depends upon whether the
 * <i>substitution</i> is defined.  The substitution is a byte array that
 *

 * is initially set to the encoder's default substitution, which is often
 * (but not always) the byte array&nbsp;<tt>{</tt>&nbsp;<tt>(byte)'?'</tt>&nbsp;<tt>}</tt>;




 *
 * its value may be changed via the {@link #substitute(byte[])
 * substitute} method.  If the substitution is defined then it is copied to the
 * output buffer each time an encoding error occurs.  If the substitution is
 * not defined then an appropriate exception, either {@link
 * MalformedInputException} or {@link UnmappableCharacterException}, is thrown
 * when such an error occurs.
 *
 * <p> This class is designed to handle all of the bookkeeping required by the
 * encoding process.  An encoder for a specific charset, which is a concrete
 * subclass of this class, need only implement the abstract {@link #encodeStep
 * encodeStep} method, which encapsulates a basic encoding step.  A subclass
 * that maintains internal state should, additionally, override the {@link
 * #flush flush} and {@link #reset reset} methods.
 *
 * <p> Instances of this class are not safe for use by multiple concurrent
 * threads.  </p>
 *
 *
 * @version 1.23, 01/05/08
 * @author Mark Reinhold
 * @since 1.4
 *
 * @see ByteBuffer
 * @see CharBuffer
 * @see Charset
 * @see CharsetDecoder
 */

public abstract class CharsetEncoder {

    private byte[] subst;
    private float averageBytesPerChar;
    private float maxBytesPerChar;

    /**
     * Initializes a new encoder.  The new encoder will have the given
     * bytes-per-char and substitution values. </p>
     *
     * @param  averageBytesPerChar
     *         A positive float value indicating the expected number of
     *         bytes that will be produced for each input character
     *
     * @param  maxBytesPerChar
     *         A positive float value indicating the maximum number of
     *         bytes that will be produced for each input character
     *
     * @param  substitution
     *         The initial substitution; may not be <tt>null</tt>
     */
    protected
    CharsetEncoder(float averageBytesPerChar,
		   float maxBytesPerChar,
		   byte[] substitution)
    {
	if (substitution == null)
	    throw new IllegalArgumentException("Null substitution");
	if (averageBytesPerChar <= 0.0f)
	    throw new IllegalArgumentException("Non-positive "
					       + "averageBytesPerChar");
	if (maxBytesPerChar <= 0.0f)
	    throw new IllegalArgumentException("Non-positive "
					       + "maxBytesPerChar");
	subst = substitution;
	this.averageBytesPerChar = averageBytesPerChar;
	this.maxBytesPerChar = maxBytesPerChar;
    }

    /**
     * Initializes a new encoder.  The new encoder will have the given
     * bytes-per-char values and its substitution will be the
     * byte array <tt>{</tt>&nbsp;<tt>(byte)'?'</tt>&nbsp;<tt>}</tt>. </p>
     *
     * @param  averageBytesPerChar
     *         A positive float value indicating the expected number of
     *         bytes that will be produced for each input character
     *
     * @param  maxBytesPerChar
     *         A positive float value indicating the maximum number of
     *         bytes that will be produced for each input character
     */
    protected CharsetEncoder(float averageBytesPerChar,
			     float maxBytesPerChar)
    {
	this(averageBytesPerChar, maxBytesPerChar,
	     new byte[] { (byte)'?' });
    }

    /**
     * Returns this encoder's substitution, or <tt>null</tt> if the
     * substitution is not defined. </p>
     *
     * @return  The current substitution of this encoder;
     *          may be <tt>null</tt>
     */
    public final byte[] substitution() {
	return subst;
    }

    /**
     * Sets this encoder's substitution.  To disable substitution,
     * set the substitution to <tt>null</tt>. </p>
     *
     * @param  newSubstitution
     *         The new substitution, or <tt>null</tt> if
     *         substitution is not to be performed
     */
    public final void substitute(byte[] newSubstitution) {
	subst = newSubstitution;
    }

    /**
     * Returns the average number of bytes that will be produced for each
     * character of input.  This value may be used to estimate the size of the
     * output buffer required for a given input sequence. </p>
     *
     * @return  The average number of bytes produced
     *          per character of input
     */
    public final float averageBytesPerChar() {
	return averageBytesPerChar;
    }

    /**
     * Returns the maximum number of bytes that will be produced for each
     * character of input.  This value may be used to compute the worst-case size
     * of the output buffer required for a given input sequence. </p>
     *
     * @return  The maximum number of bytes produced per character of input
     */
    public final float maxBytesPerChar() {
	return maxBytesPerChar;
    }

    /**
     * Encodes as many characters as possible from the given input buffer,
     * writing the results to the given output buffer.
     *
     * <p> The buffers are read from, and written to, starting at their current
     * positions.  At most {@link Buffer#remaining in.remaining()} characters
     * will be read, and at most {@link Buffer#remaining out.remaining()}
     * bytes will be written.  The buffers' positions will be advanced to
     * reflect the characters read and the bytes written, but their marks and
     * limits will not be modified.
     *
     * <p> This method either completes successfully, having read at least one
     * character from the input buffer and written zero or more bytes to the
     * output buffer, or it throws one of the exceptions listed below.  If it
     * throws an exception then no characters will have been consumed, no
     * bytes will have been produced, and the buffers' positions will have
     * been restored to their initial values.
     *
     * <p> The <tt>endOfInput</tt> parameter advises this method as to whether
     * the invoker can provide further input beyond that contained in the given
     * input buffer.  If there is a possibility of providing additional input
     * then the invoker should pass <tt>false</tt> for this parameter; if there
     * is no possibility of providing further input then the invoker should
     * pass <tt>true</tt>.  It is not erroneous, and in fact it is quite
     * common, to pass <tt>false</tt> in one invocation and later discover that
     * no further input was actually available.  It is critical, however, that
     * the final invocation of this method in a sequence of invocations always
     * pass <tt>true</tt> so that a {@link MalformedInputException} rather than
     * a {@link BufferUnderflowException} can be thrown if there is
     * insufficient input to complete the <a href="#steps">encoding
     * operation</a>.
     *
     * <p> If the substitution is defined then it will be copied to the output
     * buffer each time an encoding error occurs; if the substitution is not
     * defined then an appropriate exception will be thrown.
     *
     * <p> This method works by repeatedly invoking the {@link #encodeStep
     * encodeStep} method.  If that method throws an exception before producing
     * any output then the loop is terminated, the buffers' positions are
     * reset, and the exception is rethrown.  If that method throws an
     * exception after some output has been produced then the loop is
     * terminated, the buffers are reset to their states prior to the most
     * recent step, and the method returns normally. </p>
     *
     * @param  in
     *         The input character buffer
     *
     * @param  out
     *         The output byte buffer
     *
     * @param  endOfInput
     *         <tt>true</tt> if, and only if, this is the final invocation of
     *         this method in the current encoding operation
     *
     * @return  <tt>false</tt> if, and only if, the invocation terminated
     *          after encoding one or more characters because the next
     *          encoding step would have thrown a character-coding exception
     *
     * @throws  BufferUnderflowException
     *          If more input characters are required.  Any characters remaining in
     *          the given input buffer should be preserved so that they are the
     *          first characters read by the next invocation of this method.
     *
     * @throws  BufferOverflowException
     *          If there is insufficient room in the output buffer
     *
     * @throws  IllegalStateException
     *          If the previous step of the current encoding operation was an
     *          invocation neither of the {@link #reset reset} method nor of
     *          this method with a value of <tt>false</tt> for the
     *          <tt>endOfInput</tt> parameter
     *
     * @throws  MalformedInputException
     *          If the character sequence starting at the input buffer's current
     *          position is not a legal sixteen-bit Unicode sequence and the substitution is not defined
     *
     * @throws  UnmappableCharacterException
     *          If the character sequence starting at the input buffer's current
     *          position cannot be mapped to an equivalent byte sequence and 
     *          the substitution is not defined
     */
    public final boolean encode(CharBuffer in, ByteBuffer out,
				boolean endOfInput)
	throws CharacterCodingException
    {
	if (!in.hasRemaining())
	    throw new BufferUnderflowException();
	int inLimit = in.limit();
	int outStart = out.position();
	int inMark = 0;
	int outMark = 0;
	try {
	    try {
		for (;;) {
		    inMark = in.position();
		    outMark = out.position();
		    if (inMark >= inLimit)
			// Avoid creating a new exception object
			throw bufferUnderflowException;
		    try {
			encodeStep(in, out);
		    } catch (CharacterCodingException x) {
			if (subst == null)
			    throw x;
			// Insert the substitution and skip over
			// the troublesome input
			out.put(subst);
			int n = ((x instanceof MalformedInputException)
				 ? ((MalformedInputException)x).getInputLength()
				 : 1);
			n = Math.min(in.remaining(), n);
			if (n == 0)
			    return true;
			in.position(in.position() + n);
		    }
		}
	    } catch (BufferUnderflowException x) {
		if (!handle(in, out, outStart, inMark, outMark)) {
		    if (!endOfInput)
			throw x;
		    if (subst != null)
			out.put(subst);
		    else
			throw new MalformedInputException(in.remaining());
		}
		return true;
	    } catch (CharacterCodingException x) {
		if (!handle(in, out, outStart, inMark, outMark))
		    throw x;
		return false;
	    }
	} catch (BufferOverflowException x) {
	    if (!handle(in, out, outStart, inMark, outMark))
		throw x;
	    return true;
	}
    }

    // Cached exception used above in order to avoid creating a new exception
    // object in every invocation of the encode method.  Some clients, e.g.,
    // java.io.PrintStream in auto-flushing mode, code many very small input
    // sequences.
    //
    private static BufferUnderflowException bufferUnderflowException
	= new BufferUnderflowException();

    // Utility method for handling exceptions
    //
    private static boolean handle(Buffer in, Buffer out,
				  int outStart, int inMark, int outMark)
    {
	in.position(inMark);
	out.position(outMark);
	return (out.position() > outStart);
    }

    /**
     * Flushes this encoder.
     *
     * <p> Some encoders maintain internal state and may need to write some
     * final bytes to the output buffer once the overall input sequence has
     * been read.
     *
     * <p> Any additional output is written to the output buffer beginning at
     * its current position.  At most {@link Buffer#remaining out.remaining()}
     * bytes will be written.  The buffer's position will be advanced
     * appropriately, but its mark and limit will not be modified.
     *
     * <p> This method either completes successfully, having written zero or
     * more bytes to the output buffer, or it throws a {@link
     * BufferOverflowException}.  If it throws an exception then no bytes
     * will have been written to the output buffer.  If this happens then this
     * method must be invoked again, with an output buffer that has more room,
     * in order to complete the current <a href="#steps">encoding
     * operation</a>.
     *
     * <p> The default implementation of this method does nothing.  This method
     * should be overridden by encoders that may need to write final bytes
     * to the output buffer once the input sequence has been read. </p>
     *
     * @param  out
     *         The output byte buffer
     *
     * @throws  BufferOverflowException
     *          If there is insufficient room in the output buffer for all of
     *          the bytes that must be written
     *
     * @throws  IllegalStateException
     *          If the previous step of the current encoding operation was an
     *          invocation neither of the {@link #reset reset} method nor of
     *          the three-argument {@link
     *          #encode(CharBuffer,ByteBuffer,boolean) encode} method
     *          with a value of <tt>true</tt> for the <tt>endOfInput</tt>
     *          parameter
     */
    public void flush(ByteBuffer out) { }

    /**
     * Resets this encoder, clearing any internal state.
     *
     * <p> The default implementation of this method does nothing.  This method
     * should be overridden by encoders that maintain internal state. </p>
     */
    public void reset() { }

    /**
     * Encodes one or more characters into one or more bytes.
     *
     * <p> This method performs a single encoding step; it is intended to be
     * invoked by the {@link #encode encode} method, which handles the
     * bookkeeping involved in the encoding process.  This method does not
     * attempt to encode the entire input buffer; rather, it encodes as few
     * characters as possible.
     *
     * <p> Upon invocation this method assumes that there is at least one
     * character remaining in the input buffer.  This method may advance the
     * position of each buffer, but it does not change their marks or limits.
     *
     * <p> An invocation of method either completes successfully, having
     * consumed one or more characters from the input buffer and written zero or
     * more bytes to the output buffer, or it throws one of the exceptions
     * listed below.
     *
     * <p> If an invocation of this method throws an exception because an
     * encoding error occurs then the invoking {@link #encode encode} method
     * will check whether the substitution is defined.  If so then it will
     * insert the substitution and continue the <a href="#steps">encoding
     * operation</a>.  Otherwise, and for all other types of exceptions, the
     * {@link #encode encode} method will reset the buffers' positions to the
     * values they had prior to the failing invocation and then propagate the
     * exception.
     *
     * <p> An implementation of this method may perform arbitrary lookahead
     * simply by reading forward in the input buffer.  If there is insufficient
     * input then a {@link BufferUnderflowException} will be thrown and the
     * {@link #encode encode} method will re-invoke this method with additional
     * input if any is available. </p>
     *
     * @param  in
     *         The input character buffer; it must have at least one remaining
     *         character, and its mark must be equal to its position
     *
     * @param  out
     *         The output byte buffer; its mark must be equal to its
     *         position
     *
     * @throws  BufferUnderflowException
     *          If more input characters are required.  Any characters remaining in
     *          the given input buffer should be preserved so that they are the
     *          first characters read by the next invocation of this method.
     *
     * @throws  BufferOverflowException
     *          If there is insufficient room in the output buffer
     *
     * @throws  MalformedInputException
     *          If the character sequence starting at the input buffer's current
     *          position is not a legal sixteen-bit Unicode sequence and the substitution is not defined
     *
     * @throws  UnmappableCharacterException
     *          If the character sequence starting at the input buffer's current
     *          position cannot be mapped to an equivalent byte sequence and 
     *          the substitution is not defined
     */
    protected abstract void encodeStep(CharBuffer in, ByteBuffer out)
	throws CharacterCodingException;

    /**
     * Convenience method that encodes the entire content of a single input
     * character buffer into a newly-allocated byte buffer.
     *
     * <p> This method implements an entire <a href="#steps">encoding
     * operation</a>; that is, it resets this encoder, then it encodes the
     * characters in the given character buffer, and finally it flushes this
     * encoder.  This method should therefore not be invoked if a encoding
     * operation is already in progress.  </p>
     *
     * @param  in
     *         The input character buffer
     *
     * @return A newly-allocated byte buffer containing the result of the
     *         encoding operation.  The buffer's position will be zero and its
     *         limit will follow the last byte written.
     *
     * @throws  IllegalStateException
     *          If an encoding operation is already in progress
     *
     * @throws  MalformedInputException
     *          If the character sequence starting at the input buffer's current
     *          position is not a legal sixteen-bit Unicode sequence and the substitution is not defined
     *
     * @throws  UnmappableCharacterException
     *          If the character sequence starting at the input buffer's current
     *          position cannot be mapped to an equivalent byte sequence and 
     *          the substitution is not defined
     */
    public final ByteBuffer encode(CharBuffer in)
	throws CharacterCodingException
    {
	int n = (int)(in.remaining() * averageBytesPerChar());
	ByteBuffer out = ByteBuffer.allocate(n);
	for (;;) {
	    try {
		if (in.hasRemaining())
		    encode(in, out, true);
		else {
		    flush(out);
		    break;
		}
	    } catch (BufferOverflowException x) {
		n *= 2;
		ByteBuffer o = ByteBuffer.allocate(n);
		out.flip();
		o.put(out);
		out = o;
	    }
	}
	out.flip();
	return out;
    }





























    private boolean canEncode(CharBuffer cb) {
	byte[] subst = substitution();
	try {
	    substitute(null);
	    encode(cb);
	} catch (CharacterCodingException x) {
	    return false;
	} finally {
	    substitute(subst);
	}
	return true;
    }

    /**
     * Tells whether or not this encoder can encode the given character.
     *
     * <p> This method returns <tt>false</tt> if the given character is a
     * surrogate character; such characters can be interpreted only when they
     * are members of a pair consisting of a high surrogate followed by a low
     * surrogate.  The {@link #canEncode(java.lang.CharSequence)
     * canEncode(CharSequence)} method may be used to test whether or not a
     * character sequence can be encoded.
     *
     * <p> This method may modify this encoder's state; it should therefore not
     * be invoked if an <a href="#steps">encoding operation</a> is already in
     * progress.
     *
     * <p> The default implementation of this method is not very efficient; it
     * should generally be overridden to improve performance.  </p>
     *
     * @return  <tt>true</tt> if, and only if, this encoder can encode
     *          the given character without throwing any exceptions and without
     *          performing any substitutions
     *
     * @throws  IllegalStateException
     *          If an encoding operation is already in progress
     */
    public boolean canEncode(char c) {
	CharBuffer cb = CharBuffer.allocate(1);
	cb.put(c);
	cb.flip();
	return canEncode(cb);
    }

    /**
     * Tells whether or not this encoder can encode the given character
     * sequence.
     *
     * <p> If this method returns <tt>false</tt> for a particular character
     * sequence then more information about why the sequence cannot be encoded
     * may be obtained by performing a full <a href="#steps">encoding
     * operation</a>.
     *
     * <p> This method may modify this encoder's state; it should therefore not
     * be invoked if an encoding operation is already in progress.
     *
     * <p> The default implementation of this method is not very efficient; it
     * should generally be overridden to improve performance.  </p>
     *
     * @return  <tt>true</tt> if, and only if, this encoder can encode
     *          the given character without throwing any exceptions and without
     *          performing any substitutions
     *
     * @throws  IllegalStateException
     *          If an encoding operation is already in progress
     */
    public boolean canEncode(CharSequence cs) {
	CharBuffer cb;
	if (cs instanceof CharBuffer)
	    cb = ((CharBuffer)cs).duplicate();
	else
	    cb = CharBuffer.wrap(cs.toString());
	return canEncode(cb);
    }



}
