/*
 * @(#)PolicyNode.java	1.4 01/03/14
 *
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 */

package java.security.cert;

import java.util.Iterator;
import java.util.Set;

/**
 * A valid policy tree node resulting from the PKIX certification path 
 * validation algorithm.
 *
 * <p>One of the outputs of the PKIX certification path validation 
 * algorithm is a valid policy tree, which includes the policies that
 * were determined to be valid, how this determination was reached,
 * and any policy qualifiers encountered. This tree is of depth
 * <i>n</i>, where <i>n</i> is the length of the certification
 * path that has been validated.
 *
 * <p>Most applications will not need to examine the valid policy tree.
 * They can achieve their policy processing goals by setting the 
 * policy-related parameters in <code>PKIXParameters</code>. However,
 * the valid policy tree is available for more sophisticated applications,
 * especially those that process policy qualifiers.
 *
 * <p>{@link PKIXCertPathValidatorResult#getPolicyTree() 
 * PKIXCertPathValidatorResult.getPolicyTree} returns the root node of the
 * valid policy tree. The tree can be traversed using the 
 * {@link #getChildren getChildren} and {@link #getParent getParent} methods. 
 * Data about a particular node can be retrieved using other methods of 
 * <code>PolicyNode</code>.
 *
 * <p><b>Concurrent Access</b>
 * <p>A <code>PolicyNode</code> object may (or may not) be immutable and 
 * thread-safe. The {@link #isImmutable isImmutable} method allows an 
 * application to determine if a <code>PolicyNode</code> is immutable and
 * thread-safe. If this method returns <code>true</code>, this 
 * <code>PolicyNode</code> object
 * and any objects that it returns through public methods are thread-safe and
 * immutable. Multiple threads may concurrently invoke the methods defined 
 * in this class on a single immutable <code>PolicyNode</code> object (or 
 * more than one) with no ill effects.
 *
 * <p>Once marked immutable, a <code>PolicyNode</code> must remain immutable
 * (and thread-safe) for the lifetime of the object. Typically, a 
 * <code>PolicyNode</code> is mutable while the valid policy tree is being
 * built. At the end of the validation process, all <code>PolicyNode</code>
 * objects in the valid policy tree are made immutable and a 
 * <code>PKIXCertPathValidatorResult</code> object is created containing a 
 * reference to the root of the tree.
 *
 * @version     1.4 03/14/01
 * @since       1.4
 * @author      Sean Mullan
 */
public interface PolicyNode {

    /**
     * Returns the parent of this node, or <code>null</code> if this is the 
     * root node.
     *
     * @return the parent of this node, or <code>null</code> if this is the 
     * root node
     */
    PolicyNode getParent();

    /**
     * Returns an iterator over the children of this node. Any attempts to 
     * modify the children of this node through the 
     * <code>Iterator</code>'s remove method must throw an
     * <code>UnsupportedOperationException</code>.
     *
     * @return an iterator over the children of this node
     */
    Iterator getChildren();

    /**
     * Returns the depth of this node in the valid policy tree.
     *
     * @return the depth of this node (0 for the root node, 1 for its
     * children, and so on)
     */
    int getDepth();

    /**
     * Returns the valid policy represented by this node.
     *
     * @return the <code>String</code> OID of the valid policy
     * represented by this node, or the special value "any-policy". For 
     * the root node, this method always returns the special value "any-policy".
     */
    String getValidPolicy();

    /**
     * Returns the set of policy qualifiers associated with the
     * valid policy represented by this node.
     *
     * @return an immutable <code>Set</code> of 
     * <code>PolicyQualifierInfo</code>s. For the root node, this
     * is always an empty <code>Set</code>.
     */
    Set getPolicyQualifiers();

    /**
     * Returns the set of expected policies that would satisfy this
     * node's valid policy in the next certificate to be processed.
     *
     * @return an immutable <code>Set</code> of expected policy 
     * <code>String</code> OIDs, or an immutable <code>Set</code> with
     * the single special value "any-policy". For the root node, this method 
     * always returns a <code>Set</code> with the single value "any-policy".
     */
    Set getExpectedPolicies();

    /**
     * Returns the criticality indicator of the certificate policy extension
     * in the most recently processed certificate.
     *
     * @return <code>true</code> if extension marked critical, 
     * <code>false</code> otherwise. For the root node, <code>false</code>
     * is always returned.
     */
    boolean isCritical(); 

    /**
     * Returns a <code>boolean</code> indicating whether this 
     * <code>PolicyNode</code> object is immutable and thread-safe. Once a 
     * <code>PolicyNode</code> is marked immutable, it must remain immutable
     * and thread-safe for the lifetime of the object.
     *
     * @return <code>true</code> if this <code>PolicyNode</code> object
     * is immutable and thread-safe, <code>false</code> otherwise
     */
    boolean isImmutable();
}
