/*
 * @(#)IdentityHashMap.java	1.5 01/04/17
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package java.util;

/**
 * This class implements the <tt>Map</tt> interface with a hash table, using
 * reference-equality in place of object-equality when comparing keys (and
 * values).  In other words, in an <tt>IdentityHashMap</tt>, two keys
 * <tt>k1</tt> and <tt>k2</tt> are considered equal if and only if
 * <tt>(k1==k2)</tt>.  (In normal <tt>Map</tt> implementations (like
 * <tt>HashMap</tt>) two keys <tt>k1</tt> and <tt>k2</tt> are considered equal
 * if and only if <tt>(k1==null ? k2==null : k1.equals(k2))</tt>.)
 *
 * <p><b>This class is <i>not</i> a general-purpose <tt>Map</tt>
 * implementation!  While this class implements the <tt>Map</tt> interface, it
 * intentionally violates <tt>Map's</tt> general contract, which mandates the
 * use of the <tt>equals</tt> method when comparing objects.  This class is
 * designed for use only in the rare cases wherein reference-equality
 * semantics are required.</b>
 *
 * <p>A typical use of this class is <i>topology-preserving object graph
 * transformations</i>, such as serialization or deep-copying.  To perform such
 * a transformation, a program must maintain a "node table" that keeps track
 * of all the object references that have already been processed.  The node
 * table must not equate distinct objects even if they happen to be equal.
 * Another typical use of this class is to maintain <i>proxy objects</i>.  For
 * example, a debugging facility might wish to maintain a proxy object for
 * each object in the program being debugged.
 *
 * <p>This class provides all of the optional map operations, and permits
 * <tt>null</tt> values and the <tt>null</tt> key.  This class makes no
 * guarantees as to the order of the map; in particular, it does not guarantee
 * that the order will remain constant over time.
 *
 * <p>This class provides constant-time performance for the basic
 * operations (<tt>get</tt> and <tt>put</tt>), assuming the system
 * identity hash function ({@link System#identityHashCode(Object)})
 * disperses elements properly among the buckets.
 *
 * <p>This class has one tuning parameter (which affects performance but not
 * semantics): <i>expected maximum size</i>.  This parameter is the maximum
 * number of key-value mappings that the map is expected to hold.  Internally,
 * this parameter is used to determine the number of buckets initially
 * comprising the hash table.  The precise relationship between the expected
 * maximum size and the number of buckets is unspecified.
 *
 * <p>If the size of the map (the number of key-value mappings) sufficiently
 * exceeds the expected maximum size, the number of buckets is increased
 * Increasing the number of buckets ("rehashing") may be fairly expensive, so
 * it pays to create identity hash maps with a sufficiently large expected
 * maximum size.  On the other hand, iteration over collection views requires
 * time proportional to the the number of buckets in the hash table, so it
 * pays not to set the expected maximum size too high if you are especially
 * concerned with iteration performance or memory usage.
 *
 * <p><b>Note that this implementation is not synchronized.</b> If multiple
 * threads access this map concurrently, and at least one of the threads
 * modifies the map structurally, it <i>must</i> be synchronized externally.
 * (A structural modification is any operation that adds or deletes one or
 * more mappings; merely changing the value associated with a key that an
 * instance already contains is not a structural modification.)  This is
 * typically accomplished by synchronizing on some object that naturally
 * encapsulates the map.  If no such object exists, the map should be
 * "wrapped" using the <tt>Collections.synchronizedMap</tt> method.  This is
 * best done at creation time, to prevent accidental unsynchronized access to
 * the map: <pre>
 *     Map m = Collections.synchronizedMap(new HashMap(...));
 * </pre>
 *
 * <p>The iterators returned by all of this class's "collection view methods"
 * are <i>fail-fast</i>: if the map is structurally modified at any time after
 * the iterator is created, in any way except through the iterator's own
 * <tt>remove</tt> or <tt>add</tt> methods, the iterator will throw a
 * <tt>ConcurrentModificationException</tt>.  Thus, in the face of concurrent
 * modification, the iterator fails quickly and cleanly, rather than risking
 * arbitrary, non-deterministic behavior at an undetermined time in the
 * future.
 *
 * <p>Implementation note: This is a simple <i>linear-probe</i> hash table,
 * as described for example in texts by Sedgewick and Knuth.  The array
 * alternates holding keys and values.  (This has better locality for large
 * tables than does using separate arrays.)  For many JRE implementations
 * and operation mixes, this class will yield better performance than
 * {@Link HashMap} (which uses <i>chaining</i> rather than linear-probing). 
 *
 * @see     System#identityHashCode(Object)
 * @see     Object#hashCode()
 * @see     Collection
 * @see	    Map
 * @see	    HashMap
 * @see	    TreeMap
 * @author  Doug Lea and Josh Bloch
 * @since   1.4
 */

public class IdentityHashMap extends AbstractMap implements Map,
                               java.io.Serializable, Cloneable
{
    /**
     * The initial capacity used by the no-args constructor.
     * MUST be a power of two.  The value 32 corresponds to the
     * (specified) expected maximum size of 21, given a load factor
     * of 2/3.
     */
    private static final int DEFAULT_CAPACITY = 32;

    /**
     * The minimum capacity, used if a lower value is implicitly specified
     * by either of the constructors with arguments.  The value 4 corresponds
     * to an expected maximum size of 2, given a load factor of 2/3.
     * MUST be a power of two.
     */
    private static final int MINIMUM_CAPACITY = 4;

    /**
     * The maximum capacity, used if a higher value is implicitly specified
     * by either of the constructors with arguments.
     * MUST be a power of two <= 1<<29.
     */
    private static final int MAXIMUM_CAPACITY = 1 << 29;

    /**
     * Special value used to mark slots as deleted.
     */
    private static final Object DELETED = new Object();

    /**
     * Special value used to mark slots as empty.
     */
    private static final Object EMPTY = new Object();

    /**
     * The table, resized as necessary. Length MUST always be a power of two.
     */
    private transient Object[] table;

    /**
     * The number of key-value mappings contained in this identity hash map.
     *
     * @serial
     */
    private int size;

    /**
     * The number of modifications, to support fast-fail iterators
     */
    private transient int modCount = 0;

    /**
     * The next size value at which to resize (capacity * load factor).
     */
    private transient int threshold;

    /**
     * Constructs a new, empty identity hash map with a default expected
     * maximum size (21).
     */
    public IdentityHashMap() {
        init(DEFAULT_CAPACITY);
    }

    /**
     * Constructs a new identity hash map containing the keys-value mappings
     * in the specified map.
     *
     * @param m the map whose mappings are to be placed into this map.
     */
    public IdentityHashMap(Map m) {
        // If m is is an IdentityHashMap do a fast copy (essentially a clone)
        if (m instanceof IdentityHashMap) {
            IdentityHashMap im = (IdentityHashMap) m;
            table = (Object[])(im.table.clone());
            size = im.size();
            threshold = im.threshold;
        } else {
            // Allow for 33% growth (i.e., capacity is >= 2* size()).
            init(capacity((m.size()*4)/3));
            putAll(m);
        }
    }

    /**
     * Constructs a new, empty map with the specified expected maximum size.
     * Putting more than the expected number of key-value mappings into
     * the map may cause the internal data structure to grow, which may be
     * somewhat time-consuming.
     *
     * @param expectedMaxSize the expected maximum size of the map.
     * @throws IllegalArgumentException if <tt>expectedMaxSize</tt> is negative
     */
    public IdentityHashMap(int expectedMaxSize) {
        if (expectedMaxSize < 0)
            throw new IllegalArgumentException("expectedMaxSize is negative");

        init(capacity(expectedMaxSize));
    }

    /**
     * Returns the appropriate capacity for the specified expected maximum
     * size.  Returns the smallest power of two between MINIMUM_CAPACITY
     * and MAXIMUM_CAPACITY, inclusive, that is greater than
     * (3 * expectedMaxSize)/2, if such a number exists.  Otherwise
     * returns MAXIMUM_CAPACITY.  If (3 * expectedMaxSize)/2 is negative, it
     * is assumed that overflow has occurred, and MAXIMUM_CAPACITY is returned.
     */
    private int capacity(int expectedMaxSize) {
        // Compute min capacity for expectedMaxSize given a load factor of 2/3
        int minCapacity = (3 * expectedMaxSize)/2;

        // Compute the appropriate capacity
        int result;
        if (minCapacity > MAXIMUM_CAPACITY || minCapacity < 0) {
            result = MAXIMUM_CAPACITY;
        } else {
            result = MINIMUM_CAPACITY;
            while (result < minCapacity)
                result <<= 1;
        }
        return result;
    }

    /**
     * Initialize object to be an empty map with the specified initial
     * capacity, which is assumed to be a power of two between
     * MINIMUM_CAPACITY and MAXIMUM_CAPACITY inclusive.
     */
    private void init(int initCapacity) {
        // assert (initCapacity & -initCapacity) == initCapacity; // power of 2
        // assert initCapacity >= MINIMUM_CAPACITY;
        // assert initCapacity <= MAXIMUM_CAPACITY;

        threshold = (initCapacity * 2)/3;
        table = new Object[2 * initCapacity];
        for (int i = 0; i < table.length; i += 2)
            table[i] = EMPTY;
    }

    /**
     * Returns the number of key-value mappings in this identity hash map.
     *
     * @return the number of key-value mappings in this map.
     */
    public int size() {
        return size;
    }

    /**
     * Returns <tt>true</tt> if this identity hash map contains no key-value
     * mappings.
     *
     * @return <tt>true</tt> if this identity hash map contains no key-value
     *         mappings.
     */
    public boolean isEmpty() {
        return size == 0;
    }

    /**
     * Return index for Object x.
     */
    private static int hash(Object x, int length) {
        int h = System.identityHashCode(x);
        // Multiply by 127 (quickly, via shifts), and mix in some high
        // bits to help guard against bunching of codes that are
        // consecutive or equally spaced.
        return ((h << 7) - h + (h >>> 9) + (h >>> 17)) & (length-2);
    }

    /**
     * Returns the value to which the specified key is mapped in this identity
     * hash map, or <tt>null</tt> if the map contains no mapping for
     * this key.  A return value of <tt>null</tt> does not <i>necessarily</i>
     * indicate that the map contains no mapping for the key; it is also
     * possible that the map explicitly maps the key to <tt>null</tt>. The
     * <tt>containsKey</tt> method may be used to distinguish these two
     * cases.
     *
     * @param   key the key whose associated value is to be returned.
     * @return  the value to which this map maps the specified key, or
     *          <tt>null</tt> if the map contains no mapping for this key.
     * @see #put(Object, Object)
     */
    public Object get(Object key) {
        int i = hash(key, table.length);
        while (true) {
            Object item = table[i];
            if (item == key) 
                return table[i+1];
            if (item == EMPTY)
                return null;
            if ((i+=2) >= table.length)
                i = 0;
        }
    }

    /**
     * Tests whether the specified object reference is a key in this identity
     * hash map.
     * 
     * @param   key   possible key.
     * @return  <code>true</code> if the specified object reference is a key
     *          in this map. 
     * @see     #containsValue(Object)
     */
    public boolean containsKey(Object key) {
        int i = hash(key, table.length);
        while (true) { 
            Object item = table[i];
            if (item == key) 
                return true;
            if (item == EMPTY)
                return false;
            if ((i+=2) >= table.length)
                i = 0;
        }
    }

    /**
     * Tests whether the specified object reference is a value in this identity
     * hash map.
     *
     * @param value value whose presence in this map is to be tested.
     * @return <tt>true</tt> if this map maps one or more keys to the
     *         specified object reference.
     * @see     #containsKey(Object)
     */
    public boolean containsValue(Object value) {
        for (int i = 1; i < table.length; i+= 2)
            if (table[i] == value)
                return true;

        return false;
    }
 
    /**
     * Tests if the specified key-value mapping is in the map.
     * 
     * @param   key   possible key.
     * @param   value possible value.
     * @return  <code>true</code> if and only if the specified key-value
     *          mapping is in map.
     */
    private boolean containsMapping(Object key, Object value) {
        int i = hash(key, table.length);
        while (true) { 
            Object item = table[i];
            if (item == key)
                return table[i+1] == value;
            if (item == EMPTY)
                return false;
            if ((i+=2) >= table.length)
                i = 0;
        }
    }

    /**
     * Associates the specified value with the specified key in this identity
     * hash map.  If the map previously contained a mapping for this key, the
     * old value is replaced.
     *
     * @param key the key with which the specified value is to be associated.
     * @param value the value to be associated with the specified key.
     * @return the previous value associated with <tt>key</tt>, or
     *	       <tt>null</tt> if there was no mapping for <tt>key</tt>.  (A
     *         <tt>null</tt> return can also indicate that the map previously
     *         associated <tt>null</tt> with the specified key.)
     * @see     Object#equals(Object)
     * @see     #get(Object)
     * @see     #containsKey(Object)
     */
    public Object put(Object key, Object value) {
        /*
         * insertionIndex is the index of the first DELETED
         * entry passed over while checking if x is already
         * present. If such a slot exists, we should use it
         * rather than trailing null slot
         */
        int insertionIndex = -1; 
        int i = hash(key, table.length);

        while (true) {
            Object item = table[i];

            if (item == EMPTY) {
                if (insertionIndex < 0)
                    insertionIndex = i;
                table[insertionIndex] = key;
                table[insertionIndex+1] = value;

                ++modCount;
                if (++size >= threshold)
                    resize();
                return null;
            }
            if (item == key) {
                Object oldValue = table[++i];
                table[i] = value;
                return oldValue;
            }
            if (item == DELETED && insertionIndex < 0)
                insertionIndex = i; 
            if ((i+=2) >= table.length)
                i = 0;
        }
    }

    /**
     *  Double the size of the table
     */
    private void resize() {
        int oldTableSize = table.length;
        if (oldTableSize == 2*MAXIMUM_CAPACITY) { // can't expand any further
            if (threshold == MAXIMUM_CAPACITY-1)
                throw new IllegalStateException("Capacity exhausted.");
            threshold = MAXIMUM_CAPACITY-1;  // Gigantic map!
            return;
        }

        int newSize = 2 * oldTableSize;
        threshold = (oldTableSize * 2)/3;

        Object[] oldTable = table;
        table = new Object[newSize];
        for (int i = 0; i < table.length; i +=2)
            table[i] = EMPTY;

        for (int j = 0; j < oldTable.length; j+=2) {
            Object key = oldTable[j];
            if (key != EMPTY && key != DELETED) {
                Object value = oldTable[j+1];

                int i = hash(key, table.length);
                while (table[i] != EMPTY)
                    if ((i+=2) == table.length)
                        i = 0;

                table[i] = key;
                table[i+1] = value;
            }
        }
    }

    /**
     * Removes the mapping for this key from this map if present.
     *
     * @param key key whose mapping is to be removed from the map.
     * @return previous value associated with specified key, or <tt>null</tt>
     *	       if there was no entry for key.  (A <tt>null</tt> return can
     *	       also indicate that the map previously associated <tt>null</tt>
     *	       with the specified key.)
     */
    public Object remove(Object key) {
        int i = hash(key, table.length);
        while (true) {
            Object item = table[i];
            if (item == key) {
                Object oldValue = table[i+1];
                markAsDeleted(i);
                return oldValue;
            }
            if (item == EMPTY)
                return null;
            if ((i+=2) >= table.length)
                i = 0;
        }
    }

    /**
     * Removes the specified key-value mapping from the map if it is present.
     * 
     * @param   key   possible key.
     * @param   value possible value.
     * @return  <code>true</code> if and only if the specified key-value
     *          mapping was in map.
     */
    private boolean removeMapping(Object key, Object value) {
        int i = hash(key, table.length);
        while (true) {
            Object item = table[i];
            if (item == key) {
                if (table[i+1] != value)
                    return false;
                markAsDeleted(i);
                return true;
            }
            if (item == EMPTY) 
                return false;
            if ((i+=2) >= table.length)
                i = 0;
        }
    }

    /**
     * Mark table[index] as either DELETED or, if possible, EMPTY.
     */
    private void markAsDeleted(int index) {
        /*
         * Because table[index] could have been between two real items
         * without an intervening EMPTY, it must normally be marked as
         * DELETED so that linear probing continues to work. But if it is
         * part of a sequence of DELETEDs ending in a EMPTY, table[index] and
         * other members of the sequence can be set to EMPTY, which will
         * shorten subsequent searches, especially after bursts of
         * removals. It also guarantees that an empty table has no DELETED
         * markers. In practice, this keeps the number of DELETED markers
         * low enough to not hurt search times much in the presence of
         * deletions.
         *
         * Note that the alternative of re-inserting old elements rather
         * than using a DELETED marker cannot be used here because this
         * could re-arrange items in the midst of an iteration.
         */
        ++modCount;   // record deletion
        --size;

        table[index+1] = null; // null out value to aid GC

        int j = index;
        while (true) { // Traverse starting at next slot after index
            if ((j+=2) == table.length)
                j = 0;

            Object item = table[j];
            if (item == EMPTY)           // Found a sequence ending in EMPTY
                break;

            if (item != DELETED) {       // no such luck
                table[index] = DELETED;
                return;
            }
        }

        while (true) {                   // Run backwards until not DELETED
            if ((j-=2) < 0)
                j = table.length - 2;

            if (j == index || table[j] == DELETED)
                table[j] = EMPTY;
            else
                return;
        }
    }

    /**
     * Removes all mappings from this map.
     */
    public void clear() {
        for (int i = 0; i < table.length; i++) {
            table[i++] = EMPTY;
            table[i]   = null;
        }
        size = 0;
        ++modCount;
    }

    // Views

    /**
     * This field is initialized to contain an instance of the entry set
     * view the first time this view is requested.  The view is stateless,
     * so there's no reason to create more than one.
     */

    private transient volatile Set entrySet = null;

    /**
     * Returns an identity-based set view of the keys contained in this map.
     * The set is backed by the map, so changes to the map are reflected in
     * the set, and vice-versa.  If the map is modified while an iteration
     * over the set is in progress, the results of the iteration are
     * undefined.  The set supports element removal, which removes the
     * corresponding mapping from the map, via the <tt>Iterator.remove</tt>,
     * <tt>Set.remove</tt>, <tt>removeAll</tt> <tt>retainAll</tt>, and
     * <tt>clear</tt> methods.  It does not support the <tt>add</tt> or
     * <tt>addAll</tt> methods.
     *
     * <p><b>While the object returned by this method implements the
     * <tt>Set</tt> interface, it does <i>not</i> obey <tt>Set's</tt> general
     * contract.  Like its backing map, the set returned by this method
     * defines element equality as reference-equality rather than
     * object-equality.  This affects the behavior of its <tt>contains</tt>,
     * <tt>remove</tt>, <tt>containsAll</tt>, <tt>equals</tt>, and
     * <tt>hashCode</tt> methods.</b>
     *
     * <p>The <tt>equals</tt> method of the returned set returns <tt>true</tt>
     * only if the specified object is a set containing exactly the same
     * object references as the returned set.  The symmetry and transitivity
     * requirements of the <tt>Object.equals</tt> contract may be violated if
     * the set returned by this method is compared to a normal set.  However,
     * the <tt>Object.equals</tt> contract is guaranteed to hold among sets
     * returned by this method.</b>
     *
     * <p>The <tt>hashCode</tt> method of the returned set returns the sum of
     * the <i>identity hashcodes</i> of the elements in the set, rather than
     * the sum of their hashcodes.  This is mandated by the change in the
     * semantics of the <tt>equals</tt> method, in order to enforce the
     * general contract of the <tt>Object.hashCode</tt> method among sets
     * returned by this method.
     *
     * @return an identity-based set view of the keys contained in this map.
     * @see Object#equals(Object)
     * @see System#identityHashCode(Object)
     */
    public Set keySet() {
        Set ks = keySet;
        if (ks != null)
            return ks;

        return keySet = new AbstractSet() {
            public Iterator iterator() {
                return new KeyIterator();
            }
            public int size() {
                return size;
            }
            public boolean contains(Object o) {
                return containsKey(o);
            }
            public boolean remove(Object o) {
                int oldSize = size;
                IdentityHashMap.this.remove(o);
		return size != oldSize;
            }
            /*
             * Must revert from AbstractSet's impl to AbstractCollection's, as
             * the former contains an optimization that results in incorrect
             * behavior when c is a smaller "normal" (non-identity-based) Set.
             */
            public boolean removeAll(Collection c) {
                boolean modified = false;
                for (Iterator i = iterator(); i.hasNext(); ) {
                    if(c.contains(i.next())) {
                        i.remove();
                        modified = true;
                    }
                }
                return modified;
            }
            public void clear() {
                IdentityHashMap.this.clear();
            }
            public int hashCode() {
                int result = 0;
                for (Iterator i = iterator(); i.hasNext(); )
                    result += System.identityHashCode(i.next());
                return result;
            }
        };
    }

    /**
     * <p>Returns a collection view of the values contained in this map.  The
     * collection is backed by the map, so changes to the map are reflected in
     * the collection, and vice-versa.  If the map is modified while an
     * iteration over the collection is in progress, the results of the
     * iteration are undefined.  The collection supports element removal,
     * which removes the corresponding mapping from the map, via the
     * <tt>Iterator.remove</tt>, <tt>Collection.remove</tt>,
     * <tt>removeAll</tt>, <tt>retainAll</tt> and <tt>clear</tt> methods.
     * It does not support the <tt>add</tt> or <tt>addAll</tt> methods.
     *
     * <p><b>While the object returned by this method implements the
     * <tt>Collection</tt> interface, it does <i>not</i> obey
     * <tt>Collection's</tt> general contract.  Like its backing map,
     * the collection returned by this method defines element equality as
     * reference-equality rather than object-equality.  This affects the
     * behavior of its <tt>contains</tt>, <tt>remove</tt> and
     * <tt>containsAll</tt> methods.</b>
     *
     * @return a collection view of the values contained in this map.
     */
    public Collection values() {
        Collection vs = values;
        if (vs != null)
            return vs;
        return values = new AbstractCollection() {
            public Iterator iterator() {
                return new ValueIterator();
            }
            public int size() {
                return size;
            }
            public boolean contains(Object o) {
                return containsValue(o);
            }
            public boolean remove(Object o) {
                for(Iterator i = iterator(); i.hasNext(); ) {
                    if (i.next() == o) {
                        i.remove();
                        return true;
                    }
                }
                return false;
            }
            public void clear() {
                IdentityHashMap.this.clear();
            }
        };
    }

    /**
     * Returns a set view of the mappings contained in this map.  Each element
     * in the returned set is a reference-equality-based <tt>Map.Entry</tt>.
     * The set is backed by the map, so changes to the map are reflected in
     * the set, and vice-versa.  If the map is modified while an iteration
     * over the set is in progress, the results of the iteration are
     * undefined.  The set supports element removal, which removes the
     * corresponding mapping from the map, via the <tt>Iterator.remove</tt>,
     * <tt>Set.remove</tt>, <tt>removeAll</tt>, <tt>retainAll</tt> and
     * <tt>clear</tt> methods.  It does not support the <tt>add</tt> or
     * <tt>addAll</tt> methods.
     *
     * <p>Like the backing map, the <tt>Map.Entry</tt> objects in the set
     * returned by this method define key and value equality as
     * reference-equality rather than object-equality.  This affects the
     * behavior of the <tt>equals</tt> and <tt>hashCode</tt> methods of these
     * <tt>Map.Entry</tt> objects.  A reference-equality based <tt>Map.Entry
     * e</tt> is equal to an object <tt>o</tt> if and only if <tt>o</tt> is a
     * <tt>Map.Entry</tt> and <tt>e.getKey()==o.getKey() &&
     * e.getValue()==o.getValue()</tt>.  To accommodate these equals
     * semantics, the <tt>hashCode</tt> method returns
     * <tt>System.identityHashCode(e.getKey()) ^
     * System.identityHashCode(e.getValue())</tt>.
     *
     * <p><b>Owing to the reference-equality-based semantics of the
     * <tt>Map.Entry</tt> instances in the set returned by this method,
     * it is possible that the symmetry and transitivity requirements of
     * the {@link Object#equals(Object)} contract may be violated if any of
     * the entries in the set is compared to a normal map entry, or if
     * the set returned by this method is compared to a set of normal map
     * entries (such as would be returned by a call to this method on a normal
     * map).  However, the <tt>Object.equals</tt> contract is guaranteed to
     * hold among identity-based map entries, and among sets of such entries.
     * </b>
     *
     * @return a set view of the identity-mappings contained in this map.
     */
    public Set entrySet() {
        Set es = entrySet;
        if (es != null) 
            return es;
        return entrySet = new AbstractSet() {
            public Iterator iterator() {
                return new EntryIterator();
            }
            public boolean contains(Object o) {
                if (!(o instanceof Map.Entry))
                    return false;
                Map.Entry entry = (Map.Entry)o;
                return containsMapping(entry.getKey(), entry.getValue());
            }
            public boolean remove(Object o) {
                if (!(o instanceof Map.Entry))
                    return false;
                Map.Entry entry = (Map.Entry)o;
                return removeMapping(entry.getKey(), entry.getValue());
            }
            public int size() {
                return size;
            }
            public void clear() {
                IdentityHashMap.this.clear();
            }
            /*
             * Must revert from AbstractSet's impl to AbstractCollection's, as
             * the former contains an optimization that results in incorrect
             * behavior when c is a smaller "normal" (non-identity-based) Set.
             */
            public boolean removeAll(Collection c) {
                boolean modified = false;
                for (Iterator i = iterator(); i.hasNext(); ) {
                    if(c.contains(i.next())) {
                        i.remove();
                        modified = true;
                    }
                }
                return modified;
            }
        };
    }

    private abstract class IdentityHashMapIterator implements Iterator {
        int index = 0;                   // current slot. 
        int expectedModCount = modCount; // to support fast-fail
        int lastReturnedIndex = -1;      // to allow remove()

        public boolean hasNext() {
            Object[] tab = table;
            int i;
            for (i = index; i < tab.length; i+=2) {
                Object key = tab[i];
                if (key != EMPTY && key != DELETED) {
                    index = i;
                    return true;
                }
            }
            index = i;
            return false;
        }

        public void remove() {
            if (lastReturnedIndex == -1)
                throw new IllegalStateException();
            if (modCount != expectedModCount)
                throw new ConcurrentModificationException();

            markAsDeleted(lastReturnedIndex);
            lastReturnedIndex = -1;
            expectedModCount = modCount;
        }
    }
  
    private class KeyIterator extends IdentityHashMapIterator {
        public Object next() {
            if (modCount != expectedModCount)
                throw new ConcurrentModificationException();
      
            Object[] tab = table;
            for (int i = index; i < tab.length; i+=2) {
                Object key = tab[i];
                if (key != EMPTY && key != DELETED) {
                    lastReturnedIndex = i;
                    index = i + 2;
                    return key;
                }
            }
            throw new NoSuchElementException();    
        }
    }

    private class ValueIterator extends IdentityHashMapIterator {
        public Object next() {
            if (modCount != expectedModCount)
                throw new ConcurrentModificationException();
      
            Object[] tab = table;
            for (int i = index; i < tab.length; i+=2) {
                Object key = tab[i];
                if (key != EMPTY && key != DELETED) {
                    lastReturnedIndex = i;
                    index = i + 2;
                    return tab[i+1];
                }
            }
            throw new NoSuchElementException();    
        }
    }

    private class EntryIterator extends IdentityHashMapIterator {
        private Map.Entry e = new IdentityHashMapEntry();

        public Object next() {
            if (modCount != expectedModCount)
                throw new ConcurrentModificationException();
      
            Object[] tab = table;
            for (int i = index; i < tab.length; i+=2) {
                Object key = tab[i];
                if (key != EMPTY && key != DELETED) {
                    lastReturnedIndex = i;
                    index = i + 2;
                    return e;
                }
            }
            throw new NoSuchElementException();    
        }


        private class IdentityHashMapEntry implements Map.Entry {
            public Object getKey() {
                return table[lastReturnedIndex];
            }
    
            public Object getValue() {
                return table[lastReturnedIndex + 1];
            }

            public Object setValue(Object value) {
                Object oldValue = table[lastReturnedIndex+1];
                table[lastReturnedIndex+1] = value;
                return oldValue;
            }

            public boolean equals(Object o) {
                if (!(o instanceof Map.Entry))
                    return false;
                Map.Entry e = (Map.Entry)o;
                return e.getKey()   == table[lastReturnedIndex] &&
                       e.getValue() == table[lastReturnedIndex+1];
            }

            public int hashCode() {
                return System.identityHashCode(table[index]) ^
                       System.identityHashCode(table[index+1]);
            }

            public String toString() {
                return table[index] + "=" + table[index+1];
            }
        }
    }

    /**
     * Compares the specified object with this map for equality.  Returns
     * <tt>true</tt> if the given object is also a map and the two maps
     * represent identical object-reference mappings.  More formally, this
     * map is equal to another map <tt>m</tt> if and only if
     * map <tt>this.entrySet().equals(m.entrySet())</tt>.
     *
     * <p><b>Owing to the reference-equality-based semantics of this map it is
     * possible that the symmetry and transitivity requirements of the
     * <tt>Object.equals</tt> contract may be violated if this map is compared
     * to a normal map.  However, the <tt>Object.equals</tt> contract is
     * guaranteed to hold among <tt>IdentityHashMap</tt> instances.</b>
     *
     * @param  o object to be compared for equality with this map.
     * @return <tt>true</tt> if the specified object is equal to this map.
     * @see Object#equals(Object)
     */
    public boolean equals(Object o) {
        if (o == this) {
            return true;
        } else if (o instanceof IdentityHashMap) {
            IdentityHashMap m = (IdentityHashMap) o;
            if (m.size() != size)
                return false;

            Object[] tab = m.table;
            for (int i = 0; i < tab.length; i+=2) {
                Object k = tab[i];
                if (k!=EMPTY && k!=DELETED && !containsMapping(k, tab[i+1]))
                    return false;
            }
            return true;
        } else if (o instanceof Map) {
            Map m = (Map)o;
            return entrySet().equals(m.entrySet());
        } else {
            return false;  // o is not a Map
        }
    }
  
    /**
     * Returns the hash code value for this map.  The hash code of a map
     * is defined to be the sum of the hashcode of each entry in the map's
     * entrySet view.  This ensures that <tt>t1.equals(t2)</tt> implies
     * that <tt>t1.hashCode()==t2.hashCode()</tt> for any two 
     * <tt>IdentityHashMap</tt> instances <tt>t1</tt> and <tt>t2</tt>, as
     * required by the general contract of {@link Object#hashCode()}.
     *
     * <p><b>Owing to the reference-equality-based semantics of the
     * <tt>Map.Entry</tt> instances in the set returned by this map's
     * <tt>entrySet</tt> method, it is possible that the contractual
     * requirement of <tt>Object.hashCode</tt> mentioned in the previous
     * paragraph will be violated if one of the two objects being compared is
     * an <tt>IdentityHashMap</tt> instance and the other is a normal map.</b>
     *
     * @return the hash code value for this map.
     * @see Object#hashCode()
     * @see Object#equals(Object)
     * @see #equals(Object)
     */
    public int hashCode() {
        int h = 0;
        for (int i = 0; i < table.length; i +=2) {
            Object key = table[i];
            if (key != EMPTY && key != DELETED) {
                h += System.identityHashCode(key) ^ 
                     System.identityHashCode(table[i+1]);
            }
        }
        return h;
    }

    /**
     * Returns a shallow copy of this identity hash map: the keys and values
     * themselves are not cloned.
     *
     * @return a shallow copy of this map.
     */
    public Object clone() {
        try { 
            IdentityHashMap t = (IdentityHashMap)super.clone();
            t.entrySet = null;
            t.table = (Object[])(table.clone());
            return t;
        } catch (CloneNotSupportedException e) { 
            throw new InternalError();
        }
    }

    /**
     * Save the state of the <tt>IdentityHashMap</tt> instance to a stream
     * (i.e., serialize it).
     *
     * @serialData The <i>size</i> of the HashMap (the number of key-value
     *	        mappings) (<tt>int</tt>), followed by the key (Object) and
     *          value (Object) for each key-value mapping represented by the
     *          IdentityHashMap.  The key-value mappings are emitted in no
     *          particular order.
     */
    private void writeObject(java.io.ObjectOutputStream s)
        throws java.io.IOException  {
        // Write out and any hidden stuff
        s.defaultWriteObject();

        // Write out size (number of Mappings)
        s.writeInt(size);

        // Write out keys and values (alternating)
        for (int i = 0; i < table.length; i += 2) {
            Object key = table[i];
            if (key != EMPTY && key != DELETED) {
                s.writeObject(key);
                s.writeObject(table[i+1]);
            }
        }
    }

    /**
     * Reconstitute the <tt>IdentityHashMap</tt> instance from a stream (i.e.,
     * deserialize it).
     */
    private void readObject(java.io.ObjectInputStream s)
        throws java.io.IOException, ClassNotFoundException  {
        // Read in any hidden stuff
        s.defaultReadObject();
    
        // Read in size (number of Mappings)
        int size = s.readInt();

        // Allow for 33% growth (i.e., capacity is >= 2* size()).
        init(capacity((size*4)/3));

        // Read the keys and values, and put the mappings in the table
        for (int i=0; i<size; i++) {
            Object key = s.readObject();
            Object value = s.readObject();
            put(key, value);
        }
    }
}
