/*
 * 1.1 00/10/17
 * 
 * Copyright 1996-2000 Sun Microsystems, Inc. 901 San Antonio Road,
 * Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
 */

package java.util;
import java.io.Serializable;
import java.io.IOException;

/**
 * <p>Hash table and linked list implementation of the <tt>Map</tt> interface,
 * with predictable iteration order.  This implementation differs from
 * <tt>HashMap</tt> in that it maintains a doubly-linked list running through
 * all of its entries.  This linked list defines the iteration ordering,
 * which is normally the order in which keys were inserted into the map
 * (<i>insertion-order</i>).  Note that insertion order is not affected
 * if a key is <i>re-inserted</i> into the map.  (A key <tt>k</tt> is
 * reinserted into a map <tt>m</tt> if <tt>m.put(k, v)</tt> is invoked when
 * <tt>m.containsKey(k)</tt> would return <tt>true</tt> immediately prior to
 * the invocation.)
 *
 * <p>This implementation spares its clients from the unspecified, generally
 * chaotic ordering provided by {@link HashMap} (and {@link Hashtable}),
 * without incurring the increased cost associated with {@link TreeMap}.  It
 * can be used to produce a copy of a map that has the same order as the
 * original, regardless of the original map's implementation:
 * <pre>
 *     void foo(Map m) {
 *         Map copy = new LinkedHashMap(m);
 *         ...
 *     }
 * </pre>
 * This technique is particularly useful if a module takes a map on input,
 * copies it, and later returns results whose order is determined by that of
 * the copy.  (Clients generally appreciate having things returned in the same
 * order they were presented.)
 *
 * <p>A special {@link #LinkedHashMap(int,float,boolean) constructor} is
 * provided to create a linked hash map whose order of iteration is the order
 * in which its entries were last accessed, from least-recently accessed to
 * most-recently (<i>access-order</i>).  This kind of map is well-suited to
 * building LRU caches.  Invoking the <tt>put</tt> or <tt>get</tt> method
 * results in an access to the corresponding entry (assuming it exists after
 * the invocation completes).  The <tt>putAll</tt> method generates one entry
 * access for each mapping in the specified map, in the order that key-value
 * mappings are provided by the specified map's entry set iterator.  <i>No
 * other methods generate entry accesses.</i> In particular, operations on
 * collection-views do <i>not</i> affect the order of iteration of the backing
 * map.
 *
 * <p>This class provides all of the optional <tt>Map</tt> operations, and
 * permits null elements.  Like <tt>HashMap</tt>, it provides constant-time
 * performance for the basic operations (<tt>add</tt>, <tt>contains</tt> and
 * <tt>remove</tt>), assuming the the hash function disperses elements
 * properly among the buckets.  Performance is likely to be just slightly
 * below that of <tt>HashMap</tt>, due to the added expense of maintaining the
 * linked list, with one exception: Iteration over the collection-views
 * of a <tt>LinkedHashMap</tt> requires time proportional to the <i>size</i>
 * of the map, regardless of its capacity.  Iteration over a <tt>HashMap</tt>
 * is likely to be more expensive, requiring time proportional to its
 * <i>capacity</i>.
 *
 * <p>A linked hash map has two parameters that affect its performance:
 * <i>initial capacity</i> and <i>load factor</i>.  They are defined precisely
 * as for <tt>HashMap</tt>.  Note, however, that the penalty for choosing an
 * excessively high value for initial capacity is less severe for this class
 * than for <tt>HashMap</tt>, as iteration times for this class are unaffected
 * by capacity.
 * 
 * <p><strong>Note that this implementation is not synchronized.</strong> If
 * multiple threads access a linked hash map concurrently, and at least
 * one of the threads modifies the map structurally, it <em>must</em> be
 * synchronized externally.  This is typically accomplished by synchronizing
 * on some object that naturally encapsulates the map.  If no such object
 * exists, the map should be "wrapped" using the
 * <tt>Collections.synchronizedMap</tt>method.  This is best done at creation
 * time, to prevent accidental unsynchronized access:
 * <pre>
 *    Map m = Collections.synchronizedMap(new LinkedHashMap(...));
 * </pre>
 * A structural modification is any operation that adds or deletes one or more
 * mappings or, in the case of access-ordered linked hash maps, affects
 * iteration order.  In insertion-ordered linked hash maps, merely changing
 * the value associated with a key that is already contained in the map is not
 * a structural modification.  <strong>In access-ordered linked hash maps,
 * merely querying the map with <tt>get</tt> is a structural
 * modification.</strong>)
 *
 * <p>The iterators returned by the <tt>iterator</tt> methods of the
 * collections returned by all of this class's collection view methods are
 * <em>fail-fast</em>: if the map is structurally modified at any time after
 * the iterator is created, in any way except through the iterator's own
 * remove method, the iterator will throw a
 * <tt>ConcurrentModificationException</tt>.  Thus, in the face of concurrent
 * modification, the Iterator fails quickly and cleanly, rather than risking
 * arbitrary, non-deterministic behavior at an undetermined time in the
 * future.
 *
 * @author  Josh Bloch
 * @version 1.1 10/17/00
 * @see     Object#hashCode()
 * @see     Collection
 * @see     Map
 * @see     HashMap
 * @see     TreeMap
 * @see     Hashtable
 * @since   JDK1.4
 */

public class LinkedHashMap extends HashMap {
    /**
     * The head of the double linked list.
     */
    private transient Entry header;
                                             
    /**
     * The iteration ordering method for this linked hash map: <tt>true</tt>
     * for access-order, <tt>false</tt> for insertion-order.
     *
     * @serial
     */
    private final boolean accessOrder;

    /**
     * Constructs a new, empty insertion-ordered <tt>LinkedHashMap</tt>
     * instance with the specified initial capacity and load factor.
     *
     * @param      initialCapacity   the initial capacity of the LinkedHashMap.
     * @param      loadFactor        a number between 0.0 and 1.0.
     * @exception  IllegalArgumentException  if the initial capacity is less
     *               than or equal to zero, or if the load factor is less than
     *               or equal to zero.
     */
    public LinkedHashMap(int initialCapacity, float loadFactor) {
        this(initialCapacity, loadFactor, false);
    }

    /**
     * Constructs a new, empty insertion-ordered <tt>LinkedHashMap</tt>
     * instance with the specified initial capacity and a default load factor
     * (0.75). 
     *
     * @param   initialCapacity   the initial capacity of the LinkedHashMap.
     */
    public LinkedHashMap(int initialCapacity) {
	this(initialCapacity, 0.75f, false);
    }

    /**
     * Constructs a new, empty insertion-ordered <tt>LinkedHashMap</tt>
     * instance with a default capacity and load factor (0.75).
     */
    public LinkedHashMap() {
	this(11, 0.75f, false);
    }

    /**
     * Constructs a new insertion-ordered <tt>LinkedHashMap</tt> instance with
     * the same mappings as the specified map.  The <tt>LinkedHashMap</tt>
     * instance is created with a capacity of thrice the number of mappings in
     * the specified map or 11 (whichever is greater), and a default load
     * factor (0.75).
     * 
     */
    public LinkedHashMap(Map t) {
	this(Math.max(3*t.size(), 11), 0.75f, false);
	putAll(t);
    }

    /**
     * Constructs a new, empty linked hash map with the specified initial 
     * capacity, load factor and ordering mode.
     *
     * @param      initialCapacity   the initial capacity of the HashMap.
     * @param      loadFactor        the load factor of the HashMap
     * @param      accessOrder       the ordering mode - <tt>true</tt> for
     *               access-order, <tt>false</tt> for insertion-order.
     * @throws     IllegalArgumentException  if the initial capacity is less
     *               than zero, or if the load factor is nonpositive.
     */
    public LinkedHashMap(int initialCapacity, float loadFactor,
                         boolean accessOrder) {
        super(initialCapacity, loadFactor);
        this.accessOrder = accessOrder;
        init();
    }

    void init() {
        header = new Entry(-1, null, null, null);
        header.before = header.after = header;
    }

    /**
     * Returns <tt>true</tt> if this map maps one or more keys to the
     * specified value.
     *
     * @param value value whose presence in this map is to be tested.
     * @return <tt>true</tt> if this map maps one or more keys to the
     *         specified value.
     */
    public boolean containsValue(Object value) {
        // Overridden to take advantage of faster iterator
        if (value==null) {
            for (Entry e = header.after; e != header; e = e.after)
                if (e.value==null)
                    return true;
        } else {
            for (Entry e = header.after; e != header; e = e.after)
                if (value.equals(e.value))
                    return true;
        }
        return false;
    }

    /**
     * Returns the value to which this map maps the specified key.  Returns
     * <tt>null</tt> if the map contains no mapping for this key.  A return
     * value of <tt>null</tt> does not <i>necessarily</i> indicate that the
     * map contains no mapping for the key; it's also possible that the map
     * explicitly maps the key to <tt>null</tt>.  The <tt>containsKey</tt>
     * operation may be used to distinguish these two cases.
     *
     * @return the value to which this map maps the specified key.
     * @param key key whose associated value is to be returned.
     */
    public Object get(Object key) {
        Entry e = (Entry)getEntry(key);
        if (e == null)
            return null;
        e.recordAccess(this);
        return e.value;
    }

    /**
     * Removes the mapping for this key from this map if present.
     *
     * @param key key whose mapping is to be removed from the map.
     * @return previous value associated with specified key, or <tt>null</tt>
     *	       if there was no mapping for key.  A <tt>null</tt> return can
     *	       also indicate that the map previously associated <tt>null</tt>
     *	       with the specified key.
     */
    public Object remove(Object key) {
	Entry e = (Entry) removeEntryForKey(key);
        if (e == null)
            return null;
        e.remove();
        return e.value;
    }

    /**
     * Removes all mappings from this map.
     */
    public void clear() {
        super.clear();
        header.before = header.after = header;
    }


    /**
     * LinkedHashMap entry.
     */
    private static class Entry extends HashMap.Entry {
        // These fields comprise the doubly linked list used for iteration.
        Entry before, after;

	Entry(int hash, Object key, Object value, HashMap.Entry next) {
            super(hash, key, value, next);
        }

        /**
         * Remove this entry from the linked list.
         */
        private void remove() {
            before.after = after;
            after.before = before;
        }

        /**                                             
         * Insert this entry before the specified existing entry in the list.
         */
        private void addBefore(Entry existingEntry) {
            after  = existingEntry;
            before = existingEntry.before;
            before.after = this;
            after.before = this;
        }

        /**
         * This method is invoked by the superclass whenever the value
         * of a pre-existing entry is read by Map.get or modified by Map.set.
         * If the enclosing Map is access-ordered, it moves the entry
         * to the end of the list; otherwise, it does nothing. 
         */
        void recordAccess(HashMap m) {
            LinkedHashMap lm = (LinkedHashMap)m;
            if (lm.accessOrder) {
                lm.modCount++;
                remove();
                addBefore(lm.header);
            }
        }
    }

    private abstract class LinkedHashIterator implements Iterator {
	Entry nextEntry    = header.after;
	Entry lastReturned = null;

	/**
	 * The modCount value that the iterator believes that the backing
	 * List should have.  If this expectation is violated, the iterator
	 * has detected concurrent modification.
	 */
	int expectedModCount = modCount;

	public boolean hasNext() {
            return nextEntry != header;
	}

	public void remove() {
	    if (lastReturned == null)
		throw new IllegalStateException();
	    if (modCount != expectedModCount)
		throw new ConcurrentModificationException();
            removeEntry(lastReturned);
            lastReturned.remove();
            expectedModCount++;
            lastReturned = null;
	}

	Entry nextEntry() {
	    if (modCount != expectedModCount)
		throw new ConcurrentModificationException();
            if (nextEntry == header)
                throw new NoSuchElementException();

            Entry e = lastReturned = nextEntry;
            nextEntry = e.after;
            return e;
	}
    }

    private class KeyIterator extends LinkedHashIterator {
	public Object next() { return nextEntry().key; }
    }

    private class ValueIterator extends LinkedHashIterator {
	public Object next() { return nextEntry().value; }
    }

    private class EntryIterator extends LinkedHashIterator {
	public Object next() { return nextEntry(); }
    }

    /**
     * Rehashes the contents of the LinkedHashMap into a LinkedHashMap with a 
     * larger capacity.  This version is faster than HashMap's because
     * of our faster iterator.
     */
    void rehash() {
	HashMap.Entry oldTable[] = table;
	int oldCapacity = oldTable.length;
	int newCapacity = oldCapacity * 2 + 1;
	Entry newTable[] = new Entry[newCapacity];

	modCount++;
	threshold = (int)(newCapacity * loadFactor);
	table = newTable;

        for (Entry e = header.after; e != header; e = e.after) {
            int index = (e.hash & 0x7FFFFFFF) % newCapacity;
            e.next = newTable[index];
            newTable[index] = e;
        }
    }

    /**
     * Returns a shallow copy of this <tt>LinkedHashMap</tt> instance: the
     * keys and values themselves are not cloned.
     *
     * @return a shallow copy of this map.
     */
    public Object clone() {
        LinkedHashMap result = (LinkedHashMap)super.clone();

        // Populate clone
        result.init();
        result.size = 0;
        for (Entry e = header.after; e != header; e = e.after)
            result.put(e.key, e.value);
        return result;
    }

    // These Overrides alter the behavior of superclass view iterator() methods
    Iterator newKeyIterator()   { return new KeyIterator();   }
    Iterator newValueIterator() { return new ValueIterator(); }
    Iterator newEntryIterator() { return new EntryIterator(); }

    /*
     * This override alters behavior of superclass put method. It causes newly
     * allocated entries to get inserted at the end of the linked list.
     */
    HashMap.Entry newEntry(int hash, Object k, Object v, HashMap.Entry next) {
        Entry e = new Entry(hash, k, v, next);
        e.addBefore(header);
        return e;
    }

    // This override causes entrySet().remove(Object) to update linked list.
    HashMap.Entry removeMapping(Object key) {
        Entry e = (Entry)super.removeMapping(key);
        if (e != null)
            e.remove();
        return e;
    }
}
