/*
 * @(#)ListResourceBundle.java	1.19 00/03/09
 *
 * Copyright 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

/*
 * (C) Copyright Taligent, Inc. 1996, 1997 - All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998 - All Rights Reserved
 *
 * The original version of this source code and documentation
 * is copyrighted and owned by Taligent, Inc., a wholly-owned
 * subsidiary of IBM. These materials are provided under terms
 * of a License Agreement between Taligent and Sun. This technology
 * is protected by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package java.util;
import java.util.Hashtable;

/**
 * <code>ListResourceBundle</code> is an abstract subclass of
 * <code>ResourceBundle</code> that manages resources for a locale
 * in a convenient and easy to use list. See <code>ResourceBundle</code> for
 * more information about resource bundles in general.
 *
 * <P>
 * Subclasses must override <code>getContents</code> and provide an array,
 * where each item in the array is a pair of objects.
 * The first element of each pair is a <code>String</code> key, and the second
 * is the value associated with that key.  [Right now, there's no error-checking
 * code to enforce this, so you could specify key-value pairs that have
 * something other than a String as a key.  But since the interfaces are defined
 * in terms of String, any value with a key that isn't a String will be
 * inaccessible.]
 *
 * <p>
 * The following <a name="sample">example</a> shows two members of a resource
 * bundle family with the base name "MyResources".
 * "MyResources" is the default member of the bundle family, and
 * "MyResources_fr" is the French member.
 * These members are based on <code>ListResourceBundle</code>
 * (a related <a href="PropertyResourceBundle.html#sample">example</a> shows
 * how you can add a bundle to this family that's based on a properties file).
 * The keys in this example are of the form "s1" etc. The actual
 * keys are entirely up to your choice, so long as they are the same as
 * the keys you use in your program to retrieve the objects from the bundle.
 * Keys are case-sensitive.
 * <blockquote>
 * <pre>
 * 
 * public class MyResources extends ListResourceBundle {
 *     public Object[][] getContents() {
 *         return contents;
 *     }
 *     static final Object[][] contents = {
 *     // LOCALIZE THIS
 *         {"s1", "The disk \"{1}\" contains {0}."},  // MessageFormat pattern
 *         {"s2", "1"},                               // location of {0} in pattern
 *         {"s3", "My Disk"},                         // sample disk name
 *         {"s4", "no files"},                        // first ChoiceFormat choice
 *         {"s5", "one file"},                        // second ChoiceFormat choice
 *         {"s6", "{0,number} files"},                // third ChoiceFormat choice
 *         {"s7", "3 Mar 96"},                        // sample date
 *         {"s8", new Dimension(1,5)}                 // real object, not just string
 *     // END OF MATERIAL TO LOCALIZE
 *     };
 * }
 *
 * public class MyResources_fr extends ListResourceBundle {
 *     public Object[][] getContents() {
 *         return contents;
 *     }
 *     static final Object[][] contents = {
 *     // LOCALIZE THIS
 *         {"s1", "Le disque \"{1}\" {0}."},          // MessageFormat pattern
 *         {"s2", "1"},                               // location of {0} in pattern
 *         {"s3", "Mon disque"},                      // sample disk name
 *         {"s4", "ne contient pas de fichiers"},     // first ChoiceFormat choice
 *         {"s5", "contient un fichier"},             // second ChoiceFormat choice
 *         {"s6", "contient {0,number} fichiers"},    // third ChoiceFormat choice
 *         {"s7", "3 mars 1996"},                     // sample date
 *         {"s8", new Dimension(1,3)}                 // real object, not just string
 *     // END OF MATERIAL TO LOCALIZE
 *     };
 * }
 * </pre>
 * </blockquote>
 * @see ResourceBundle
 * @see PropertyResourceBundle
 * @since JDK1.1
 */
public abstract class ListResourceBundle extends ResourceBundle {
    /**
     * Sole constructor.  (For invocation by subclass constructors, typically
     * implicit.)
     */
    public ListResourceBundle() {
    }

    // Implements java.util.ResourceBundle.handleGetObject; inherits javadoc specification.
    public final Object handleGetObject(String key) {
        // lazily load the lookup hashtable.
        if (lookup == null) {
            loadLookup();
        }
        return lookup.get(key); // this class ignores locales
    }

    /**
     * Implementation of ResourceBundle.getKeys.
     */
    public Enumeration getKeys() {
        // lazily load the lookup hashtable.
        if (lookup == null) {
            loadLookup();
        }
        Enumeration result = null;
        if (parent != null) {
            final Enumeration myKeys = lookup.keys();
            final Enumeration parentKeys = parent.getKeys();

            result = new Enumeration() {
                public boolean hasMoreElements() {
                    if (temp == null)
                        nextElement();
                    return temp != null;
                }

                public Object nextElement() {
                    Object returnVal = temp;
                    if (myKeys.hasMoreElements())
                        temp = myKeys.nextElement();
                    else {
                        temp = null;
                        while (temp == null && parentKeys.hasMoreElements()) {
                            temp = parentKeys.nextElement();
                            if (lookup.containsKey(temp))
                                temp = null;
                        }
                    }
                    return returnVal;
                }

                Object temp = null;
            };
        } else {
            result = lookup.keys();
        }
        return result;
    }

    /**
     * See class description.
     */
    abstract protected Object[][] getContents();

    // ==================privates====================

    /**
     * We lazily load the lookup hashtable.  This function does the
     * loading.
     */
    private synchronized void loadLookup() {
        if (lookup != null)
            return;

        Object[][] contents = getContents();
        Hashtable temp = new Hashtable(contents.length);
        for (int i = 0; i < contents.length; ++i) {
            temp.put(contents[i][0],contents[i][1]);
        }
        lookup = temp;
    }

    private Hashtable lookup = null;
}
