/*
 * @(#)LogManager.java	1.13 01/02/25
 *
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */


package java.util.logging;

import java.io.*;
import java.util.*;
import java.security.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.net.URL;
import sun.security.action.GetPropertyAction;

/**
 * There is a single global LogManager object that is used to
 * maintain a set of shared state about Loggers and log services.
 * <p>
 * This LogManager object:
 * <ul>
 * <li> Manages a hierarchical namespace of Logger objects.  All
 *      named Loggers are stored in this namespace.
 * <li> Manages a set of logging control properties.  These are
 *      simple key-value pairs that can be used by Handlers and
 *      other logging objects to configure themselves.
 * <li> Manages a list of "global" logging handlers.  By default
 *      Loggers will publish all (relevant) LogRecords to all
 *      the global handlers.
 * </ul>
 * <p>
 * The global LogManager object can be retrieved using LogManager.getLogManager().
 * The LogManager object is created during class initialization and
 * cannot subsequently be changed.
 * <p>
 * <b>Note to reviewers:</b>  The main reason for having a LogManager object
 * rather than simply a class with static methods is to allow container
 * applications (especially J2EE containers) to replace the standard
 * LogManager class with their own subclasses.  We do not support 
 * on-the-fly replacement of the LogManager object as that would 
 * introduce tricky sychnronzation problems in propagating state
 * between the old and the new objects.  Making the LogManager object
 * non-replaceable also allows client code to cache the object.
 * <p>
 * At startup the LogManager class is located using the 
 * java.util.logging.manager system property.
 * <p>
 * By default, the LogManager reads its initial configuration from 
 * a properties file "lib/logging.properties" in the JRE directory.
 * If you edit that property file you can change the default logging
 * configuration for all uses of that JRE.
 * <p>
 * In addition, the LogManager uses two optional system properties that
 * allow more control over reading the initial configuration:
 * <ul>
 * <li>"java.util.logging.config.class"
 * <li>"java.util.logging.config.file"
 * </ul>
 * These two properties may be set via the Preferences API, or as
 * command line property definitions to the "java" command, or as
 * system property definitions passed to JNI_CreateJavaVM.
 * <p>
 * If the "java.util.logging.config.class" property is set, then the
 * property value is treated as a class name.  The given class will be
 * loaded, an object will be instantiated, and that object's constructor
 * is responsible for reading in the initial configuration.  (That object
 * may use other system properties to control its configuration.)
 * <p>
 * If "java.util.logging.config.class" property is <b>not</b> set,
 * then the "java.util.logging.config.file" system property can be used
 * to specify a properties file (in java.util.Properties format). The
 * initial logging configuration will be read from this file.
 * <p>
 * If neither of these properties is defined then, as described
 * above, the LogManager will read its initial configuration from 
 * a properties file "lib/logging.properties" in the JRE directory.
 * <p>
 * The properties for loggers and Handlers will have names starting
 * with the dot-separated name for the handler or logger.
 * <p>
 * The global logging properties may include:
 * <ul>
 * <li>A property "handlers".  This defines a whitespace separated
 * list of class names for handler classes to load and register as
 * global handlers.  Each class name must be for a Handler class
 * which has a default constructor.
 *
 * <li>A property "config".  This property is intended to allow
 * arbitrary configuration code to be run.  The property defines a
 * whitespace separated list of class names.  A new instance will be
 * created for each named class.  The default constructor of each class
 * may execute arbitrary code to update the logging configuration, such as
 * setting logger levels, adding handlers, adding filters, etc.
 * </ul>
 * <p>
 * Note that all classes loaded during LogManager configuration must
 * be on the system class path.  That includes the LogManager class,
 * any config classes, and any handler classes.
 * <p>
 * Loggers are organized into a naming hierarchy based on their
 * dot separated names.  Thus "a.b.c" is a child of "a.b", but
 * "a.b1" and a.b2" are peers.
 * <p>
 * All properties whose names end with ".level" are assumed to define
 * log levels for Loggers.  Thus "foo.level" defines a log level for
 * the logger called "foo" and (recursively) for any of its children
 * in the naming hierarchy.  Log Levels are applied in the order they 
 * are defined in the properties file.  Thus level settings for child
 * nodes in the tree should come after settings for their parents.
 * The property name ".level" can be used to set the level for the
 * root of the tree.
 * <p> 
 * All methods on the LogManager object are multi-thread safe.
 *
 * @version 1.13, 02/25/01
 * @since 1.4
*/

public class LogManager {
    // The global LogManager object
    private static LogManager manager;

    private Vector handlers = new Vector();
    private Properties props = new LogProperties();
    private PropertyChangeSupport changes
			 = new PropertyChangeSupport(LogManager.class);

    // Tree of known loggers
    private LogNode root = new LogNode();

    static {
	if (System.getSecurityManager() == null) {
	    initialize();
	} else {
            // We need to raise privilege here.
	    AccessController.doPrivileged(new PrivilegedAction() {
		public Object run() {
		    initialize();
		    return null;
		}
	    });
	}
    }

    static void initialize() {
	String cname = null;
	try {
	    GetPropertyAction a = new GetPropertyAction("java.util.logging.manager");
	    cname = ((String) AccessController.doPrivileged(a));
	    if (cname != null) {
		Class clz = ClassLoader.getSystemClassLoader().loadClass(cname);
		manager = (LogManager) clz.newInstance();
	    }
	} catch (Exception ex) {
	    System.err.println("Could not load Logmanager \"" + cname + "\"");
	    ex.printStackTrace();
	}
	if (manager == null) {
	    manager = new LogManager();
	}
	try {
	    manager.readConfiguration();
	} catch (Exception ex) {
	    System.err.println("Can't read logging configuration:");
	    System.err.println("" + ex);
	}
    }


    // This private class is used as a shutdown hook.  It tries to close
    // all the open global handlers.
    private class Cleaner extends Thread {
	public void run() {
	    // Make a reasonable effort to close all global handlers.
	    Handler handlers[] = { };
	    try {
	        handlers = getGlobalHandlers();
	        removeAllGlobalHandlers();
	    } catch (Exception ex) {
		// Problems removing handlers?  Keep going..
	    }
	    for (int i = 0; i < handlers.length; i++) {
		try {
		    handlers[i].close();
		} catch (Exception ex) {
		    // Problems closing a handler?  Keep going...
		}
	    }

 	}
    }


    /**     
     * Protected constructor.  This is protected so that container applications
     * (such as J2EE containers) can subclass the object.  It is non-public as
     * it is intended that there only be one LogManager object, whose value is
     * retrieved by calling Logmanager.getLogManager.
     */
    protected LogManager() {
	// Add a shutdown hook to close the global handlers.
	Runtime.getRuntime().addShutdownHook(new Cleaner());
    }

    /**
     * Return the global LogManager object.
     */
    public static LogManager getLogManager() {
	return manager;
    }

    /**
     * Add an event listener to be invoked when the logging
     * properties are re-read.
     *
     * @param l  event listener
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     */
    public void addPropertyChangeListener(PropertyChangeListener l) throws SecurityException {
	checkAccess();
	changes.addPropertyChangeListener(l);
    }

    /**
     * Remove an event listener for property change events.
     * <P>
     * Returns silently if the given listener is not found.
     * 
     * @param l  event listener 
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     */
    public void removePropertyChangeListener(PropertyChangeListener l) throws SecurityException {
	checkAccess();
	changes.removePropertyChangeListener(l);
    }

    /**
     * Add a named logger.  This does nothing and returns false if a logger
     * with the same name is already registered.
     * <p>
     * The Logger factory methods call this method to register each
     * newly created Logger.
     * <p>
     * The application must retain its own reference to the Logger 
     * object to avoid it being garbage collected.  The LogManager
     * will only retain a weak reference.
     *
     * @param   l the new logger.
     * @return  true if the argument logger was registered successfully,
     *          false if a logger of that name already exists.
     * @exception NullPointerException if the logger name is null.
     */
    public synchronized boolean addLogger(Logger l) {
	// Check there is a non-null name.  Do a cheap operation so
	// we'll get a NullPointerException if the name is null.
	l.getName().length();
	LogNode node = findNode(l.getName(), true);
	if (node.getLogger() == null) {
	    node.setLogger(l);
	    return true;
	}
	// We already have a registered logger with the given name.
	return false;
    }

    // Find a node in our tree of logger nodes.
    // If "create" is true, create the node if it doesn't exist.
    private LogNode findNode(String name, boolean create) {
	if (name == null || name.equals("")) {
	    return root;
	}
	LogNode node = root;
	while (name.length() > 0) {
	    int ix = name.indexOf(".");
	    String head;
	    if (ix > 0) {
		head = name.substring(0,ix);
		name = name.substring(ix+1);
	    } else {
		head = name;
		name = "";
	    }
	    if (node.children == null) {
		if (!create) {
		    return null;
		}
		node.children = new HashMap();
	    }
	    LogNode child = (LogNode)node.children.get(head);
	    if (child == null) {
		if (!create) {
		    return null;
		}
		child = new LogNode();
		node.children.put(head, child);
		child.level = node.level;
	    }
	    node = child;
	}
	return node;
    }

    /**
     * Method to find a named logger.
     * <p>
     * Note that since untrusted code may create loggers with
     * arbitrary names this method should not be relied on to
     * find Loggers for security sensitive logging.
     * <p>
     * @param name name of the logger 
     * @return  matching logger or null if none is found
     */
    public Logger getLogger(String name) {
	LogNode node = findNode(name, false);
	if (node == null) {
	    return null;
	}
	return (Logger) node.getLogger();
    }

    /**
     * Set a log level for a given set of loggers.  Subsequently the
     * target loggers will only log messages whose types are greater
     * than or equal to the given level.  The level value Level.OFF
     * can be used to turn off logging.
     * <p>
     * The given log level applies to the named Logger (if it exists),
     * and on any other named Loggers below that name in the naming
     * hierarchy.  The name and level are recorded, and will be applied 
     * to any new Loggers that are later created matching the given name.
     * <p>
     * Thus, setting a level on "x.y" would affect loggers called
     * "x.y", "x.y.z", "x.y.foo.bah" but would not affect loggers
     * called "x.yy" or "x.z".
     * <p>
     * The empty string "" can be used to identify the root.
     *
     * @param   name    The logger name to update.
     * @param   level   the new log level, e.g. Level.SEVERE or Level.FINER
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     */
    public synchronized void setLevel(String name, Level level) throws SecurityException {
	if (level == null || name == null) {
	    throw new NullPointerException();
	}
	checkAccess();
	LogNode node = findNode(name, true);
	node.walkAndSet(level);
    }

    /**
     * Get the log level for a given logger name.
     * See the setLevel method for more details.
     *
     * @param   name    The logger name
     * @return the log level
     */
    public synchronized Level getLevel(String name) {
	LogNode node = root;
	Level level = root.level;

	while (node != null && name.length() > 0) {

	    int ix = name.indexOf(".");
	    String head;
	    if (ix > 0) {
		head = name.substring(0,ix);
		name = name.substring(ix+1);
	    } else {
		head = name;
		name = "";
	    }
	    if (node.children == null) {
		break;
	    }
	    node = (LogNode)node.children.get(head);
	    if (node == null) {
		break;
	    }	    
	    level = node.level;
	}
	return level;
    }

    /**
     * Add a global log Handler to receive output from all loggers.
     * 
     * @param	handler	An output Handler
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     */
    public synchronized void addGlobalHandler(Handler handler) throws SecurityException {
	// Check for null handler
	handler.getClass();
	checkAccess();
	handlers.add(handler);
    }

    /**
     * Remove a global log handler.
     * <P>
     * Returns silently if the given Handler is not found.
     * 
     * @param	handler   An output Handler
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     */
    public synchronized void removeGlobalHandler(Handler handler) throws SecurityException {
	checkAccess();
	handlers.remove(handler);
    }

    /**
     * Remove all global log handlers.
     * 
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     */
    public synchronized void removeAllGlobalHandlers() throws SecurityException {
	checkAccess();
	handlers.clear();
    }

    /**
     * Get an array of all the global Handlers.
     *
     * @return  array of global Handlers
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     */
    public synchronized Handler[] getGlobalHandlers() throws SecurityException {
	checkAccess();
	Handler result[] = new Handler[handlers.size()];
	handlers.copyInto(result);
	return result;
    }

    /**
     * Get an enumeration of known logger names.
     * <p>
     * Note:  Loggers may be added dynamically as new classes are loaded.
     * This method only reports on the loggers that are currently registered.
     * <p>
     * @return  enumeration of logger name strings
     */
    public synchronized Enumeration getLoggerNames() {
	return new LogEnumerator(root);
    }

    /**
     * Reinitialize the logging properties and reread the logging configuration.
     * <p>
     * The same rules are used for locating the configuration properties
     * as are used at startup.  So normally the logging properties will
     * be re-read from the same file that was used at startup.
     * <P>
     * All existing global Handlers will be closed and a new set of global
     * Handlers will be created based on the new properties.
     * <p>
     * Any log level definitions in the new configuration file will be 
     * applied using LogManager.setLevel() in the order they are read.
     * <p>
     * A PropertyChangeEvent will be fired after the properties are read.
     * 
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     * @exception  IOException if there are IO problems reading the configuration.
     */
    public void readConfiguration() throws IOException, SecurityException {
	checkAccess();
	reset();

	// if a configuration class is specified, load it and use it.
	String cname = System.getProperty("java.util.logging.config.class");
	if (cname != null) {
	    try {
		// Instantiate the named class.  It its contructor's
		// repsonsibility to initialize the logging configuration, by
		// calling readConfiguration(InputStream) with a suitable stream.
		Class clz = ClassLoader.getSystemClassLoader().loadClass(cname);
		clz.newInstance();
		return;
	    } catch (Exception ex) {
	        System.err.println("Logging configuration class \"" + cname + "\" failed");
	        System.err.println("" + ex);	    
	        // keep going and useful config file.
	    }
	}

	String fname = System.getProperty("java.util.logging.config.file");
	if (fname == null) {
	    fname = System.getProperty("java.home");
	    if (fname == null) {
		throw new Error("Can't find java.home ??");
	    }
	    File f = new File(fname, "lib");
	    f = new File(f, "logging.properties");
	    fname = f.getCanonicalPath();
	}
	InputStream in = new FileInputStream(fname);
	BufferedInputStream bin = new BufferedInputStream(in);
	try {
	    readConfiguration(bin);
	} finally {
	    if (in != null) {
	    	in.close();
	    }
	}
    }

    // Reset state before (re)reading the configuration file.
    private void reset() {
	props = new LogProperties();
	// Close any global handlers based on old configuration file.
	while (handlers.size() > 0) {
	    Handler h = (Handler)handlers.elementAt(0);
	    removeGlobalHandler(h);
	    try {
		h.close();
	    } catch (Exception ex) {
	    }
	}
    }

    // get a list of whitesace separated classnames from a property.
    private String[] parseClassNames(String propertyName) {
	String hands = props.getProperty(propertyName, "");
	hands = hands.trim();
	int ix = 0;
	Vector result = new Vector();
	while (ix < hands.length()) {
	    int end = ix;
	    while (end < hands.length()) {
		if (Character.isWhitespace(hands.charAt(end))) {
		    break;
		}
		if (hands.charAt(end) == ',') {
		    break;
		}
		end++;
	    }
	    String word = hands.substring(ix, end);
	    ix = end+1;
	    word = word.trim();
	    if (word.length() == 0) {
		continue;
	    }
	    result.add(word);
	}
	return (String[]) result.toArray(new String[result.size()]);
    }

    /**
     * Reinitialize the logging properties and reread the logging configuration
     * from the given stream, which should be in java.util.Properties format.
     * A PropertyChangeEvent will be fired after the properties are read.
     * <p>
     * All existing global Handlers will be closed and a new set of global
     * Handlers will be created based on the new properties.
     * <p>
     * Any log level definitions in the new configuration file will be 
     * applied using LogManager.setLevel() in the order they are read.
     * 
     * @param ins	stream to read properties from
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     * @exception  IOException if there are problems reading from the stream.
     */
    public void readConfiguration(InputStream ins) throws IOException, SecurityException {
	checkAccess();
	reset();

	// Logger levels will get defined as a side effect of reading in
	// the properties.  See LogProperties.put() below.
	props.load(ins);

	// Add new global handlers.
	String names[] = parseClassNames("handlers");
	for (int i = 0; i < names.length; i++) {
	    String word = names[i];
	    try {
		Class clz = ClassLoader.getSystemClassLoader().loadClass(word);
		Handler h = (Handler) clz.newInstance();
		addGlobalHandler(h);
		try {
		    // Check if there is a property defining the
		    // handler's level.
		    String levs = props.getProperty(word + ".level");
		    if (levs != null) {
			h.setLevel(Level.parse(levs));
		    }
		} catch (Exception ex) {
		    // Probably a bad level.
		    // Drop through.
		}
	    } catch (Exception ex) {
		System.err.println("Can't load log handler \"" + word + "\"");
		System.err.println("" + ex);
		ex.printStackTrace();
	    }
	}

	// Instantiate new configuration objects.
	names = parseClassNames("config");
	for (int i = 0; i < names.length; i++) {
	    String word = names[i];
	    try {
		Class clz = ClassLoader.getSystemClassLoader().loadClass(word);
		clz.newInstance();
	    } catch (Exception ex) {
		System.err.println("Can't load config class \"" + word + "\"");
		System.err.println("" + ex);
		// ex.printStackTrace();
	    }
	}

	// Notify any interested parties that our properties have changed.
	changes.firePropertyChange(null, null, null);
    }

    /**
     * Get the value of a logging property.
     * @param name	property name
     * @return		property value
     */
    public String getProperty(String name) {
	return props.getProperty(name);
    }

    // Package private method to get a String property.
    // If the property is not defined we return the given
    // default value.
    String getStringProperty(String name, String defaultValue) {
	String val = getProperty(name);
	if (val == null) {
	    return defaultValue;
	}
	return val.trim();
    }

    // Package private method to get an integer property.
    // If the property is not defined or cannot be parsed
    // we return the given default value.
    int getIntProperty(String name, int defaultValue) {
	String val = getProperty(name);
	if (val == null) {
	    return defaultValue;
	}
	try {
	    return Integer.parseInt(val.trim());
	} catch (Exception ex) {
	    return defaultValue;
	}
    }

    // Package private method to get a Level property.
    // If the property is not defined or cannot be parsed
    // we return the given default value.
    Level getLevelProperty(String name, Level defaultValue) {
	String val = getProperty(name);
	if (val == null) {
	    return defaultValue;
	}
	try {
	    return Level.parse(val.trim());
	} catch (Exception ex) {
	    return defaultValue;
	}
    }

    // Package private method to get a filter property.
    // We return an instance of the class named by the "name" 
    // property. If the property is not defined or has problems
    // we return the defaultValue.
    Filter getFilterProperty(String name, Filter defaultValue) {
	String val = getProperty(name);
	try {
	    if (val != null) {
		Class clz = ClassLoader.getSystemClassLoader().loadClass(val);
	        return (Filter) clz.newInstance();
	    }
	} catch (Exception ex) {
	    // We got one of a variety of exceptions in creating the
	    // class or creating an instance.
	    // Drop through.
	}
	// We got an exception.  Return the defaultValue.
	return defaultValue;
    }


    // Package private method to get a formatter property.
    // We return an instance of the class named by the "name" 
    // property. If the property is not defined or has problems
    // we return the defaultValue.
    Formatter getFormatterProperty(String name, Formatter defaultValue) {
	String val = getProperty(name);
	try {
	    if (val != null) {
		Class clz = ClassLoader.getSystemClassLoader().loadClass(val);
	        return (Formatter) clz.newInstance();
	    }
	} catch (Exception ex) {
	    // We got one of a variety of exceptions in creating the
	    // class or creating an instance.
	    // Drop through.
	}
	// We got an exception.  Return the defaultValue.
	return defaultValue;
    }


    /**
     * Method to post a log record to all global handlers.
     * <p>
     * This is used by Loggers when they wish to publish a LogRecord
     * to all the global handlers.
     */
    public void publish(LogRecord record) {
	for (int i = 0; i < handlers.size(); i++) {
	    Handler h = (Handler)handlers.elementAt(i);
	    h.publish(record);
	}
    }

    /**
     * Flush all the global handlers.
     */
    public void flush() {
	for (int i = 0; i < handlers.size(); i++) {
	    Handler h = (Handler)handlers.elementAt(i);
	    h.flush();
	}
    }

    private Permission ourPermission = new LoggingPermission("control", null);

    /**
     * Check that the current context is trusted to modify the logging
     * configuration.  This requires LoggingPermission("control").
     * <p>
     * If the check fails we throw a SecurityException, otherwise
     * we return normally.
     *
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     */
    public void checkAccess() throws SecurityException {
	SecurityManager sm = System.getSecurityManager();
	if (sm == null) {
	    return;
	}
        sm.checkPermission(ourPermission);
    }

    // Native methods to get thread and process IDs.
    private static native int getThreadID();

    // Nested class to represent a node in our tree of named loggers.
    private static class LogNode {
	HashMap children;
	java.lang.ref.WeakReference loggerRef;
	Level level = Level.INFO;

	Logger getLogger() {
	    if (loggerRef == null) {
		return null;
	    }
	    return (Logger) loggerRef.get();
	}

	void setLogger(Logger l) {
	    loggerRef = new java.lang.ref.WeakReference(l);
	}

	// Internal method to walk the tree below a node and set
        // a given level on each node.
        void walkAndSet(Level level) {
	    this.level = level;
	    Logger l = getLogger();
	    if (l != null) {
	        l.setLevel(level);
	    }
	    if (children == null) {
	        return;
	    }
	    Iterator values = children.values().iterator();
	    while (values.hasNext()) {
	        LogNode node = (LogNode) values.next();
	    	node.walkAndSet(level);
	    }
	}
    }

    // Nested class to enumerate over our logger names.
    private static class LogEnumerator implements Enumeration {
	private Vector names;
	private int index = 0;

	private LogEnumerator(LogNode root) {
	    names = new Vector();
	    addTree(root, "");
   	}

	private void addTree(LogNode node, String name) {
	    // Does the current mode have an associated logger?
	    if (node.getLogger() != null) {
		names.addElement(name);
	    }
	    // Walk this nodes children.
	    if (node.children == null) {
	        return;
	    }
	    Iterator keys = node.children.keySet().iterator();
	    while (keys.hasNext()) {
	        String key = (String) keys.next();
		LogNode child = (LogNode)node.children.get(key);
		String cname;
	  	if (name.length() == 0) {
		    cname = key;
		} else {
		    cname = name + "." + key;
		}
	    	addTree(child, cname);
	    }
	    
	}

	public boolean hasMoreElements() {
	    return (index < names.size());
	}

	public Object nextElement() {
	    if (index >= names.size()) {
		throw new NoSuchElementException();
	    }
	    Object result = names.elementAt(index);
	    index++;
	    return result;
  	}
    }

    // We extend java.util.Properties so that when we are reading
    // properties we can intercept "put" calls and do setLevels
    // in the order they were defined in the property file.
    private class LogProperties extends java.util.Properties {
	public Object put(Object keyObject, Object valueObject) {
	    Object result = super.put(keyObject, valueObject);
	    String key = (String) keyObject;
	    if (!key.endsWith(".level")) {
		// Not a level definition.
		return result;
	    }
	    int ix = key.length() - 6;
	    String name = key.substring(0, ix);
	    String value = (String)valueObject;
	    Level level;
	    try {
		level = Level.parse(value);
	    } catch (Exception ex) {
		System.err.println("Bad level value for property: " + key);
		return result;
	    }
	    setLevel(name, level);
	    return result;
	}
    }
}
