/*
 * @(#)Logger.java	1.18 01/03/06
 *
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */


package java.util.logging;

import java.util.*;
import java.security.*;

/**
 * A Logger object is used to log messages for a specific
 * system or application component.  Loggers are normally named,
 * using a hierarchical dot-separated namespace.  Logger names
 * can be arbitrary strings, but they should normally be based on
 * the package name or class name of the logged component, such
 * as java.net or javax.swing.  In additon it is possible to create
 * "anonymous" Loggers that are not stored in the Logger namespace.
 * <p>
 * Logger objects may be obtained by calls on one of the getLogger
 * factory methods.  These will either create a new Logger or
 * return a suitable existing Logger.
 * <p>
 * Logging messages will be forwarded to registered Handler
 * objects, which can forward the messages to a variety of
 * destinations, including consoles, files, OS logs, etc.
 * <p>
 * On each logging call the Logger initially performs a cheap
 * check of the request level (e.g. SEVERE or FINE) against a
 * log level maintained by the logger.  If the request level is
 * lower than the log level, the logging call returns immediately.
 * <p>
 * The log level is originally initialized based on the properties
 * from the logging configuration file, as described in the description
 * of the LogManager class.  However it may also be dynamically changed
 * by calls on the Logger.setLevel or LogManager.setLevel methods.
 * <p>
 * After passing this initial (cheap) test, the Logger will allocate
 * a LogRecord to describe the logging message.  It will then call a 
 * Filter (if present) to do a more detailed check on whether the
 * record should be published.  If that passes it will then publish
 * the LogRecord to the target Handlers.
 * <p>
 * Most of the logger output methods take a "msg" argument.  This
 * msg argument may be either a raw value or a localization key.
 * During formatting, if the logger has a localization ResourceBundle
 * and if the ResourceBundle has a mapping for the msg string, then
 * the msg string is replaced by the localized value.  Otherwise the
 * original msg string is used.
 * <p>
 * Formatting (including localization) is the responsibility of
 * the output Handler, which will typically call a Formatter.
 * <p>
 * Note that formatting need not occur synchronously.  It may be delayed
 * until a LogRecord is actually written to an external sink.
 * <p>
 * There are various logging convenience methods:
 * <p>
 * First, to make it easier to identify the source of a logging request
 * there are a set of logging methods that take the source class name
 * and the source method name.  For example:
 * <ul>
 * <li> void warning(String sourceClass, string sourceMethod, String msg);
 * </ul>
 * Second, for developers who do not wish to provide this
 * source information there are also methods that do not require
 * these arguments.  For example:
 * <ul>
 * <li> void warning(String msg);
 * </ul>
 * For this second set of methods, the Logging framework will make
 * a "best effort" to determine which class and method called into
 * the logging method.  However, it is important to realize that
 * this automatically inferred information may only be approximate
 * (or may even be quite wrong!).  Virtual machines are allowed to
 * do extensive optimizations when JITing and may entirely remove
 * stack frames, making it impossible to reliably locate the calling
 * class and method.
 * <P>
 * All methods on Logger are multi-thread safe.
 * <p>
 * <b>Subclassing Information:</b> Subclasses of Logger are required
 * to provide their own getLogger factory methods.  In order to
 * intercept all logging output, subclasses need only override the
 * log(LogRecord) method.  All the other log() methods are implemented
 * as calls on this log(LogRecord) method.
 *
 * @version 1.18, 03/06/01
 * @since 1.4
 */


public class Logger {
    private static final int offValue = Level.OFF.intValue();
    private LogManager manager = LogManager.getLogManager();
    private String name;
    private Level levelObject;
    private int levelValue;
    private ArrayList handlers;
    private String resourceBundleName;
    private ResourceBundle catalog;
    private HashMap catalogCache;
    private boolean useGlobalHandlers = true;
    private Filter filter;
    private boolean anonymous;

    /**
     * The "global" Logger object is provided as a convenience to developers
     * who are making casual use of the Logging package.  Developers
     * who are making serious use of the logging package (for example
     * in products) should create and use their own Logger objects,
     * with appropriate names, so that logging can be controlled on a
     * suitable per-Logger granularity.
     * <p>
     * The global logger is initialized by calling Logger.getLogger("global").
     */
    public static final Logger global = getLogger("global");

    /**
     * Protected method to construct a logger for a named subsystem.
     * <p>
     * The logger will be initially configured with a level of Level.ALL
     * and with useGlobalHandlers true.
     *
     * @param	name	A name for the logger.  This should
     *				be a dot-separated name and should normally
     *				be based on the package name or class name
     *				of the subsystem, such as java.net
     *				or javax.swing.  It may be null for anonymous Loggers.
     * @param 	resourceBundleName  name of ResourceBundle to be used for localizing
     *				messages for this logger.  May be null if none
     *				of the messages require localization.
     * @throws MissingResourceException if the ResourceBundleName is non-null and
     *		   no corresponding resource can be found.
     */
    protected Logger(String name, String resourceBundleName) {
	this.resourceBundleName = resourceBundleName;
	if (resourceBundleName != null) {
	    this.catalog = ResourceBundle.getBundle(resourceBundleName);
	}
	this.name = name;
	levelObject = Level.ALL;
	levelValue = levelObject.intValue();
    }


    /**
     * Find or create a logger for a named subsystem.  If a logger has
     * already been created with the given name it is returned.  Otherwise
     * a new logger is created.
     * <p>
     * If a new logger is created its log level will be configured
     * based on the LogManager configuration and it will configured
     * to send logging output to all global output Handlers.  It will
     * be registered in the LogManager global namespace.  This will
     * mean it will be affected by subsequent LogManager.setLevel calls.
     * 
     * @param	name		A name for the logger.  This should
     *				be a dot-separated name and should normally
     *				be based on the package name or class name
     *				of the subsystem, such as java.net
     *				or javax.swing
     * @return a suitable Logger
     */
    public static synchronized Logger getLogger(String name) {
	if (name.equals("")) {
	    throw new IllegalArgumentException("need a logger name");
	}
	LogManager manager = LogManager.getLogManager();
	Logger result = manager.getLogger(name);
	if (result == null) {
	    manager.addLogger(new Logger(name, null));
	    result = manager.getLogger(name);
	    result.levelObject = manager.getLevel(name);
	    result.levelValue = result.levelObject.intValue();
	}
	return result;
    }

    /**
     * Find or create a logger for a named subsystem.  If a logger has 
     * already been created with the given name it is returned.  Otherwise
     * a new logger is created.
     * <p>
     * If a new logger is created its log level will be configured
     * based on the LogManager and it will configured to send logging
     * output to all global output Handlers.  It will be registered in
     * the LogManager global namespace.  This will mean it will be
     * affected by subsequent LogManager.setLevel calls.
     * <p>
     * If the named Logger already exists and does not yet have a
     * localization resource bundle then the given resource bundle 
     * name is used.  If the named Logger already exists and has
     * a different resource bundle name then an IllegalArgumentException
     * is thrown.
     * <p>
     * @param	name	A name for the logger.  This should
     *				be a dot-separated name and should normally
     *				be based on the package name or class name
     *				of the subsystem, such as java.net
     *				or javax.swing
     * @param 	resourceBundleName  name of ResourceBundle to be used for localizing
     *				messages for this logger.
     * @return a suitable Logger
     * @throws MissingResourceException if the named ResourceBundle cannot be found.
     * @throws IllegalArgumentName if the Logger already exists and uses
     *		   a different resource bundle name.
     */
    public static synchronized Logger getLogger(String name, String resourceBundleName) {
	if (name.equals("")) {
	    throw new IllegalArgumentException("need a logger name");
	}
	LogManager manager = LogManager.getLogManager();
	Logger result = manager.getLogger(name);
	if (result == null) {
	    // First check that an appropriate resource bundle exists.
	    // This will throw a MissingResourceBundle if it doesn't.
	    ResourceBundle.getBundle(resourceBundleName);
	    // Now create a new logger.
	    manager.addLogger(new Logger(name, null));
	    result = manager.getLogger(name);
	    result.levelObject = manager.getLevel(name);
	    result.levelValue = result.levelObject.intValue();
	}
	if (result.resourceBundleName == null) {
	    // Note: We are careful to call ResourceBundle.getBundle (which may throw
	    // a MissingResourceException) before setting the resourceBundleName.
	    result.catalog = ResourceBundle.getBundle(resourceBundleName);
	    result.resourceBundleName = resourceBundleName;
	} else if (!result.resourceBundleName.equals(resourceBundleName)) {
	    throw new IllegalArgumentException(result.resourceBundleName +
				" != " + resourceBundleName);
	}
	return result;
    }


    /**
     * Create an anonymous Logger.  The newly created Logger is not
     * registered in the LogManager namespace and will not be
     * affected by updates to the logging configuration or by calls
     * on LogManager.setLevel.  There will also be no access checks on
     * updates to the logger.
     * <p>
     * This factory method is primarily intended for use from applets.
     * Because the resulting Logger is anonymous it can be kept private
     * by the creating class.  This removes the need for normal security
     * checks, which in turn allows untrusted applet code to update
     * the control state of the Logger.  For example an applet can do
     * a setLevel or an addHandler on an anonymous Logger.
     * <p>
     * The new logger is initially configured with the logging level
     * for the root name ("") from the LogManager class.  It is
     * initially configured to use the LogManager's global handlers.
     *
     * @return a newly created private Logger
     */
    public static synchronized Logger getAnonymousLogger() {
	LogManager manager = LogManager.getLogManager();
	Logger result = new Logger(null, null);
	result.anonymous = true;
	result.setLevel(manager.getLevel(""));
	return result;
    }

    /**
     * Create an anonymous Logger.  The newly created Logger is not
     * registered in the LogManager namespace and will not be
     * affected by updates to the logging configuration or by calls
     * on LogManager.setLevel.  There will also be no access checks on
     * updates to the Logger.
     * <p>
     * This factory method is primarily intended for use from applets.
     * Because the resulting Logger is anonymous it can be kept private
     * by the creating class.  This removes the need for normal security
     * checks, which in turn allows untrusted applet code to update
     * the control state of the Logger.  For example an applet can do
     * a setLevel or an addHandler on an anonymous Logger.
     * <p>
     * The new logger is initially configured with the logging level
     * for the root name ("") from the LogManager class.  It is
     * initially configured to use the LogManager's global handlers.
     * <p>
     * @param 	resourceBundleName  name of ResourceBundle to be used for localizing
     *				messages for this logger.
     * @return a newly created private Logger
     * @throws MissingResourceException if the named ResourceBundle cannot be found.
     */
    public static synchronized Logger getAnonymousLogger(String resourceBundleName) {
	LogManager manager = LogManager.getLogManager();
	Logger result = new Logger(null, null);
	result.anonymous = true;
	result.setLevel(manager.getLevel(""));
	result.resourceBundleName = resourceBundleName;
	result.catalog = ResourceBundle.getBundle(resourceBundleName);
	return result;
    }

    /**
     * Retrieve the localization resource bundle for this
     * logger.
     * @return localization bundle (may be null)
     */
    public ResourceBundle getResourceBundle() {
	return catalog;
    }

    /**
     * Retrieve the localization resource bundle name for this
     * logger.
     * @return localization bundle name (may be null)
     */
    public String getResourceBundleName() {
	return resourceBundleName;
    }

    /**
     * Set a filter to control output on this Logger.
     * <P>
     * After passing the initial "level" check, the Logger will
     * call this Filter to check if a log record should really
     * be published.
     *
     * @param   newFilter  a filter object (may be null)
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     */
    public void setFilter(Filter newFilter) throws SecurityException {
	if (!anonymous) {
	    manager.checkAccess();
	}
	filter = newFilter;
    }

    /**
     * Get the current filter for this Logger.
     *
     * @return  a filter object (may be null)
     */
    public Filter getFilter() {
	return filter;
    }

    /**
     * Log a LogRecord.
     * <p>
     * All the other logging methods in this class call through
     * this method to actually perform any logging.  Subclasses can
     * override this single method to capture all log activity.
     * <p>
     * The current logger's name, resource bundle, and resource bundle
     * name will be set into the LogRecord.
     *
     * @param record the LogRecord to be published
     */
    public synchronized void log(LogRecord record) {
	record.setLoggerName(name);
	record.setResourceBundle(catalog);
	record.setResourceBundleName(resourceBundleName);

	if (record.getLevel().intValue() < levelValue || levelValue == offValue) {
	    return;
	}
	if (filter != null && !filter.isLoggable(record)) {
	    return;
	}
	// First post to all global handlers.
	if (useGlobalHandlers) {
	    manager.publish(record);
	}
	// Now post to all local handlers.
	if (handlers != null) {
	    for (int i = 0; i < handlers.size(); i++) {
		Handler h = (Handler)handlers.get(i);
		h.publish(record);
	    }
	}
    }

    //================================================================
    // Start of convenience methods WITH className and methodName
    //================================================================

    /**
     * Log a message, specifying source class and method,
     * with no arguments.
     * <p>
     * If the logger is currently enabled for the given message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param	level	One of the message level identifiers, e.g. SEVERE
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void log(Level level, String sourceClass, String sourceMethod, String msg) {
	if (level.intValue() < levelValue || levelValue == offValue) {
	    return;
	}
	LogRecord lr = new LogRecord(level, msg);
	lr.setSourceClassName(sourceClass);
	lr.setSourceMethodName(sourceMethod);
	log(lr);
    }

    /**
     * Log a message, specifying source class and method,
     * with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the given message 
     * level then a corresponding LogRecord is created and forwarded 
     * to all the registered output Handler objects.
     * <p>
     * @param	level   One of the message level identifiers, e.g. SEVERE
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void log(Level level, String sourceClass, String sourceMethod,
						String msg, Object params[]) {
	if (level.intValue() < levelValue || levelValue == offValue) {
	    return;
	}
	LogRecord lr = new LogRecord(level, msg);
	lr.setSourceClassName(sourceClass);
	lr.setSourceMethodName(sourceMethod);
	lr.setParameters(params);
	log(lr);
    }

    /**
     * Log a message, specifying source class and method,
     * with associated Throwable information.
     * <p>
     * If the logger is currently enabled for the given message 
     * level then the given arguments are stored in a LogRecord
     * which is forwarded to all registered output handlers.
     * <p>
     * Note that the thrown argument is stored in the LogRecord thrown
     * property, rather than the LogRecord parameters property.  Thus is it
     * processed specially by output Formatters and is not treated
     * as a formatting parameter to the LogRecord message property.
     * <p>
     * @param	level   One of the message level identifiers, e.g. SEVERE
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     * @param   thrown  Throwable associated with log message.
     */
    public void log(Level level, String sourceClass, String sourceMethod,
							String msg, Throwable thrown) {
	if (level.intValue() < levelValue || levelValue == offValue) {
	    return;
	}
	LogRecord lr = new LogRecord(level, msg);
	lr.setSourceClassName(sourceClass);
	lr.setSourceMethodName(sourceMethod);
	lr.setThrown(thrown);
	log(lr);
    }

    /**
     * Log a procedure entry.
     * <p>
     * This is a convenience method that can be used to log entry
     * to a method.  A LogRecord with message "ENTRY", log level
     * FINER, and the given sourceMethod and sourceClass is logged.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that is being entered
     */
    public void entering(String sourceClass, String sourceMethod) {
	if (Level.FINER.intValue() < levelValue) {
	    return;
	}
	log(Level.FINER, sourceClass, sourceMethod, "ENTRY");
    }

    /**
     * Log a procedure entry, with parameters.
     * <p>
     * This is a convenience method that can be used to log entry
     * to a method.  A LogRecord with message "ENTRY", log level
     * FINER, and the given sourceMethod, sourceClass, and params
     * is logged.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that is being entered
     * @param   params	array of parameters to the method being entered
     */
    public void entering(String sourceClass, String sourceMethod, Object params[]) {
	if (Level.FINER.intValue() < levelValue) {
	    return;
	}
	log(Level.FINER, sourceClass, sourceMethod, "ENTRY", params);
    }

    /**
     * Log a procedure return.
     * <p>
     * This is a convenience method that can be used to log returning
     * from a method.  A LogRecord with message "RETURN", log level
     * FINER, and the given sourceMethod and sourceClass is logged.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of the method 
     */
    public void exiting(String sourceClass, String sourceMethod) {
	if (Level.FINER.intValue() < levelValue) {
	    return;
	}
	log(Level.FINER, sourceClass, sourceMethod, "RETURN");
    }

    /**
     * Log a procedure return, with result object.
     * <p>
     * This is a convenience method that can be used to log returning
     * from a method.  A LogRecord with message "RETURN", log level
     * FINER, and the gives sourceMethod, sourceClass, and result
     * object is logged.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of the method 
     * @param   result  Object that is being returned
     */
    public void exiting(String sourceClass, String sourceMethod, Object result) {
	if (Level.FINER.intValue() < levelValue) {
	    return;
	}
 	Object params[] = { result };
	log(Level.FINER, sourceClass, sourceMethod, "RETURN", params);
    }

    /**
     * Log throwing an exception.
     * <p>
     * This is a convenience method to log that a method is
     * terminating by throwing an exception.  The logging is done 
     * using the FINER level.
     * <p>
     * If the logger is currently enabled for the given message 
     * level then the given arguments are stored in a LogRecord
     * which is forwarded to all registered output handlers.  The
     * LogRecord's message is set to "THROW".
     * <p>
     * Note that the thrown argument is stored in the LogRecord thrown
     * property, rather than the LogRecord parameters property.  Thus is it
     * processed specially by output Formatters and is not treated
     * as a formatting parameter to the LogRecord message property.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod  name of the method.
     * @param   thrown  The Throwable that is being thrown.
     */
    public void throwing(String sourceClass, String sourceMethod, Throwable thrown) {
	if (Level.FINER.intValue() < levelValue) {
	    return;
	}
	LogRecord lr = new LogRecord(Level.FINER, "THROW");
	lr.setSourceClassName(sourceClass);
	lr.setSourceMethodName(sourceMethod);
	lr.setThrown(thrown);
	log(lr);
    }

    /**
     * Log a SEVERE message, specifying source class and method,
     * with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the SEVERE message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void severe(String sourceClass, String sourceMethod, String msg, Object params[]) {
	if (Level.SEVERE.intValue() < levelValue) {
	    return;
	}
	log(Level.SEVERE, sourceClass, sourceMethod, msg, params);
    }

    /**
     * Log a SEVERE message, specifying source class and method.
     * <p>
     * If the logger is currently enabled for the SEVERE message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void severe(String sourceClass, String sourceMethod, String msg) {
	if (Level.SEVERE.intValue() < levelValue) {
	    return;
	}
	log(Level.SEVERE ,sourceClass, sourceMethod, msg);
    }

    /**
     * Log a WARNING message, specifying source class and method,
     * with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the WARNING message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void warning(String sourceClass, String sourceMethod, String msg, Object params[]) {
	if (Level.WARNING.intValue() < levelValue) {
	    return;
	}
	log(Level.WARNING, sourceClass, sourceMethod, msg, params);
    }

    /**
     * Log a WARNING message, specifying source class and method.
     * <p>
     * If the logger is currently enabled for the WARNING message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void warning(String sourceClass, String sourceMethod, String msg) {
	if (Level.WARNING.intValue() < levelValue) {
	    return;
	}
	log(Level.WARNING, sourceClass, sourceMethod, msg);
    }


    /**
     * Log an INFO message, specifying source class and method,
     * with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the INFO message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void info(String sourceClass, String sourceMethod, String msg, Object params[]) {
	if (Level.INFO.intValue() < levelValue) {
	    return;
	}
	log(Level.INFO, sourceClass, sourceMethod, msg, params);
    }

    /**
     * Log an INFO message, specifying source class and method.
     * <p>
     * If the logger is currently enabled for the INFO message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void info(String sourceClass, String sourceMethod, String msg) {
	if (Level.INFO.intValue() < levelValue) {
	    return;
	}
	log(Level.INFO, sourceClass, sourceMethod, msg);
    }

    /**
     * Log a CONFIG message, specifying source class and method,
     * with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the CONFIG message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void config(String sourceClass, String sourceMethod, String msg, Object params[]) {
	if (Level.CONFIG.intValue() < levelValue) {
	    return;
	}
	log(Level.CONFIG, sourceClass, sourceMethod, msg, params);
    }

    /**
     * Log a CONFIG message, specifying source class and method.
     * <p>
     * If the logger is currently enabled for the CONFIG message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void config(String sourceClass, String sourceMethod, String msg) {
	if (Level.CONFIG.intValue() < levelValue) {
	    return;
	}
	log(Level.CONFIG, sourceClass, sourceMethod, msg);
    }

    /**
     * Log a FINE message, specifying source class and method,
     * with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the FINE message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void fine(String sourceClass, String sourceMethod, String msg, Object params[]) {
	if (Level.FINE.intValue() < levelValue) {
	    return;
	}
	log(Level.FINE, sourceClass, sourceMethod, msg, params);
    }

    /**
     * Log a FINE message, specifying source class and method.
     * <p>
     * If the logger is currently enabled for the FINE message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void fine(String sourceClass, String sourceMethod, String msg) {
	if (Level.FINE.intValue() < levelValue) {
	    return;
	}
	log(Level.FINE, sourceClass, sourceMethod, msg);
    }

    /**
     * Log a FINER message, specifying source class and method,
     * with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the FINER message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void finer(String sourceClass, String sourceMethod, String msg, Object params[]) {
	if (Level.FINER.intValue() < levelValue) {
	    return;
	}
	log(Level.FINER, sourceClass, sourceMethod, msg, params);
    }

    /**
     * Log a FINER message, specifying source class and method.
     * <p>
     * If the logger is currently enabled for the FINER message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void finer(String sourceClass, String sourceMethod, String msg) {
	if (Level.FINER.intValue() < levelValue) {
	    return;
	}
	log(Level.FINER, sourceClass, sourceMethod, msg);
    }

    /**
     * Log a FINEST message, specifying source class and method,
     * with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the FINEST message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void finest(String sourceClass, String sourceMethod, String msg, Object params[]) {
	if (Level.FINEST.intValue() < levelValue) {
	    return;
	}
	log(Level.FINEST, sourceClass, sourceMethod, msg, params);
    }

    /**
     * Log a FINEST message, specifying source class and method.
     * <p>
     * If the logger is currently enabled for the FINEST message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   sourceClass    name of class that issued the logging request
     * @param   sourceMethod   name of method that issued the logging request
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void finest(String sourceClass, String sourceMethod, String msg) {
	if (Level.FINEST.intValue() < levelValue) {
	    return;
	}
	log(Level.FINEST, sourceClass, sourceMethod, msg);
    }

    //================================================================
    // Start of convenience methods WITHOUT className and methodName
    //================================================================


    // Private method to infer the caller's class and method names
    private void inferCaller(LogRecord r) {
	// This is an initial implementation.
	// We may get VM support later.
	try {
	    java.io.StringWriter sw = new java.io.StringWriter();
	    java.io.PrintWriter pw = new java.io.PrintWriter(sw);
	    (new Exception("")).printStackTrace(pw);
	    pw.close();
	    String trace = sw.toString();
	    // Find first call on Logger.
	    int ix = trace.lastIndexOf("java.util.logging.Logger");
	    ix = trace.indexOf("(", ix)+3;
	     // Now skip to the next stack frame.
	    int end = trace.indexOf("(", ix);
	    int dot = trace.lastIndexOf(".", end);
	    int start = trace.lastIndexOf(" ", dot)+1;
	    r.setSourceClassName(trace.substring(start, dot));
	    r.setSourceMethodName(trace.substring(dot+1, end));
	} catch (Exception ex) {
	    // Something went wrong.  We punt.  This is legal, we
	    // are only commited to making a "best effort" here.
	    return;
	}
    }

    /**
     * Log a message, with no arguments.
     * <p>
     * If the logger is currently enabled for the given message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param	level	One of the message level identifiers, e.g. SEVERE
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void log(Level level, String msg) {
	if (level.intValue() < levelValue || levelValue == offValue) {
	    return;
	}
	LogRecord lr = new LogRecord(level, msg);
	inferCaller(lr);
	log(lr);
    }

    /**
     * Log a message, with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the given message 
     * level then a corresponding LogRecord is created and forwarded 
     * to all the registered output Handler objects.
     * <p>
     * @param	level   One of the message level identifiers, e.g. SEVERE
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void log(Level level, String msg, Object params[]) {
	if (level.intValue() < levelValue || levelValue == offValue) {
	    return;
	}
	LogRecord lr = new LogRecord(level, msg);
	lr.setParameters(params);
	inferCaller(lr);
	log(lr);
    }

    /**
     * Log a message, with associated Throwable information.
     * <p>
     * If the logger is currently enabled for the given message 
     * level then the given arguments are stored in a LogRecord
     * which is forwarded to all registered output handlers.
     * <p>
     * Note that the thrown argument is stored in the LogRecord thrown
     * property, rather than the LogRecord parameters property.  Thus is it
     * processed specially by output Formatters and is not treated
     * as a formatting parameter to the LogRecord message property.
     * <p>
     * @param	level   One of the message level identifiers, e.g. SEVERE
     * @param   msg	The string message (or a key in the message catalog)
     * @param   thrown  Throwable associated with log message.
     */
    public void log(Level level, String msg, Throwable thrown) {
	if (level.intValue() < levelValue || levelValue == offValue) {
	    return;
	}
	LogRecord lr = new LogRecord(level, msg);
	lr.setThrown(thrown);
	inferCaller(lr);
	log(lr);
    }

    /**
     * Log a SEVERE message, with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the SEVERE message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void severe(String msg, Object params[]) {
	if (Level.SEVERE.intValue() < levelValue) {
	    return;
	}
	log(Level.SEVERE, msg, params);
    }

    /**
     * Log a SEVERE message.
     * <p>
     * If the logger is currently enabled for the SEVERE message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void severe(String msg) {
	if (Level.SEVERE.intValue() < levelValue) {
	    return;
	}
	log(Level.SEVERE, msg);
    }

    /**
     * Log a WARNING message, with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the WARNING message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void warning(String msg, Object params[]) {
	if (Level.WARNING.intValue() < levelValue) {
	    return;
	}
	log(Level.WARNING, msg, params);
    }

    /**
     * Log a WARNING message.
     * <p>
     * If the logger is currently enabled for the WARNING message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void warning(String msg) {
	if (Level.WARNING.intValue() < levelValue) {
	    return;
	}
	log(Level.WARNING, msg);
    }


    /**
     * Log an INFO message, with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the INFO message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void info(String msg, Object params[]) {
	if (Level.INFO.intValue() < levelValue) {
	    return;
	}
	log(Level.INFO, msg, params);
    }

    /**
     * Log an INFO message.
     * <p>
     * If the logger is currently enabled for the INFO message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void info(String msg) {
	if (Level.INFO.intValue() < levelValue) {
	    return;
	}
	log(Level.INFO, msg);
    }

    /**
     * Log a CONFIG message, with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the CONFIG message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void config(String msg, Object params[]) {
	if (Level.CONFIG.intValue() < levelValue) {
	    return;
	}
	log(Level.CONFIG, msg, params);
    }

    /**
     * Log a CONFIG message.
     * <p>
     * If the logger is currently enabled for the CONFIG message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void config(String msg) {
	if (Level.CONFIG.intValue() < levelValue) {
	    return;
	}
	log(Level.CONFIG, msg);
    }

    /**
     * Log a FINE message, with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the FINE message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void fine(String msg, Object params[]) {
	if (Level.FINE.intValue() < levelValue) {
	    return;
	}
	log(Level.FINE, msg, params);
    }

    /**
     * Log a FINE message.
     * <p>
     * If the logger is currently enabled for the FINE message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void fine(String msg) {
	if (Level.FINE.intValue() < levelValue) {
	    return;
	}
	log(Level.FINE, msg);
    }

    /**
     * Log a FINER message, with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the FINER message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void finer(String msg, Object params[]) {
	if (Level.FINER.intValue() < levelValue) {
	    return;
	}
	log(Level.FINER, msg, params);
    }

    /**
     * Log a FINER message.
     * <p>
     * If the logger is currently enabled for the FINER message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void finer(String msg) {
	if (Level.FINER.intValue() < levelValue) {
	    return;
	}
	log(Level.FINER, msg);
    }

    /**
     * Log a FINEST message, with an array of object arguments.
     * <p>
     * If the logger is currently enabled for the FINEST message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     * @param   params	array of parameters to the message
     */
    public void finest(String msg, Object params[]) {
	if (Level.FINEST.intValue() < levelValue) {
	    return;
	}
	log(Level.FINEST, msg, params);
    }

    /**
     * Log a FINEST message.
     * <p>
     * If the logger is currently enabled for the FINEST message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     */
    public void finest(String msg) {
	if (Level.FINEST.intValue() < levelValue) {
	    return;
	}
	log(Level.FINEST, msg);
    }

    //================================================================
    // End of convenience methods 
    //================================================================

    /**
     * Set the log level specifying which message levels will be
     * logged by this logger.  Message levels lower than this
     * value will be discarded.  The level value Level.OFF
     * can be used to turn off logging.
     * 
     * @param newLevel   the new value for the log level
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     */
    public synchronized void setLevel(Level newLevel) throws SecurityException {
	if (newLevel == null) {
	    throw new NullPointerException();
	}
	if (!anonymous) {
	    manager.checkAccess();
	}
	levelObject = newLevel;
	levelValue = levelObject.intValue();
    }

    /**
     * Get the log level specifying which messages will be
     * logged by this logger.  Message levels lower than this
     * level will be discarded.
     * @return	the level of messages being logged.
     */
    public Level getLevel() {
	return levelObject;
    }

    /**
     * Check if a message of the given level would actually be logged
     * by this logger.
     * @param	level	a message logging level
     * @return	true if the given message level is currently being logged.
     */
    public boolean isLoggable(Level level) {
	if (level.intValue() < levelValue || levelValue == offValue) {
	    return false;
	}
	return true;
    }

    /**
     * Get the name for this logger.
     * @return logger name.  Will be null for anonymous Loggers.
     */
    public String getName() {
	return name;
    }

    /**
     * Add a log Handler to receive logging messages.
     * <p>
     * By default, Loggers send their output to all the global handlers.
     * Application programmers only need to use the addHandler method if
     * they are configuring a Logger with a non-global output handler.
     * Adding a non-global handler does not affect the use of global handlers.
     *
     * @param	handler	a logging Handler
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     */
    public synchronized void addHandler(Handler handler) throws SecurityException {
	// Check for null handler
	handler.getClass();
	if (!anonymous) {
	    manager.checkAccess();
	}
	if (handlers == null) {
	    handlers = new ArrayList();
	}
	handlers.add(handler);
    }

    /**
     * Remove a log Handler.
     * <P>
     * Returns silently if the given Handler is not found.
     * 
     * @param	handler	a logging Handler
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     */
    public synchronized void removeHandler(Handler handler) throws SecurityException {
	if (!anonymous) {
	    manager.checkAccess();
	}
	if (handlers == null) {
	    return;
	}
	handlers.remove(handler);
    }

    /**
     * Get the Handlers associated with this logger.
     * <p>
     * The resulting array includes all Handlers set with addHandler,
     * but excludes global Handlers.
     * <p>
     * @return  an array of all registered Handlers
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     */
    public synchronized Handler[] getHandlers() throws SecurityException {
	if (!anonymous) {
	    manager.checkAccess();
	}
	if (handlers == null) {
	    return new Handler[0];
	}
	Handler result[] = new Handler[handlers.size()];
	result = (Handler [])handlers.toArray(result);
	return result;
    }

    /**
     * Specify whether or not this logger should send its output
     * to the set of global handlers managed by the LogManager.
     * @param useGlobalhandlers   true if output is to be sent to
     *		the LogManager's global handlers.
     * @exception  SecurityException  if a security manager exists and if
     *             the caller does not have LoggingPermission("control").
     */
    public void setUseGlobalHandlers(boolean useGlobalHandlers) {
	if (!anonymous) {
	    manager.checkAccess();
	}
	this.useGlobalHandlers = useGlobalHandlers;
    }

    /**
     * Discover whether or not this logger is sending its output
     * to the set of global handlers managed by the LogManager.
     * @return  true if output is to be sent to
     *		the LogManager's global handlers.
     */
    public boolean getUseGlobalHandlers() {
	return useGlobalHandlers;
    }
}
