/*
 * @(#)XMLFormatter.java	1.10 00/09/20
 *
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */


package java.util.logging;

import java.io.*;
import java.util.*;

/**
 * Format a LogRecord into a standard XML format.
 * <p>
 * The DTD specification is provided as Appendix A to the
 * Java Logging APIs specification.
 * <p>
 * The XMLFormatter can be used with arbitrary character encodings,
 * but it is recommended that it normally be used with UTF-8.  The
 * character encoding can be set on the output Handler.
 *
 * @version 1.10, 09/20/00
 * @since 1.4
 */

public class XMLFormatter extends Formatter {
    private LogManager manager = LogManager.getLogManager();

    // Append a two digit number.
    private void a2(StringBuffer sb, int x) {
	if (x < 10) {
	    sb.append('0');
	}
	sb.append(x);
    }

    // Append the time and date in ISO 8601 format
    private void appendISO8601(StringBuffer sb, long millis) {
	Date date = new Date(millis);
	sb.append(date.getYear() + 1900);
	sb.append('-');
	a2(sb, date.getMonth() + 1);
	sb.append('-');
	a2(sb, date.getDate());
	sb.append('T');
	a2(sb, date.getHours());
	sb.append(':');
	a2(sb, date.getMinutes());
	sb.append(':');
	a2(sb, date.getSeconds());
    }

    /**
     * Format the given message to XML.
     * @param record the log record to be formatted.
     * @return a formatted log record
     */
    public String format(LogRecord record) {
	StringBuffer sb = new StringBuffer(500);
	sb.append("<record>\n");

	sb.append("  <date>");
	appendISO8601(sb, record.getMillis());
	sb.append("</date>\n");

	sb.append("  <millis>");
	sb.append(record.getMillis());
	sb.append("</millis>\n");

	sb.append("  <sequence>");
	sb.append(record.getSequenceNumber());
	sb.append("</sequence>\n");

	sb.append("  <logger>");
	sb.append(record.getLoggerName());
	sb.append("</logger>\n");

	sb.append("  <level>"); 
	sb.append(record.getLevel().toString());
	sb.append("</level>\n");

	if (record.getSourceClassName() != null) {
	    sb.append("  <class>");
	    sb.append(record.getSourceClassName());
	    sb.append("</class>\n");
	}

	if (record.getSourceMethodName() != null) {
	    sb.append("  <method>");
	    sb.append(record.getSourceMethodName());
	    sb.append("</method>\n");
	}

	sb.append("  <thread>");
	sb.append(record.getThreadID());
	sb.append("</thread>\n");

	if (record.getMessage() != null) {
	    // Format the message string and its accompanying parameters.
	    String message = formatMessage(record);
	    // NOTE: Need to make sure this generates clean XML.
	    sb.append("  <message>");
	    sb.append(message);
	    sb.append("</message>");
	    sb.append("\n");
	}

	// If the message is being localized, output the key, resource
  	// bundle name, and params.
	ResourceBundle bundle = record.getResourceBundle();
	try {
	    if (bundle != null && bundle.getString(record.getMessage()) != null) {
	        sb.append("  <key>");
	        sb.append(record.getMessage());
	        sb.append("</key>\n");
	        sb.append("  <catalog>");
	        sb.append(record.getResourceBundleName());
	        sb.append("</catalog>\n");
		Object parameters[] = record.getParameters();
	        for (int i = 0; i < parameters.length; i++) {
	            sb.append("  <param>");
		    try {
	                sb.append(parameters[i].toString());
		    } catch (Exception ex) {
		        sb.append("???");
		    }
	            sb.append("</param>\n");
		}
	    }
	} catch (Exception ex) {
	    // The message is not in the catalog.  Drop through.
	}

	if (record.getThrown() != null) {
	    // Report on the state of the throwable.
	    // This is temporary and will get cleaned up to use the
	    // new Throwable APIs Josh is adding in Merlin.
	    sb.append("  <exception>\n");
	    try {
	        Throwable th = record.getThrown();
	        sb.append("    <message>");
	        sb.append(th.toString());
	        sb.append("</message>\n");
		String stack = record.getThrownBackTrace();
		int ix = 0;
		while (ix < stack.length()) {
		    ix = stack.indexOf("at ", ix);
		    if (ix < 0) {
			break;
		    }
		    int start = ix+3;
	            int end = stack.indexOf("(", start);
		    if (end < 0) {
			break;
		    }
		    int dot = stack.lastIndexOf(".", end);

		    sb.append("    <frame>\n");
		    sb.append("      <class>");
		    sb.append(stack.substring(start,dot));
		    sb.append("</class>\n");
		    sb.append("      <method>");
		    sb.append(stack.substring(dot+1,end));
		    sb.append("</method>\n");

		    // Check for a line number.
		    start = stack.indexOf(":", end);
		    end = stack.indexOf(")", end);
		    if (start > 0 && start < end) {
			sb.append("      <line>");
			sb.append(stack.substring(start+1, end));
			sb.append("</line>\n");
		    }
		    sb.append("    </frame>\n");
		    ix = end;
		}
	    } catch (Exception ex) {
	    }
	    sb.append("  </exception>\n");
	}

	sb.append("</record>\n");
	return sb.toString();
    }

    /**
     * Return the header string for a set of XML formatted records.
     * 
     * @param   h  The target handler.
     * @return  header string
     */
    public String getHead(Handler h) {
	StringBuffer sb = new StringBuffer();
	sb.append("<?xml version=\"1.0\"");
	String encoding = h.getEncoding();
	if (encoding != null && encoding.equals("UTF8")) {
	    encoding = "UTF-8";
	}
	if (encoding != null) {
	    sb.append(" encoding=\"");
	    sb.append(encoding);
	    sb.append("\"");
	}
	sb.append(" standalone=\"no\"?>\n");
	sb.append("<!DOCTYPE log SYSTEM \"file:logger.dtd\">\n");
	sb.append("<log>\n");
	return sb.toString();
    }

    /**
     * Return the tail string for a set of XML formatted records.
     * 
     * @param   h  The target handler.
     * @return  tail string
     */
    public String getTail(Handler h) {
	return "</log>\n";
    }
}
