/*
 * @(#)HardCodedFocusTraversalPolicy.java	1.1 00/09/06
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
package javax.swing;

import java.awt.FocusTraversalPolicy;
import java.awt.Component;
import java.awt.Container;
import java.awt.Window;
import java.util.HashMap;
import java.util.HashSet;
import java.io.*;


/**
 * A FocusTraversalPolicy which provides support for legacy applications which
 * hard code focus traversal via JComponent.setNextFocusableComponent. If a
 * specific traversal has not been hard coded, then that traversal is provided
 * by a wrapped FocusTraversalPolicy instance.
 *
 * @version 1.1, 09/06/00
 * @author David Mendenhall
 */
final class HardCodedFocusTraversalPolicy extends FocusTraversalPolicy
    implements Serializable
{
    private transient FocusTraversalPolicy delegate;
    private HashMap forwardMap = new HashMap(),
	backwardMap = new HashMap();

    HardCodedFocusTraversalPolicy(FocusTraversalPolicy delegate) {
	this.delegate = delegate;
    }

    void setNextFocusableComponent(Component left, Component right) {
	forwardMap.put(left, right);
	backwardMap.put(right, left);
    }
    void unsetNextFocusableComponent(Component left, Component right) {
	forwardMap.remove(left);
	backwardMap.remove(right);
    }

    public Component getComponentAfter(Container focusCycleRoot,
				       Component aComponent) {
	Component hardCoded = aComponent, prevHardCoded;
	HashSet sanity = new HashSet();

	do {
	    prevHardCoded = hardCoded;
	    hardCoded = (Component)forwardMap.get(hardCoded);
	    if (hardCoded == null) {
		if (delegate != null) {
		    return delegate.getComponentAfter(focusCycleRoot,
						      prevHardCoded);
		} else {
		    return null;
		}
	    }
	    if (sanity.contains(hardCoded)) {
		// cycle detected; bail
		return null;
	    }
	    sanity.add(hardCoded);
	} while (!accept(hardCoded));

	return hardCoded;
    }
    public Component getComponentBefore(Container focusCycleRoot,
					Component aComponent) {
	Component hardCoded = aComponent, prevHardCoded;
	HashSet sanity = new HashSet();

	do {
	    prevHardCoded = hardCoded;
	    hardCoded = (Component)backwardMap.get(hardCoded);
	    if (hardCoded == null) {
		if (delegate != null) {
		    return delegate.getComponentBefore(focusCycleRoot,
						       prevHardCoded);
		} else {
		    return null;
		}
	    }
	    if (sanity.contains(hardCoded)) {
		// cycle detected; bail
		return null;
	    }
	    sanity.add(hardCoded);
	} while (!accept(hardCoded));

	return hardCoded;
    }
    public Component getFirstComponent(Container focusCycleRoot) {
	if (delegate != null) {
	    return delegate.getFirstComponent(focusCycleRoot);
	} else {
	    return null;
	}
    }
    public Component getLastComponent(Container focusCycleRoot) {
	if (delegate != null) {
	    return delegate.getLastComponent(focusCycleRoot);
	} else {
	    return null;
	}
    }
    public Component getDefaultComponent(Container focusCycleRoot) {
	if (delegate != null) {
	    return delegate.getDefaultComponent(focusCycleRoot);
	} else {
	    return null;
	}
    }
    private boolean accept(Component aComponent) {
        if (!(aComponent.isVisible() && aComponent.isDisplayable() &&
              aComponent.isFocusable() && aComponent.isEnabled())) {
            return false;
        }

        // Verify that the Component is recursively enabled. Disabling a
        // heavyweight Container disables its children, whereas disabling
        // a lightweight Container does not.
        if (!(aComponent instanceof Window)) {
            for (Container enableTest = aComponent.getParent();
                 enableTest != null;
                 enableTest = enableTest.getParent())
            {
                if (!(enableTest.isEnabled() || enableTest.isLightweight())) {
                    return false;
                }
                if (enableTest instanceof Window) {
                    break;
                }
            }
        }

        return true;
    }
    private void writeObject(ObjectOutputStream out) throws IOException {
	out.defaultWriteObject();

	if (delegate instanceof Serializable) {
	    out.writeObject(delegate);
	} else {
	    out.writeObject(null);
	}
    }
    private void readObject(ObjectInputStream in)
        throws IOException, ClassNotFoundException
    {
	in.defaultReadObject();
	delegate = (FocusTraversalPolicy)in.readObject();
    }
}
