/*
 * @(#)SpringLayout.java	1.3 01/04/01
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */
package javax.swing;

import java.awt.*;
import java.util.*;

/**
 * A <code>SpringLayout</code> lays out the children of an associated container
 * according to a set of constraints which are established between the edges
 * of the components. Each edge is considered as an unbounded horizontal
 * or vertical line and is represented by a single integer specifying the distance to the
 * north-west corner of the container. Each constraint is a spring which has a minimum,
 * preferred and maximum value. When each of these values is the same, the spring may be
 * thought of as an inflexible "strut", fixing the distance between two edges
 * absolutely. By default, the components being laid out are provided with
 * two default springs: one linking their north and south edges by the
 * heights of the component's size characteristics (as returned by the
 * component's <code>getMinimumSize()</code>, <code>getPreferredSize()</code> and
 * <code>getMaximumSize()</code>) methods); the other linking the east and west
 * edges by their widths.
 * <p>
 * In a <code>SpringLayout</code>, the position of an edge is dependent on
 * the position of just one other edge. If a constraint is subsequently added
 * to create a new binding for this edge, the previous binding is discarded
 * and the edge remains dependent on a single edge. The springs that link
 * the edges to each other may be submitted by the user and may be any
 * function of the springs representing the minimum, preferred and maximum sizes
 * of the other components in the container. Springs should only be attached
 * between edges of the container and its immediate children; the behavior
 * of the <code>SpringLayout</code> when presented with constraints linking
 * the edges of components from different containers (either internal or
 * external) is currently undefined.
 * <p>
 * The <code>SpringLayout</code> has a number of advantages over the other layout managers
 * in the important domain of design tools. One significant advantage of the
 * <code>SpringLayout</code> is that the springs that control the dynamic behavior
 * of a layout are more easily visualized than the somewhat more abstract properties
 * of conventional layout managers. Secondly, it is much easier to incrementally
 * add dynamic behavior to an existing absolute layout using a
 * spring layout as the default behavior of the <code>SpringLayout</code>
 * class is very close to that of the <code>null</code> layout manager.
 * <p>
 * Despite its simplicity, it is possible to use the <code>SpringLayout</code> to
 * emulate the behavior of most of the other layout managers by adding the
 * appropriate constraints to the edges of the components (some features,
 * like the line-breaking feature of a <code>FlowLayout</code>, however cannot
 * be emulated this way and can only be accomodated in a <code>SpringLayout</code> by
 * creating a special purpose subclass of the <code>Spring</code> class).
 * The <code>SpringLayout</code> also provides a way to solve many of the difficult layout
 * problems which cannot be solved using the established technique of nesting combinations
 * of <code>Box</code>es. That said, the <code>SpringLayout</code> honors the
 * <code>LayoutManager2</code> contract correctly and so can be nested with
 * other layout mangers - a technique which is often preferable to
 * creating the constraints implied by the other layout managers.
 * <p>
 * The asymptotic complexity of the layout operation of a <code>SpringLayout</code>
 * is linear in the number of constraints (and/or components).
 *
 * @version 1.3 04/01/01
 * @author 	Philip Milne
 * @author 	Joe Winchester
 * @since       1.4
 */
public class SpringLayout implements LayoutManager2 {
    private Map componentConstraints = new HashMap();

    private Spring cyclicReference = Spring.constant(Spring.UNSET);
    private Set cyclicSprings;
    private Set acyclicSprings;


    /**
     * The north edge.
     */
    public static final String NORTH  = "North";

    /**
     * The south edge.
     */
    public static final String SOUTH  = "South";

    /**
     * The east edge.
     */
    public static final String EAST   = "East";

    /**
     * The west edge.
     */
    public static final String WEST   = "West";

    private abstract static class LocationSpring extends Spring.AbstractSpring {
        protected Component c;

        public int getMinimumValue() {
            return getPreferredValue();
        }

        public int getMaximumValue() {
            return getPreferredValue();
        }
    }

    private static class XSpring extends LocationSpring {

        public XSpring(Component c) {
            this.c = c;
        }

        public int getPreferredValue() {
            return c.getX();
        }
    }

    private static class YSpring extends LocationSpring {

        public YSpring(Component c) {
            this.c = c;
        }

        public int getPreferredValue() {
            return c.getY();
        }
    }

    private static class WidthSpring extends Spring.AbstractSpring {
        private Component c;

        public WidthSpring(Component c) {
            this.c = c;
        }

        public int getMinimumValue() {
            return c.getMinimumSize().width;
        }

        public int getPreferredValue() {
            return c.getPreferredSize().width;
        }

        public int getMaximumValue() {
            // We will be doing arithmetic with the results of this call,
            // so if a returned value is Integer.MAX_VALUE we will get
            // arithmetic overflow. Truncate such values.
            return Math.min(Short.MAX_VALUE, c.getMaximumSize().width);
        }
    }

     private static class HeightSpring extends Spring.AbstractSpring {
        private Component c;

        public HeightSpring(Component c) {
            this.c = c;
        }

        public int getMinimumValue() {
            return c.getMinimumSize().height;
        }

        public int getPreferredValue() {
            return c.getPreferredSize().height;
        }

        public int getMaximumValue() {
            return Math.min(Short.MAX_VALUE, c.getMaximumSize().height);
        }
    }

    /**
     * A <code>Constraints</code> object is used to hold the
     * constraints that govern the way a component's size
     * will change in a <code>SpringLayout</code>.
     * The <code>Constraints</code> object is basically
     * a <code>Rectangle</code>, whose <code>x</code>, <code>y</code>,
     * <code>width</code> and <code>height</code> variables are
     * <code>Spring</code>s instead of integers.
     * The constraints object also provides read/write access to
     * the edges of such a rectangle so that it may be manipulated
     * as a set of four edges instead of location and size variables.
     * <p>
     * The constraints object is maintained in a consistent state
     * so that the following formulae hold when the "setter"
     * methods of the <code>Constraints</code> object are called
     * externally:
     * <pre>
     *       west = x
     *      north = y
     *       east = x + width
     *      south = y + height
     * </pre>
     * The arithmetic operations in these formulae refer to the
     * operators in the <code>Spring</code> class. I.e. "+" here
     * refers to the "sum" method in the <code>Spring</code> class.
     * <p>
     * So if, for example, the "east" property of this constraints
     * object is set to some (spring) value and both the "x" and "width"
     * properties are already defined, a definition needs to change so
     * that the above rules will still hold. In general this class
     * assumes the "width" and "height" properties should be left
     * unaltered wherever possible and will favor changes
     * to the locations of its edges over changes to its dimensions.
     * In this example we set the "x" value to "east - width".
     */
   public static class Constraints {
       private Spring x;
       private Spring y;
       private Spring width;
       private Spring height;
       private Spring east;
       private Spring south;

       /**
        * Creates an empty <code>Constraints</code> object.
        */
       public Constraints() {
           this(null, null, null, null);
       }

       /**
        * Creates a <code>Constraints</code> object with the springs for its
        * ordinate and abscissa taking the values <code>x</code> and <code>y</code>.
        * Height and width springs have the value of null.
        *
        * @param x  the spring for the ordinate
        * @param y  the spring for the abscissa
        */
       public Constraints(Spring x, Spring y) {
           this(x, y, null, null);
       }

       /**
        * Creates a <code>Constraints</code> object with the springs for its
        * ordinate, abscissa, width and height springs taking the values
        * <code>x</code>, <code>y</code>, <code>width</code> and <code>height</code>.
        * Note that if the SpringLayout class encounters null values in the
        * constraints object of a given component it replaces them with
        * suitable defaults.
        *
        * @param x  the spring for the ordinate
        * @param y  the spring for the abscissa
        * @param width  the spring for the width
        * @param height  the spring for the height
        */
       public Constraints(Spring x, Spring y, Spring width, Spring height) {
           this.x = x;
           this.y = y;
           this.width = width;
           this.height = height;
       }

       /**
        * Sets the spring controlling the ordinate of a component's location
        * (the "x" value).
        *
        * @param x the spring controlling the ordinate of a component's location
        *
        * @see #getX
        * @see SpringLayout.Constraints
        */
       public void setX(Spring x) {
           this.x = x;
       }

       /**
        * Returns the spring controlling the ordinate of a component's location
        * (the "x" value).
        *
        * @return the spring controlling the ordinate of a component's location
        *
        * @see #setX
        * @see SpringLayout.Constraints
        */
       public Spring getX() {
           if (x == null && east != null && width != null) {
               x = Spring.difference(east, width);
           }
           return x;
       }

       /**
        * Sets the spring controlling the abscissa of a component's location
        * (the "y" value).
        *
        * @param the spring controlling the abscissa of a component's location
        *
        * @see #getY
        * @see SpringLayout.Constraints
        */
       public void setY(Spring y) {
           this.y = y;
       }

       /**
        * Returns the spring controlling the abscissa of a component's location
        * (the "y" value).
        *
        * @return the spring controlling the abscissa of a component's location
        *
        * @see #setY
        * @see SpringLayout.Constraints
        */
       public Spring getY() {
           if (y == null && south != null && height != null) {
               y = Spring.difference(south, height);
           }
            return y;
       }

       /**
        * Sets the spring controlling the width of this constraints object.
        *
        * @param width the spring controlling the width of this constraints object
        *
        * @see #getWidth
        * @see SpringLayout.Constraints
        */
       public void setWidth(Spring width) {
           this.width = width;
       }

       /**
        * Returns the spring controlling the width of this constraints object.
        *
        * @return the spring controlling the width of this constraints object
        *
        * @see #setWidth
        * @see SpringLayout.Constraints
        */
       public Spring getWidth() {
           if (width == null && x != null && east != null) {
               width = Spring.difference(east, x);
           }
           return width;
       }

       /**
        * Sets the spring controlling the height of this constraints object.
        *
        * @param height the spring controlling the height of this constraints object
        *
        * @see #getHeight
        * @see SpringLayout.Constraints
        */
       public void setHeight(Spring height) {
           this.height = height;
       }

       /**
        * Returns the spring controlling the height of this constraints object.
        *
        * @return the spring controlling the height of this constraints object
        *
        * @see #setHeight
        * @see SpringLayout.Constraints
        */
       public Spring getHeight() {
           if (height == null && y != null && south != null) {
               height = Spring.difference(south, y);
           }
           return height;
       }

       private void setEast(Spring east) {
           this.east = east;
           if (width != null) {
               x = null;
           }
       }

       private Spring getEast() {
           if (east == null && x != null && width != null) {
               east = Spring.sum(x, width);
           }
           return east;
       }

       private void setSouth(Spring south) {
           this.south = south;
           if (height != null) {
               y = null;
           }
       }

       private Spring getSouth() {
           if (south == null && y != null && height != null) {
               south = Spring.sum(y, height);
           }
           return south;
       }

       /**
        * Sets the spring controlling the specified edge of this constraints object.
        * Edges must be in the set of static values:
        * {<code>NORTH</code>, <code>SOUTH</code>, <code>EAST</code>, <code>WEST</code>}.
        *
        * @param edgeName the edge of this constraints object that is to be set
        * @param s the spring controlling the specified edge of this constraints object
        *
        * @see #getConstraint
        * @see SpringLayout.Constraints
        */
       public void setConstraint(String edgeName, Spring s) {
           edgeName = edgeName.intern();
           if (edgeName == "West") {
               setX(s);
           }
           else if (edgeName == "North") {
               setY(s);
           }
           else if (edgeName == "East") {
               setEast(s);
           }
           else if (edgeName == "South") {
               setSouth(s);
           }
       }

       /**
        * Returns the spring controlling the specified edge of this constraints object.
        * Edges must be in the set of static values:
        * {<code>NORTH</code>, <code>SOUTH</code>, <code>EAST</code>, <code>WEST</code>}.
        *
         *
        * @param edgeName the edge of this constraints object that is to be returned
        *
        * @return the spring controlling the specified edge of this constraints object
        *
        * @see #setConstraint
        * @see SpringLayout.Constraints
        */
       public Spring getConstraint(String edgeName) {
           edgeName = edgeName.intern();
           return (edgeName == "West")  ? getX() :
                  (edgeName == "North") ? getY() :
                  (edgeName == "East")  ? getEast() :
                  (edgeName == "South") ? getSouth() :
                  null;
       }

       /*pp*/ void reset() {
           if (x != null) x.setValue(Spring.UNSET);
           if (y != null) y.setValue(Spring.UNSET);
           if (width != null) width.setValue(Spring.UNSET);
           if (height != null) height.setValue(Spring.UNSET);
           if (east != null) east.setValue(Spring.UNSET);
           if (south != null) south.setValue(Spring.UNSET);
       }
   }

   private static class SpringProxy extends Spring {
       private String edgeName;
       private Component c;
       private SpringLayout l;

       public SpringProxy(String edgeName, Component c, SpringLayout l) {
           this.edgeName = edgeName;
           this.c = c;
           this.l = l;
       }

       public Spring getConstraint() {
           return l.getConstraint2(edgeName, c);
       }

       public int getMinimumValue() {
           return getConstraint().getMinimumValue();
       }

       public int getPreferredValue() {
           return getConstraint().getPreferredValue();
       }

       public int getMaximumValue() {
           return getConstraint().getMaximumValue();
       }

       public int getValue() {
           return getConstraint().getValue();
       }

       public void setValue(int size) {
           getConstraint().setValue(size);
       }

       /*pp*/ boolean isCyclic(SpringLayout l) {
           return l.isCyclic(getConstraint());
       }

       public String toString() {
           return "SpringProxy for " + edgeName + " edge of " + c.getName() + ".";
       }
    }

    /**
     * Constructs a new <code>SpringLayout</code>.
     */
    public SpringLayout() {}

    /*pp*/ boolean isCyclic(Spring s) {
        if (cyclicSprings.contains(s)) {
            return true;
        }
        if (acyclicSprings.contains(s)) {
            return false;
        }
        cyclicSprings.add(s);
        boolean result = s.isCyclic(this);
        if (!result) {
            acyclicSprings.add(s);
            cyclicSprings.remove(s);
        }
        else {
            System.err.println(s + " is cyclic. ");
        }
        return result;
    }

    // LayoutManager methods.

    public void addLayoutComponent(String name, Component c) {}

    public void removeLayoutComponent(Component c) {
        componentConstraints.remove(c);
    }

    private static Dimension addInsets(int width, int height, Container p) {
        Insets i = p.getInsets();
        return new Dimension(width + i.left + i.right, height + i.top + i.bottom);
    }

    public Dimension minimumLayoutSize(Container p) {
        setParent(p);
        return addInsets(getConstraint3("East", p).getMinimumValue(),
                         getConstraint3("South", p).getMinimumValue(), p);
    }

    public Dimension preferredLayoutSize(Container p) {
        setParent(p);
        return addInsets(getConstraint3("East", p).getPreferredValue(),
                         getConstraint3("South", p).getPreferredValue(), p);
    }

    // LayoutManager2 methods.

    public Dimension maximumLayoutSize(Container p) {
        setParent(p);
        return addInsets(getConstraint3("East", p).getMaximumValue(),
                         getConstraint3("South", p).getMaximumValue(), p);
    }

    /**
     * If the constraints object is an instance of SpringLayout.Constraints
     * register the constraints of the component with the layout manager.
     * <p>
     * @param   c the component being added
     * @param   constraints the component's constraints
     *
     * @see SpringLayout.Constraints
     */
    public void addLayoutComponent(Component component, Object constraints) {
        if (constraints instanceof Constraints) {
            putConstraints(component, (Constraints)constraints);
        }
    }

    public float getLayoutAlignmentX(Container p) {
        return 0;
    }

    public float getLayoutAlignmentY(Container p) {
        return 0;
    }

    public void invalidateLayout(Container c) {
        // System.out.println("invalidating ... ");
    }

    // End of LayoutManger2 methods

    private void setParent(Container p) {
        resetCylceStatuses();
        // Leading edges of the parent are axes, and should have the value of zero.
        // Set the bindings for the trailing edges, so as to define the default:
        // min, pref, and max values.
        Constraints constraints = getConstraints(p);
        constraints.setX(Spring.constant(0));
        constraints.setY(Spring.constant(0));
        constraints.setWidth(Spring.constant(0, 0, Integer.MAX_VALUE));
        constraints.setHeight(Spring.constant(0, 0, Integer.MAX_VALUE));
    }

    private void resetCylceStatuses() {
        cyclicSprings = new HashSet();
        acyclicSprings = new HashSet();
    }

   /**
     * Link edge <code>e1</code> of component <code>c1</code> to
     * edge <code>e2</code> of component <code>c2</code>. This
     * constraint will cause the following assignment:
     * <pre>
     *     value(e1, c1) = value(e2, c2) + pref
     * </pre>
     * to take place during all subsequent layout operations.
     * <p>
     * @param   e1 the edge of the dependent
     * @param   c1 the component of the dependent
     * @param   pref the fixed distance between dependent and anchor
     * @param   e2 the edge of the anchor
     * @param   c2 the component of the anchor
     *
     * @see #putConstraint(String, Component, Spring, String, Component)
     */
    public void putConstraint(String e1, Component c1, int pref, String e2, Component c2) {
        putConstraint(e1, c1, Spring.constant(pref), e2, c2);
    }

    /**
     * Link edge <code>e1</code> of component <code>c1</code> to
     * edge <code>e2</code> of component <code>c2</code>. As edge
     * <code>(e2, c2)</code> changes value, edge <code>(e1, c1)</code> will
     * be calculated by taking the (spring) sum of <code>(e2, c2)</code>
     * and <code>s</code>. Edges must be in the set of static values:
     * {<code>NORTH</code>, <code>SOUTH</code>, <code>EAST</code>, <code>WEST</code>}.
     * <p>
     * @param   e1 the edge of the dependent
     * @param   c1 the component of the dependent
     * @param   s the spring linking dependent and anchor
     * @param   e2 the edge of the anchor
     * @param   c2 the component of the anchor
     *
     * @see #putConstraint(String, Component, int, String, Component)
     */
    public void putConstraint(String e1, Component c1, Spring s, String e2, Component c2) {
        putConstraint(e1, c1, Spring.sum(s, getConstraint(e2, c2)));
    }

    private void putConstraint(String e, Component c, Spring s) {
        if (s != null) {
            getConstraints(c).setConstraint(e, s);
        }
     }

    private Constraints applyDefaults(Component c, Constraints constraints) {
        if (constraints == null) {
             constraints = new Constraints();
        }
        if (constraints.getX() == null) {
            constraints.setX(new XSpring(c));
        }
        if (constraints.getY() == null) {
            constraints.setY(new YSpring(c));
        }
        if (constraints.getWidth() == null) {
            constraints.setWidth(new WidthSpring(c));
        }
        if (constraints.getHeight() == null) {
            constraints.setHeight(new HeightSpring(c));
        }
        return constraints;
    }

    private void putConstraints(Component component, Constraints constraints) {
        componentConstraints.put(component, applyDefaults(component, constraints));
    }

    /**
     * Returns the constraints for the specified component.
     * Note that, unlike the <code>GridBagLayout</code>,
     * constraints are not cloned by this method. If no
     * constraint has been registered for this component
     * and the component is a <code>JComponent</code>, this
     * method checks the "constraints" client property of
     * the component. If this is also undefined, the method
     * returns a default constraints object in using the
     * current location of the component and its size
     * characteristics. Note that the size characteristics
     * are not frozen at the time this method is called;
     * instead this method returns a constraints object
     * whose charateristics track the charateristics
     * of the component as they change.
     *
     * @param       c the component whose constraints will be returned
     *
     * @return      the constraints for the specified component
     */
    public Constraints getConstraints(Component c) {
       Constraints result = (Constraints)componentConstraints.get(c);
       if (result == null) {
           if (c instanceof javax.swing.JComponent) {
                Object cp = ((javax.swing.JComponent)c).getClientProperty(SpringLayout.class);
                if (cp instanceof Constraints) {
                    return applyDefaults(c, (Constraints)cp);
                }
           }
           result = new Constraints();
           putConstraints(c, result);
       }
       return result;
    }

    private Spring getConstraint2(String e, Component c) {
       return getConstraints(c).getConstraint(e);
    }

    private Spring getConstraint3(String e, Component c) {
        Spring result = getConstraint2(e, c);
        return isCyclic(result) ? cyclicReference : result;
    }

    /**
     * Returns the spring controlling the distance between this edge of
     * the component from the top-left corner of its parent. This
     * method, instead of returning the current binding for this
     * edge, returns a proxy which will track the characteristics
     * of this edge even if the edge is subsequently rebound.
     * Proxies are intended to be used in builder envonments
     * where it is useful to allow the user to define the
     * constraints for a layout in any order. Proxies do, however,
     * provide the means to create cyclic dependencies amongst
     * the constraints of a layout. Such cycles are detected
     * internally by the <code>SpringLayout</code> class so that
     * the layout operation always terminates.
     *
     * @return the spring controlling the distance between this edge of
     * the component from the top-left corner of its parent
     */
    public Spring getConstraint(String edgeName, Component c) {
        // The interning here is unnecessary; it was added for efficiency.
        edgeName = edgeName.intern();
        return new SpringProxy(edgeName, c, this);
    }

    public void layoutContainer(Container p) {
        setParent(p);

        int n = p.getComponentCount();
        getConstraints(p).reset();
        for (int i = 0 ; i < n ; i++) {
            getConstraints(p.getComponent(i)).reset();
        }

        Insets insets = p.getInsets();

        /*
        getConstraints(p).getWidth().setValue(p.getWidth() - insets.left - insets.right);
        getConstraints(p).getHeight().setValue(p.getHeight() - insets.top - insets.bottom);
        */

        getConstraints(p).getEast().setValue(p.getWidth() - insets.left - insets.right);
        getConstraints(p).getSouth().setValue(p.getHeight() - insets.top - insets.bottom);

	for (int i = 0 ; i < n ; i++) {
	    Component c = p.getComponent(i);
            int x = getConstraint3("West", c).getValue();
            int y = getConstraint3("North", c).getValue();
            int width = getConstraint3("East", c).getValue() - x;
            int height = getConstraint3("South", c).getValue() - y;
            c.setBounds(insets.left + x, insets.top + y, width, height);
            // System.out.println(c.getName() + ": "  + c.getBounds());
	}
    }
}
