/*
 * @(#)MetalProgressBarUI.java	1.21 01/03/05
 *
 * Copyright 1998-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package javax.swing.plaf.metal;

import javax.swing.*;
import javax.swing.plaf.*;
import javax.swing.plaf.basic.*;
import java.awt.*;

/**
 * The Metal implementation of ProgressBarUI.
 * <p>
 * <strong>Warning:</strong>
 * Serialized objects of this class will not be compatible with
 * future Swing releases. The current serialization support is
 * appropriate for short term storage or RMI between applications running
 * the same version of Swing.  As of 1.4, support for long term storage
 * of all JavaBeans<sup><font size="-2">TM</font></sup>
 * has been added to the <code>java.beans</code> package.
 * Please see {@link java.beans.XMLEncoder}.
 *
 * @version 1.21 03/05/01
 * @author Michael C. Albers
 */
public class MetalProgressBarUI extends BasicProgressBarUI {

    private Rectangle innards;
    private Rectangle box;

    public static ComponentUI createUI(JComponent c) {
	return new MetalProgressBarUI();
    }

    /**
     * Draws a bit of special highlighting on the progress bar.
     * The core painting is deferred to the BasicProgressBar's
     * <code>paintDeterminate</code> method.
     */ 
    public void paintDeterminate(Graphics g, JComponent c) {
	super.paintDeterminate(g,c);

	if (progressBar.isBorderPainted()) {
	    BoundedRangeModel model = progressBar.getModel();
	    int barRectX = 0;
	    int barRectY = 0;
	    int barRectWidth = progressBar.getWidth();
	    int barRectHeight = progressBar.getHeight();
	    Insets b = progressBar.getInsets(); // area for border
	    barRectX += b.left;
	    barRectY += b.top;
	    barRectWidth -= (b.right + barRectX);
	    barRectHeight -= (b.bottom + barRectY);
	    int amountFull = getAmountFull(b, barRectWidth, barRectHeight);
	    
	    if (progressBar.getOrientation() == JProgressBar.HORIZONTAL) {
                int fillX = barRectX;
                if( !MetalUtils.isLeftToRight(c) ) {
                    fillX += barRectWidth - amountFull;
                }
		// Highlighting
		//     over the unfilled portion
		//     well, draw all the way across; let others draw over it
		g.setColor(MetalLookAndFeel.getControlShadow());
		g.drawLine(barRectX,barRectY,barRectX+barRectWidth-1,barRectY);
		
		//     line on left
		if (fillX == barRectX && amountFull > 0) {
                    // filled area is touching left edge of bar
		    g.setColor(MetalLookAndFeel.getPrimaryControlDarkShadow());
		} else {
                    // filled area is not touching left edge of bar
		    g.setColor(MetalLookAndFeel.getControlShadow());
		}
		g.drawLine(barRectX, barRectY,
                           barRectX, barRectY+barRectHeight-1);
		
		//     highlight over the filled portion
                if( amountFull > 0 ) {
                    g.setColor(MetalLookAndFeel.getPrimaryControlDarkShadow());
                    g.drawLine(fillX, barRectY, fillX+amountFull-1, barRectY);
                }
	    } else { // VERTICAL
		// Highlighting
		//     left of the unfilled portion
		//     well, draw all the way down; let others draw over it
		g.setColor(MetalLookAndFeel.getControlShadow());
		g.drawLine(barRectX, barRectY,
                           barRectX, barRectY+barRectHeight-1);
		
		//     line on bottom
		if ( amountFull <= 0 ) { // haven't begun
		    g.setColor(MetalLookAndFeel.getControlShadow());
		} else { // Some portion of bar is filled
		    g.setColor(MetalLookAndFeel.getPrimaryControlDarkShadow());
		}
		g.drawLine(barRectX, barRectY+barRectHeight-1,
			   barRectX+barRectWidth-1, barRectY+barRectHeight-1);
		
		//     left of the filled portion
		//     pick up color from the "line on bottom" above
		g.drawLine(barRectX, barRectY+barRectHeight-1,
			   barRectX, barRectY+barRectHeight-amountFull);
	    }
	}
    }

    /**
     * Draws a bit of special highlighting on the progress bar
     * and bouncing box.
     * The core painting is deferred to the BasicProgressBar's
     * <code>paintIndeterminate</code> method.
     */ 
    public void paintIndeterminate(Graphics g, JComponent c) {
        super.paintIndeterminate(g, c);

        if (!progressBar.isBorderPainted()) { //do nothing
            return;
        }

        innards = SwingUtilities.calculateInnerArea(progressBar, innards);
        box = getBox(box);

        //Draw a highlight along the top and left of the channel.
        g.setColor(MetalLookAndFeel.getControlShadow());

        g.drawLine(innards.x, innards.y,                       //top
                   innards.x + innards.width - 1, innards.y);
        g.drawLine(innards.x, innards.y,                       //left
                   innards.x, innards.y + innards.height - 1);

        if (box == null) {
            return;
        }

        g.setColor(MetalLookAndFeel.getPrimaryControlDarkShadow());

        if (progressBar.getOrientation() == JProgressBar.HORIZONTAL) {
            //Draw a highlight to emphasize the top and (if we're up against
            //the starting edge) starting edge of the bouncing box.

            g.drawLine(box.x, innards.y,                       //top
                       box.x + box.width - 1, innards.y);

            if (MetalUtils.isLeftToRight(c)) {
                if (box.x == innards.x) { //draw the left highlight
                    g.drawLine(box.x, innards.y,
                               box.x, box.y + box.height - 1);
                }
            } else { // right-to-left
                int maxBoxX = innards.x + innards.width - box.width;
                if (box.x == maxBoxX) {
                    //draw the right highlight
                    int rightmostX = box.x + box.width - 1;
                    g.drawLine(rightmostX, innards.y,
                               rightmostX, box.y + box.height - 1);
                }
            } 
        } else { //VERTICAL
            //Draw a highlight to emphasize the left and (if we're up against
            //the starting edge) bottom edge of the bouncing box.

            int maxY = innards.y + innards.height - 1;

            g.drawLine(box.x, box.y,                       //left
                       box.x, box.y + box.height - 1);

            int maxBoxY = innards.y + innards.height - box.height;
            if (box.y == maxBoxY) { //draw the bottom highlight
                g.drawLine(innards.x,
                           maxY,
                           innards.x + innards.width - 1,
                           maxY);
            } 
        }
    }
}
