/*
 * @(#)InternationalFormatter.java	1.6 01/03/05
 *
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
package javax.swing.text;

import java.awt.event.ActionEvent;
import java.io.*;
import java.text.*;
import java.util.*;
import javax.swing.*;
import javax.swing.text.*;

/**
 * <code>InternationalFormatter</code> extends <code>DefaultFormatter</code>,
 * using an instance of <code>java.text.Format</code> to handle the
 * conversion to a String, and the conversion from a String.
 * <p>
 * If <code>getAllowsInvalid()</code> is false, this will ask the
 * <code>Format</code> to format the current text on every edit.
 * <p>
 * You can specify a minimum and maximum value by way of the 
 * <code>setMinimum</code> and <code>setMaximum</code> methods. In order
 * for this to work the values returned from <code>stringToValue</code> must be
 * comparable to the min/max values by way of the <code>Comparable</code>
 * interface.
 * <p>
 * Be careful how you configure the <code>Format</code> and the
 * <code>InternationalFormatter</code>, as it is possible to create a
 * situation where certain values can not be input. Consider the date
 * format 'M/d/yy', an <code>InternationalFormatter</code> that is always
 * valid (<code>setAllowsInvalid(false)</code>), is in overwrite mode
 * (<code>setOverwriteMode(true)</code>) and the date 7/1/99. In this
 * case the user will not be able to enter a two digit month or day of
 * month. To avoid this, the format should be 'MM/dd/yy'.
 * <p>
 * If <code>InternationalFormatter</code> is configured to only allow valid
 * values (<code>setAllowsInvalid(false)</code>), every valid edit will result
 * in the text of the <code>JFormattedTextField</code> being completely reset
 * from the <code>Format</code>.
 * The cursor position will also be adjusted as literal characters are
 * added/removed from the resulting String.
 * <p>
 * <code>InternationalFormatter</code>'s behavior of
 * <code>stringToValue</code> is  slightly different than that of
 * <code>DefaultTextFormatter</code>, it does the following:
 * <ol>
 *   <li><code>parseObject</code> is invoked on the <code>Format</code>
 *       specified by <code>setFormat</code>
 *   <li>If a Class has been set for the values (<code>setValueClass</code>),
 *       supers implementation is invoked to convert the value returned
 *       from <code>parseObject</code> to the appropriate class.
 *   <li>If a <code>ParseException</code> has not been thrown, and the value
 *       is outside the min/max a <code>ParseException</code> is thrown.
 *   <li>The value is returned.
 * </ol>
 * <code>InternationalFormatter</code> implements <code>stringToValue</code>
 * in this manner so that you can specify an alternate Class than
 * <code>Format</code> may return.
 * <p>
 * <strong>Warning:</strong>
 * Serialized objects of this class will not be compatible with
 * future Swing releases. The current serialization support is
 * appropriate for short term storage or RMI between applications running
 * the same version of Swing.  As of 1.4, support for long term storage
 * of all JavaBeans<sup><font size="-2">TM</font></sup>
 * has been added to the <code>java.beans</code> package.
 * Please see {@link java.beans.XMLEncoder}.
 *
 * @see java.text.Format
 * @see java.lang.Comparable
 *
 * @version 1.6 03/05/01
 * @since 1.4
 */
public class InternationalFormatter extends DefaultFormatter {
    /**
     * Used by <code>getFields</code>.
     */
    private static final Format.Field[] EMPTY_FIELD_ARRAY =new Format.Field[0];

    /**
     * Object used to handle the conversion.
     */
    private Format format;
    /**
     * Can be used to impose a maximum value.
     */
    private Comparable max;
    /**
     * Can be used to impose a minimum value.
     */
    private Comparable min;


    /**
     * Creates an <code>InternationalFormatter</code> with no
     * <code>Format</code> specified.
     */
    public InternationalFormatter() {
        setOverwriteMode(false);
    }

    /**
     * Creates an <code>InternationalFormatter</code> with the specified
     * <code>Format</code> instance.
     *
     * @param Format instance used for converting from/to Strings
     */
    public InternationalFormatter(Format format) {
        this();
        setFormat(format);
    }

    /**
     * Sets the format that dictates the legal values that can be edited
     * and displayed.
     *
     * @param format <code>Format</code> instance used for converting
     * from/to Strings
     */
    public void setFormat(Format format) {
        this.format = format;
    }

    /**
     * Returns the format that dictates the legal values that can be edited
     * and displayed.
     *
     * @return Format instance used for converting from/to Strings
     */
    public Format getFormat() {
        return format;
    }

    /**
     * Sets the minimum permissible value. If the <code>valueClass</code> has
     * not been specified, and <code>minimum</code> is non null, the
     * <code>valueClass</code> will be set to that of the class of
     * <code>minimum</code>.
     *
     * @param minimum Minimum legal value that can be input
     * @see #setValueClass
     */
    public void setMinimum(Comparable minimum) {
        if (getValueClass() == null && minimum != null) {
            setValueClass(minimum.getClass());
        }
        min = minimum;
    }

    /**
     * Returns the minimum permissible value.
     *
     * @return Minimum legal value that can be input
     */
    public Comparable getMinimum() {
        return min;
    }

    /**
     * Sets the maximum permissible value. If the <code>valueClass</code> has
     * not been specified, and <code>max</code> is non null, the
     * <code>valueClass</code> will be set to that of the class of
     * <code>max</code>.
     *
     * @param maximum Maximum legal value that can be input
     * @see #setValueClass
     */
    public void setMaximum(Comparable max) {
        if (getValueClass() == null && max != null) {
            setValueClass(max.getClass());
        }
        this.max = max;
    }

    /**
     * Returns the maximum permissible value.
     *
     * @return Maximum legal value that can be input
     */
    public Comparable getMaximum() {
        return max;
    }

    /**
     * Returns a String representation of the Object <code>value</code>.
     * This invokes <code>format</code> on the current <code>Format</code>.
     *
     * @throws ParseException if there is an error in the conversion
     * @param value Value to convert
     * @return String representation of value
     */
    public String valueToString(Object value) throws ParseException {
        if (value == null) {
            return "";
        }
        Format f = getFormat();

        if (f == null) {
            return value.toString();
        }
        return f.format(value);
    }

    /**
     * Returns the <code>Object</code> representation of the
     * <code>String</code> <code>text</code>.
     *
     * @param text <code>String</code> to convert
     * @return <code>Object</code> representation of text
     * @throws ParseException if there is an error in the conversion
     */
    public Object stringToValue(String text) throws ParseException {
        Object value = stringToValue(text, getFormat());

        if (value != null && getValueClass() != null) {
            value = super.stringToValue(value.toString());
        }
        try {
            if (!isValidValue(value, true)) {
                throw new ParseException("Value not within min/max range", 0);
            }
        } catch (ClassCastException cce) {
            throw new ParseException("Class cast exception comparing values: "
                                     + cce, 0);
        }
        return value;
    }

    /**
     * Returns the <code>Format.Field</code> constants associated with
     * the text at <code>offset</code>. If <code>offset</code> is not
     * a valid location into the current text, this will return an
     * empty array.
     *
     * @param offset offset into text to be examined
     * @return Format.Field constants associated with the text at the
     *         given position.
     */
    public Format.Field[] getFields(int offset) {
        Filter filter = (Filter)getFilter();

        if (filter != null) {
            Map attrs = filter.getAttributes(offset);

            if (attrs != null && attrs.size() > 0) {
                ArrayList al = new ArrayList();

                al.addAll(attrs.keySet());
                return (Format.Field[])al.toArray(EMPTY_FIELD_ARRAY);
            }
        }
        return EMPTY_FIELD_ARRAY;
    }

    /**
     * Invokes <code>parseObject</code> on <code>f</code>, returning
     * its value.
     */
    Object stringToValue(String text, Format f) throws ParseException {
        if (f == null) {
            return text;
        }
        return f.parseObject(text);
    }

    /**
     * Returns true if <code>value</code> is between the min/max.
     */
    boolean isValidValue(Object value, boolean wantsCCE) {
        Comparable min = getMinimum();

        try {
            if (min != null && min.compareTo(value) > 0) {
                return false;
            }
        } catch (ClassCastException cce) {
            if (wantsCCE) {
                throw cce;
            }
            return false;
        }

        Comparable max = getMaximum();
        try {
            if (max != null && max.compareTo(value) < 0) {
                return false;
            }
        } catch (ClassCastException cce) {
            if (wantsCCE) {
                throw cce;
            }
            return false;
        }
        return true;
    }

    /**
     * Creates the <code>Filter</code> instance that is used to enforce a
     * policy in the <code>JFormattedTextField</code>.
     */
    DefaultFormatter.Filter createFilter() {
        return new Filter();
    }


    /**
     * <code>Filter</code>'s behavior is dicatated by a
     * <code>AttributedCharacterIterator</code> that is obtained from
     * the <code>Format</code>. On every edit, assuming
     * allows invalid is false, the <code>Format</code> instance is invoked
     * with <code>formatToCharacterIterator</code>. A <code>BitSet</code> is
     * also kept upto date with the non-literal characters, that is
     * for every index in the <code>AttributedCharacterIterator</code> an
     * entry in the bit set is updated based on the return value from
     * <code>isLiteral(Map)</code>. <code>isLiteral(int)</code> then uses
     * this cached information.
     * <p>
     * If allowsInvalid is false, every edit results in resetting the complete
     * text of the JTextComponent.
     * <p>
     * InternationalFormatterFilter can also provide two actions suitable for
     * incrementing and decrementing. To enable this a subclass must
     * override <code>getSupportsIncrement</code> to return true, and
     * override <code>adjustValue</code> to handle the changing of the
     * value. If you want to support changing the value outside of
     * the valid FieldPositions, you will need to override
     * <code>canIncrement</code>.
     */
    class Filter extends DefaultFormatter.Filter {
        /**
         * A bit is set for every index identified in the
         * AttributedCharacterIterator that is not considered decoration.
         * This should only be used if validMask is true.
         */
        private BitSet literalMask;
        /**
         * Used to iterate over characters.
         */
        private transient AttributedCharacterIterator iterator;
        /**
         * True if the Format was able to convert the value to a String and
         * back.
         */
        private boolean validMask;
        /**
         * Current value being displayed.
         */
        private transient String string;
        /**
         * If true, DocumentFilter methods are unconditionally allowed,
         * and no checking is done on their values. This is used when
         * incrementing/decrementing via the actions.
         */
        private boolean ignoreDocumentMutate;


        /**
         * If <code>getSupportsIncrement</code> returns true, this returns
         * two Actions suitable for incrementing/decrementing the value.
         */
        Action[] getActions() {
            if (getSupportsIncrement()) {
                return new Action[] { new IncrementAction("increment", 1),
                                      new IncrementAction("decrement", -1) };
            }
            return null;
        }

        /**
         * Subclasses may want to override this. The default implementation
         * returns true indicating <code>string</code> can be inserted.
         * This should not check if actually inserting the string results in
         * a valid value, it should just do a quick pass to make sure
         * <code>string</code> contains valid characters for this type of
         * <code>Format</code>.
         */
        boolean isValidInsertString(int offset, String string) {
            return true;
        }

        /**
         * Returns a Set of the attribute identifiers at <code>index</code>.
         */
        Map getAttributes(int index) {
            if (isValidMask()) {
                AttributedCharacterIterator iterator = getIterator();

                if (index >= 0 && index <= iterator.getEndIndex()) {
                    iterator.setIndex(index);
                    return iterator.getAttributes();
                }
            }
            return null;
        }

        /**
         * Returns the start of the first run that contains the attribute
         * <code>id</code>. This will return <code>-1</code> if the attribute
         * can not be found.
         */
        int getAttributeStart(AttributedCharacterIterator.Attribute id) {
            if (isValidMask()) {
                AttributedCharacterIterator iterator = getIterator();

                iterator.first();
                while (iterator.current() != CharacterIterator.DONE) {
                    if (iterator.getAttribute(id) != null) {
                        return iterator.getIndex();
                    }
                    iterator.next();
                }
            }
            return -1;
        }

        /**
         * Returns the <code>AttributedCharacterIterator</code> used to
         * format the last value.
         */
        AttributedCharacterIterator getIterator() {
            return iterator;
        }

        /**
         * Updates the AttributedCharacterIterator and bitset, if necessary.
         */
        void updateMaskIfNecessary() {
            if (!getAllowsInvalid() && (getFormat() != null)) {
                if (!isValidMask()) {
                    updateMask();
                }
                else {
                    String newString = getFormattedTextField().getText();

                    if (!newString.equals(string)) {
                        updateMask();
                    }
                }
            }
        }

        /**
         * Updates the AttributedCharacterIterator by invoking
         * <code>formatToCharacterIterator</code> on the <code>Format</code>.
         * If this is successful,
         * <code>updateMask(AttributedCharacterIterator)</code>
         * is then invoked to update the internal bitmask.
         */
        void updateMask() {
            if (getFormat() != null) {
                Document doc = getFormattedTextField().getDocument();

                validMask = false;
                try {
                    string = doc.getText(0, doc.getLength());
                } catch (BadLocationException ble) {
                    string = null;
                }
                if (string != null) {
                    try {
                        Object value = stringToValue(string);
                        AttributedCharacterIterator iterator = getFormat().
                                  formatToCharacterIterator(value);

                        updateMask(iterator);
                    }
                    catch (ParseException pe) {}
                    catch (IllegalArgumentException iae) {}
                    catch (NullPointerException npe) {}
                }
            }
        }

        /**
         * Returns the number of literal characters before <code>index</code>.
         */
        int getLiteralCountTo(int index) {
            int lCount = 0;

            for (int counter = 0; counter < index; counter++) {
                if (isLiteral(counter)) {
                    lCount++;
                }
            }
            return lCount;
        }

        /**
         * Returns true if the character at index is a literal, that is
         * not editable.
         */
        boolean isLiteral(int index) {
            if (isValidMask() && index < string.length()) {
                return literalMask.get(index);
            }
            return false;
        }

        /**
         * Returns the literal character at index.
         */
        char getLiteral(int index) {
            if (isValidMask() && string != null && index < string.length()) {
                return string.charAt(index);
            }
            return (char)0;
        }

        /**
         * Returns true if the character at offset is navigatable too. This
         * is implemented in terms of <code>isLiteral</code>, subclasses
         * may wish to provide different behavior.
         */
        boolean isNavigatable(int offset) {
            return !isLiteral(offset);
        }

        //
        // Following are override to invoke updateMaskIfNecessary
        //
        int getNextVisualPositionFrom(JTextComponent text, int pos,
                                      Position.Bias bias, int direction,
                                      Position.Bias[] biasRet)
                                           throws BadLocationException {
            updateMaskIfNecessary();
            return super.getNextVisualPositionFrom(text, pos, bias, direction,
                                                   biasRet);
        }

        int getInitialVisualPosition() {
            updateMaskIfNecessary();
            return super.getInitialVisualPosition();
        }

        void updateValue(Object value) {
            super.updateValue(value);
            updateMaskIfNecessary();
        }

        /**
         * This will either reposition the cursor or invoke replace based
         * on the state of the InternationalFormatter.
         */
        void remove(DocumentFilter.FilterBypass fb, int offset,
                    int length) throws BadLocationException {
            if (ignoreDocumentMutate) {
                fb.remove(offset, length);
                return;
            }
            updateMaskIfNecessary();

            if (getAllowsInvalid()) {
                fb.remove(offset, length);
                updateValue();
            }
            else {
                int selStart = getFormattedTextField().getSelectionStart();
                int selEnd = getFormattedTextField().getSelectionEnd();
                int cursor = offset;

                if (length == 1) {
                    int direction = (selStart == offset) ? 1 : -1;

                    while (offset >= 0 && isLiteral(offset)) {
                        offset += direction;
                    }
                }
                if (offset != -1) {
                    replace(fb, offset, length, null, null, selStart,
                            cursor, -1);
                }
                else {
                    invalidEdit();
                }
            }
        }

        /**
         * Adjusts the insert position and String based on how the
         * InternationalFormatter is configured, and invokes replace.
         */
        void insertString(DocumentFilter.FilterBypass fb, int offset,
                String string, AttributeSet attr) throws BadLocationException {
            if (ignoreDocumentMutate) {
                fb.insertString(offset, string, attr);
                return;
            }
            updateMaskIfNecessary();

            if (string != null && string.length() > 0) {
                if (!getAllowsInvalid()) {
                    if (!isValidInsertString(offset, string)) {
                        string = null;
                    }
                    else if (getOverwriteMode()) {
                        // skip literals
                        string = getOverwriteString(string, offset);
                        replace(fb, offset, string.length(), string, null,
                                offset + string.length(),
                                offset + string.length(), 1);
                        return;
                    }
                    else {
                        // skip as many literals at start of string that match,
                        // then insert rest.
                        int skipIndex = 0;
                        int max = string.length();

                        for (int counter = 0; counter < max; counter++) {
                            if (isLiteral(offset) && getLiteral(offset) ==
                                string.charAt(counter)) {
                                skipIndex++;
                                offset++;
                            }
                            else {
                                break;
                            }
                        }
                        if (skipIndex == string.length()) {
                            repositionCursor(offset, 1);
                            return;
                        }
                        else if (skipIndex > 0) {
                            string = string.substring(skipIndex);
                        }
                    }
                    if (string != null) {
                        replace(fb, offset, 0, string, attr, offset,
                                offset + string.length(), 1);
                    }
                    else {
                        invalidEdit();
                    }
                }
                else if (getOverwriteMode()) {
                    int rLength = Math.min(string.length(), fb.getDocument().
                                           getLength() - offset);

                    if (rLength > 0) {
                        fb.remove(offset, rLength);
                    }
                    fb.insertString(offset, string, attr);
                    updateValue();
                    repositionCursor(offset + string.length(), 1);
                }
                else {
                    fb.insertString(offset, string, attr);
                    updateValue();
                }
            }
        }

        /**
         * Replaces <code>length</code> characters at <code>offset</code>
         * with <code>string</code>. The cursor is then adjusted as
         * necessary.
         */
        void replace(DocumentFilter.FilterBypass fb,
                     int offset, int length, String string,
                     AttributeSet attr, int startCursor,
                     int endCursor, int direction) throws
                                 BadLocationException {
            String result = getReplaceString(offset, length, string);

            try {
                String newValue = valueToString(stringToValue(result));
                int oldLCount = getLiteralCountTo(startCursor);

                fb.remove(0, fb.getDocument().getLength());
                fb.insertString(0, newValue, null);
                updateValue();

                int newLCount = getLiteralCountTo(endCursor);

                if (newLCount != endCursor) {
                    endCursor -= oldLCount;
                    for (int counter = 0; counter < endCursor; counter++) {
                        if (isLiteral(counter)) {
                            endCursor++;
                        }
                    }
                }
/*
                int newLCount = getLiteralCountTo(endCursor);

                if (oldLCount != newLCount && !getOverwriteMode()) {
                    int delta = newLCount - oldLCount;

                    if (direction * delta > 0) {
                        endCursor += delta;
                    }
                }
*/
                repositionCursor(endCursor, 1 /*direction*/);
            } catch (ParseException pe) {
                invalidEdit();
            }
        }

        /**
         * Returns the character from the mask that has been buffered
         * at <code>index</code>.
         */
        char getBufferedChar(int index) {
            if (isValidMask()) {
                if (string != null && index < string.length()) {
                    return string.charAt(index);
                }
            }
            return (char)0;
        }

        /**
         * Returns true if the current mask is valid.
         */
        boolean isValidMask() {
            return validMask;
        }

        /**
         * Returns true if <code>attributes</code> is null or empty.
         */
        boolean isLiteral(Map attributes) {
            return ((attributes == null) || attributes.size() == 0);
        }

        /**
         * Returns the String to insert at offset, skipping over literal
         * characters as necessary.
         */
        private String getOverwriteString(String string, int offset) {
            if (string != null) {
                StringBuffer sb = new StringBuffer();

                for (int iCounter = 0, iLength = string.length();
                     iCounter < iLength; iCounter++, offset++) {
                    char sChar = string.charAt(iCounter);

                    if (isLiteral(offset)) {
                        char l = getLiteral(offset);

                        if (l == sChar) {
                            sb.append(sChar);
                        }
                        else {
                            sb.append(l);
                            iCounter--;
                        }
                    }
                    else {
                        sb.append(sChar);
                    }
                }
                return sb.toString();
            }
            return string;
        }

        /**
         * Updates the interal bitset from <code>iterator</code>. This will
         * set <code>validMask</code> to true if <code>iterator</code> is
         * non-null.
         */
        private void updateMask(AttributedCharacterIterator iterator) {
            if (iterator != null) {
                validMask = true;
                this.iterator = iterator;

                // Update the literal mask
                if (literalMask == null) {
                    literalMask = new BitSet();
                }
                else {
                    for (int counter = literalMask.length() - 1; counter >= 0;
                         counter--) {
                        literalMask.clear(counter);
                    }
                }

                iterator.first();
                while (iterator.current() != CharacterIterator.DONE) {
                    Map attributes = iterator.getAttributes();
                    boolean set = isLiteral(attributes);
                    int start = iterator.getIndex();
                    int end = iterator.getRunLimit();

                    while (start < end) {
                        if (set) {
                            literalMask.set(start);
                        }
                        else {
                            literalMask.clear(start);
                        }
                        start++;
                    }
                    iterator.setIndex(start);
                }
            }
        }

        /**
         * Returns true if <code>attributes</code> is no empty.
         * Subclasses that wish to allow incrementing to happen outside of
         * the known fields will need to override this.
         */
        boolean canIncrement(Map attributes, int cursorPosition) {
            return (attributes != null && attributes.size() > 0);
        }

        /**
         * Selects the FieldPosition identified by <code>fp.getField()</code>.
         * If <code>fp</code> is null, this has no effect.
         */
        void selectField(AttributedCharacterIterator.Attribute field) {
            int start = getAttributeStart(field);

            if (start != -1) {
                AttributedCharacterIterator iterator = getIterator();

                iterator.setIndex(start);

                int end = iterator.getRunLimit(field);

                if (end != -1) {
                    getFormattedTextField().select(start, end);
                }
            }
            else {
                getFormattedTextField().selectAll();
            }
        }

        /**
         * Selects the fields identified by <code>attributes</code>.
         */
        void selectField(Map attributes) {
            if (attributes.size() > 0) {
                Iterator iterator = attributes.keySet().iterator();

                selectField((AttributedCharacterIterator.Attribute)iterator.
                            next());
            }
        }

        /**
         * Subclasses supporting incrementing must override this to handle
         * the actual incrementing. <code>value</code> is the current value,
         * <code>attributes</code> gives the field the cursor is in (may be
         * null depending upon <code>canIncrement</code>) and 
         * <code>direction</code> is the amount to increment by.
         */
        Object adjustValue(Object value, Map attributes, int direction) throws
                      BadLocationException, ParseException {
            return null;
        }

        /**
         * Returns false, indicating InternationalFormatter does not allow
         * incrementing of the value. Subclasses that wish to support 
         * incrementing/decrementing the value should override this and
         * return true. Subclasses should also override
         * <code>adjustValue</code>.
         */
        boolean getSupportsIncrement() {
            return false;
        }

        /**
         * Resets the value of the JFormattedTextField to be
         * <code>value</code>.
         */
        void resetValue(Object value) throws BadLocationException,
                                             ParseException {
            Document doc = getFormattedTextField().getDocument();
            String string = valueToString(value);

            try {
                ignoreDocumentMutate = true;
                doc.remove(0, doc.getLength());
                doc.insertString(0, string, null);
            } finally {
                ignoreDocumentMutate = false;
            }
            updateValue(value);
        }


        /**
         * Subclassed to update the internal representation of the mask after
         * the default read operation has completed.
         */
        private void readObject(ObjectInputStream s) 
            throws IOException, ClassNotFoundException {
            s.defaultReadObject();
            updateMaskIfNecessary();
        }

        /**
         * IncrementAction is used to increment the value by a certain amount.
         * It calls into <code>adjustValue</code> to handle the actual
         * incrementing of the value.
         */
        private class IncrementAction extends AbstractAction {
            private int direction;

            IncrementAction(String name, int direction) {
                super(name);
                this.direction = direction;
            }

            public void actionPerformed(ActionEvent ae) {
                if (getAllowsInvalid()) {
                    // This will work if the currently edited value is valid.
                    updateMask();
                }

                boolean validEdit = false;

                if (isValidMask()) {
                    int start = getFormattedTextField().getSelectionStart();

                    if (start != -1) {
                        AttributedCharacterIterator iterator = getIterator();

                        iterator.setIndex(start);

                        Map attributes = iterator.getAttributes();

                        if (canIncrement(attributes, start)) {
                            try {
                                Object value = stringToValue(
                                         getFormattedTextField().getText());

                                value = adjustValue(value, attributes,
                                                    direction);
                                if (value != null && isValidValue(value,
                                                                  false)) {
                                    resetValue(value);
                                    selectField(attributes);
                                    validEdit = true;
                                }
                            }
                            catch (ParseException pe) { }
                            catch (BadLocationException ble) { }
                        }
                    }
                }
                if (!validEdit) {
                    invalidEdit();
                }
            }
        }        
    }
}
