/*
 * @(#)AccessibleHTML.java	1.6 01/03/25
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */
package javax.swing.text.html;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import javax.swing.text.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.plaf.TextUI;
import java.util.*;

import javax.accessibility.*;
import java.text.BreakIterator;
import java.awt.geom.*;
import java.awt.image.ImageObserver;

/**
 * This class implements accessibility support for the
 * <code>HTMLEditorKit</code> class.  It provides an implementation of 
 * the Java Accessibility API appropriate to menu user-interface elements.
 * <p>
 * <strong>Warning:</strong>
 * Serialized objects of this class will not be compatible with
 * future Swing releases. The current serialization support is
 * appropriate for short term storage or RMI between applications running
 * the same version of Swing.  As of 1.4, support for long term storage
 * of all JavaBeans<sup><font size="-2">TM</font></sup>
 * has been added to the <code>java.beans</code> package.
 * Please see {@link java.beans.XMLEncoder}.
 *
 * @author  Lynn Monsanto
 * @version 1.6 03/25/01
 */
public class AccessibleHTML implements Accessible {

    private AccessibleContext accessibleContext;
    private JEditorPane editor;
    private Vector children = new Vector();

    /*
     * constructors
     */
    private AccessibleHTML() {}	// should not be called

    public AccessibleHTML(JEditorPane editor) {
	this.editor = editor;
    }

    /*
     * returns the AccessibleContext for the root view
     */
    public AccessibleContext getAccessibleContext() {
	if (accessibleContext == null) {
	    if (editor == null) {
		return null;
	    }
	    View rootView = editor.getUI().getRootView(editor);
	    accessibleContext = new AccessibleHTMLRootView(rootView, editor, 0);
	}
	return accessibleContext;
    }

    /*
     * the AccessibleContext for the root view
     */
    private class AccessibleHTMLRootView extends AccessibleHTMLContext
        implements DocumentListener {

	/*
	 * constructor
	 */
	private AccessibleHTMLRootView(View view, Accessible parent,
					 int indexInParent) {
	    super(view, parent, indexInParent);

	    // build a list of AccessibleContexts for the children
	    // of the root view
	    buildAccessibleChildrenList(view);

	    // register for document changes
	    editor.getDocument().addDocumentListener(this);
	}

	// document listener
	
	/**
	 * Gives notification that there was an insert into the document.  The 
	 * range given by the DocumentEvent bounds the freshly inserted region.
	 *
	 * @param e the document event
	 */
	public void insertUpdate(DocumentEvent e) {
	    // XXX update the children list
	}
	
	/**
	 * Gives notification that a portion of the document has been 
	 * removed.  The range is given in terms of what the view last
	 * saw (that is, before updating sticky positions).
	 *
	 * @param e the document event
	 */
	public void removeUpdate(DocumentEvent e) {
	    // XXX update the children list
	}
	
	/**
	 * Gives notification that an attribute or set of attributes changed.
	 *
	 * @param e the document event
	 */
	public void changedUpdate(DocumentEvent e) {
	    // XXX update the children list
	}

	// ----- AccessibleContext -----

	/*
	 * returns the AccessibleContext for this Accessible
	 */
	public AccessibleContext getAccessibleContext() {
	    return this;
	}
	
	/*
	 * returns the AccessibleComponent for this Accessible
	 */
	public AccessibleComponent getAccessibleComponent() {
	    return this;
	}
	
	/**
	 * Gets the role of this object.  The role of the object is the generic
	 * purpose or use of the class of this object.  For example, the role
	 * of a push button is AccessibleRole.PUSH_BUTTON.  The roles in 
	 * AccessibleRole are provided so component developers can pick from
	 * a set of predefined roles.  This enables assistive technologies to
	 * provide a consistent interface to various tweaked subclasses of 
	 * components (e.g., use AccessibleRole.PUSH_BUTTON for all components
	 * that act like a push button) as well as distinguish between sublasses
	 * that behave differently (e.g., AccessibleRole.CHECK_BOX for check boxes
	 * and AccessibleRole.RADIO_BUTTON for radio buttons).
	 * <p>Note that the AccessibleRole class is also extensible, so 
	 * custom component developers can define their own AccessibleRole's
	 * if the set of predefined roles is inadequate.
	 *
	 * @return an instance of AccessibleRole describing the role of the object
	 * @see AccessibleRole
	 */
	public AccessibleRole getAccessibleRole() {
	    return AccessibleRole.TEXT;
	}

/**
	 * Returns the number of accessible children of the object.
	 *
	 * @return the number of accessible children of the object.
	 */
	public int getAccessibleChildrenCount() {
	    return children.size();
	}
	
	/**
	 * Returns the specified Accessible child of the object.  The Accessible
	 * children of an Accessible object are zero-based, so the first child 
	 * of an Accessible child is at index 0, the second child is at index 1,
	 * and so on.
	 *
	 * @param i zero-based index of child
	 * @return the Accessible child of the object
	 * @see #getAccessibleChildrenCount
	 */
	public Accessible getAccessibleChild(int i) {
	    if (i < 0 || i >= children.size()) {
		return null;
	    } else {
		return (Accessible)children.elementAt(i);
	    }
	}

	/*
	 * builds a list of AccessibleContexts that are the children
	 * of the specified parent view
	 */
	private void buildAccessibleChildrenList(View parent) {

	    Document doc = parent.getDocument();

	    for (int i = 0; i < parent.getViewCount(); i++) {
		View child = parent.getView(i);
		
		if (child instanceof TableView) { // AccessibleTable
		    AccessibleHTMLTable table = 
			new AccessibleHTMLTable((TableView)child, this, i);
		    children.addElement(table);

		} else if (child instanceof InlineView) { // AccessibleText
		    // skip view containing a single newline
		    try {
			String text = doc.getText(child.getStartOffset(),
						  child.getEndOffset() -
						  child.getStartOffset());
			if (text.equals("\n")) {
			    continue;
			}
		    } catch (BadLocationException e) {
		    }

		    // is this text a hyperlink?
		    boolean isHyperlink = false;
		    HTMLDocument hdoc = (HTMLDocument)editor.getDocument();
		    Element elem = hdoc.getCharacterElement(child.getStartOffset());
		    AttributeSet a = elem.getAttributes();
		    if (a != null) {
			AttributeSet anchor = (AttributeSet)a.
			    getAttribute(HTML.Tag.A);
			if (anchor != null) {
			    String href = (String)anchor.getAttribute
				(HTML.Attribute.HREF);
			    if (href != null) {
				isHyperlink = true;
				AccessibleHTMLAnchorSet anchorSet =
				    new AccessibleHTMLAnchorSet((InlineView)child, this, i, href);
				children.add(anchorSet);
			    }
			}
		    }
		    if (!isHyperlink) {
			AccessibleHTMLText text = 
			    new AccessibleHTMLText((InlineView)child, this, i);
			children.addElement(text);
		    }
			
		} else if (child instanceof ImageView) { // AccessibleImage
		    AccessibleHTMLImage image = 
			new AccessibleHTMLImage((ImageView)child, this, i);
		    children.addElement(image);

		} else if (child instanceof FormView) {	// Form
		    Component c = ((FormView)child).getComponent();
		    if (c instanceof Accessible)
			children.addElement((Accessible)c);
		}
		
		if (child.getViewCount() > 0) {
		    buildAccessibleChildrenList(child);
		}
	    }
	}
    }
    
    /*
     * base class for the AccessibleContexts associated with the views
     * in the editor pane
     */
    private abstract class AccessibleHTMLContext extends AccessibleContext
       implements Accessible, AccessibleComponent {

	protected View view;
	protected Accessible parent;
	protected int indexInParent;

	/*
	 * constructor
	 */
	protected AccessibleHTMLContext(View view, Accessible parent,
					int indexInParent) {
	    this.view = view;
	    this.parent = parent;
	    this.indexInParent = indexInParent;
	}

	/*
	 * returns the AccessibleContext for this Accessible
	 */
	public AccessibleContext getAccessibleContext() {
	    return this;
	}

	/*
	 * returns the AccessibleComponent for this Accessible
	 */
	public AccessibleComponent getAccessibleComponent() {
	    return this;
	}
	
	/*
	 * returns the parent for this Accessible
	 */
	public Accessible getAccessibleParent() {
	    return parent;
	}

	/*
	 * returns the start offset in the model for this view
	 */
	protected int getStartOffset() {
	    return view.getStartOffset();
	}

	/*
	 * returns the end offset in the model for this view
	 */
	protected int getEndOffset() {
	    return view.getEndOffset();
	}

	// ----- AccessibleContext -----

	/**
	 * Gets the accessibleName property of this object.  The accessibleName
	 * property of an object is a localized String that designates the purpose
	 * of the object.  For example, the accessibleName property of a label
	 * or button might be the text of the label or button itself.  In the
	 * case of an object that doesn't display its name, the accessibleName
	 * should still be set.  For example, in the case of a text field used
	 * to enter the name of a city, the accessibleName for the en_US locale
	 * could be 'city.'
	 *
	 * @return the localized name of the object; null if this 
	 * object does not have a name
	 *
	 * @see #setAccessibleName
	 */
	public String getAccessibleName() {
	    Document doc = view.getDocument();
	    if (doc != null) {
		return (String)doc.getProperty(Document.TitleProperty);
	    } else {
		return null;
	    }
	}

        /**
         * Gets the accessibleDescription property of this object.  If this
         * property isn't set, returns the content type of this
         * <code>JEditorPane</code> instead (e.g. "plain/text", "html/text").
         *
         * @return the localized description of the object; <code>null</code>
         * 	if this object does not have a description
         *
         * @see #setAccessibleName
         */
        public String getAccessibleDescription() {
	    return editor.getContentType();
        }

        /**
         * Gets the state set of this object.
         *
         * @return an instance of AccessibleStateSet describing the states
         * of the object
         * @see AccessibleStateSet
         */
        public AccessibleStateSet getAccessibleStateSet() {
            AccessibleStateSet states = new AccessibleStateSet();
	    AccessibleComponent acomp = getAccessibleComponent();
            if (acomp != null && acomp.isEnabled()) {
                states.add(AccessibleState.ENABLED);
            }
            if (acomp != null && acomp.isFocusTraversable()) {
                states.add(AccessibleState.FOCUSABLE);
            }
            if (acomp != null && acomp.isVisible()) {
                states.add(AccessibleState.VISIBLE);
            }
            if (acomp != null && acomp.isShowing()) {
                states.add(AccessibleState.SHOWING);
            }
            return states;
        }

	/**
	 * Gets the 0-based index of this object in its accessible parent.
	 *
	 * @return the 0-based index of this object in its parent; -1 if this 
	 * object does not have an accessible parent.
	 *
	 * @see #getAccessibleParent 
	 * @see #getAccessibleChildrenCount
	 * @see #getAccessibleChild
	 */
	public int getAccessibleIndexInParent() {
	    return indexInParent;
	}
	
	/**
	 * Returns the number of accessible children of the object.
	 *
	 * @return the number of accessible children of the object.
	 */
	public abstract int getAccessibleChildrenCount();
	
	/**
	 * Returns the specified Accessible child of the object.  The Accessible
	 * children of an Accessible object are zero-based, so the first child 
	 * of an Accessible child is at index 0, the second child is at index 1,
	 * and so on.
	 *
	 * @param i zero-based index of child
	 * @return the Accessible child of the object
	 * @see #getAccessibleChildrenCount
	 */
	public abstract Accessible getAccessibleChild(int i);
	
	/** 
	 * Gets the locale of the component. If the component does not have a 
	 * locale, then the locale of its parent is returned.  
	 *
	 * @return this component's locale.  If this component does not have 
	 * a locale, the locale of its parent is returned.
	 *
	 * @exception IllegalComponentStateException 
	 * If the Component does not have its own locale and has not yet been 
	 * added to a containment hierarchy such that the locale can be
	 * determined from the containing parent. 
	 */
	public Locale getLocale() throws IllegalComponentStateException {
	    return editor.getLocale();
	}


	// ----- AccessibleComponent -----

	/**
	 * Gets the background color of this object.
	 *
	 * @return the background color, if supported, of the object; 
	 * otherwise, null
	 * @see #setBackground
	 */
	public Color getBackground() {
	    return view.getContainer().getBackground();
	}
	
	/**
	 * Sets the background color of this object.
	 *
	 * @param c the new Color for the background
	 * @see #setBackground
	 */
	public void setBackground(Color c) {
	    // should be handled by the view
	}
	
	/**
	 * Gets the foreground color of this object.
	 *
	 * @return the foreground color, if supported, of the object; 
	 * otherwise, null
	 * @see #setForeground
	 */
	public Color getForeground() {
	    return view.getContainer().getForeground();
	}
	
	/**
	 * Sets the foreground color of this object.
	 *
	 * @param c the new Color for the foreground
	 * @see #getForeground
	 */
	public void setForeground(Color c) {
	    // should be handled by view
	}
	
	/**
	 * Gets the Cursor of this object.
	 *
	 * @return the Cursor, if supported, of the object; otherwise, null
	 * @see #setCursor
	 */
	public Cursor getCursor() {
	    return view.getContainer().getCursor();
	}
	
	/**
	 * Sets the Cursor of this object.
	 *
	 * @param c the new Cursor for the object
	 * @see #getCursor
	 */
	public void setCursor(Cursor cursor) {
	    // should be handled by view
	}
	
	/**
	 * Gets the Font of this object.
	 *
	 * @return the Font,if supported, for the object; otherwise, null
	 * @see #setFont
	 */
	public Font getFont() {
	    return view.getContainer().getFont();
	}
	
	/**
	 * Sets the Font of this object.
	 *
	 * @param f the new Font for the object
	 * @see #getFont
	 */
	public void setFont(Font f) {
	    // should be handled by view
	}
	
	/**
	 * Gets the FontMetrics of this object.
	 *
	 * @param f the Font
	 * @return the FontMetrics, if supported, the object; otherwise, null
	 * @see #getFont
	 */
	public FontMetrics getFontMetrics(Font f) {
	    // XXX
	    return null;
	}
	
	/**
	 * Determines if the object is enabled.  Objects that are enabled
	 * will also have the AccessibleState.ENABLED state set in their
	 * AccessibleStateSets.
	 *
	 * @return true if object is enabled; otherwise, false
	 * @see #setEnabled
	 * @see AccessibleContext#getAccessibleStateSet
	 * @see AccessibleState#ENABLED
	 * @see AccessibleStateSet
	 */
	public boolean isEnabled() {
	    return view.getContainer().isEnabled();
	}
	
	/**
	 * Sets the enabled state of the object.
	 *
	 * @param b if true, enables this object; otherwise, disables it 
	 * @see #isEnabled
	 */
	public void setEnabled(boolean b) {
	    // should be handled by view
	}
	
	/**
	 * Determines if the object is visible.  Note: this means that the
	 * object intends to be visible; however, it may not be
	 * showing on the screen because one of the objects that this object
	 * is contained by is currently not visible.  To determine if an object is
	 * showing on the screen, use isShowing().
	 * <p>Objects that are visible will also have the 
	 * AccessibleState.VISIBLE state set in their AccessibleStateSets.
	 *
	 * @return true if object is visible; otherwise, false
	 * @see #setVisible
	 * @see AccessibleContext#getAccessibleStateSet
	 * @see AccessibleState#VISIBLE
	 * @see AccessibleStateSet
	 */
	public boolean isVisible() {
	    return view.getContainer().isVisible();
	}
	
	/**
	 * Sets the visible state of the object.
	 *
	 * @param b if true, shows this object; otherwise, hides it 
	 * @see #isVisible
	 */
	public void setVisible(boolean b) {
	    // should be handled by view
	}
	
	/**
	 * Determines if the object is showing.  This is determined by checking
	 * the visibility of the object and its ancestors.
	 * Note: this
	 * will return true even if the object is obscured by another (for 
	 * example, it is underneath a menu that was pulled down).
	 *
	 * @return true if object is showing; otherwise, false
	 */
	public boolean isShowing() {
	    return view.getContainer().isShowing();
	}

	/** 
	 * Checks whether the specified point is within this object's bounds,
	 * where the point's x and y coordinates are defined to be relative 
	 * to the coordinate system of the object. 
	 *
	 * @param p the Point relative to the coordinate system of the object
	 * @return true if object contains Point; otherwise false
	 * @see #getBounds
	 */
	public boolean contains(Point p) {
	    Rectangle r = getBounds();
	    if (r != null) {
		return r.contains(p.x, p.y);
	    } else {
		return false;
	    }
	}
	
	/** 
	 * Returns the location of the object on the screen.
	 *
	 * @return the location of the object on screen; null if this object
	 * is not on the screen
	 * @see #getBounds
	 * @see #getLocation
	 */
	public Point getLocationOnScreen() {
	    Point editorLocation = view.getContainer().getLocationOnScreen();
	    Rectangle r = getBounds();
	    if (r != null) {
		return new Point(editorLocation.x + r.x,
				 editorLocation.y + r.y);
	    } else {
		return null;
	    }
	}
	
	/** 
	 * Gets the location of the object relative to the parent in the form 
	 * of a point specifying the object's top-left corner in the screen's 
	 * coordinate space.
	 *
	 * @return An instance of Point representing the top-left corner of the 
	 * object's bounds in the coordinate space of the screen; null if
	 * this object or its parent are not on the screen
	 * @see #getBounds
	 * @see #getLocationOnScreen
	 */
	public Point getLocation() {
	    Rectangle r = getBounds();
	    if (r != null) {
		return new Point(r.x, r.y);
	    } else {
		return null;
	    }
	}

	/** 
	 * Sets the location of the object relative to the parent.
	 * @param p the new position for the top-left corner
	 * @see #getLocation
	 */
	public void setLocation(Point p) {
	    // should be done by view.
	}
	
	/** 
	 * Gets the bounds of this object in the form of a Rectangle object. 
	 * The bounds specify this object's width, height, and location
	 * relative to its parent. 
	 *
	 * @return A rectangle indicating this component's bounds; null if 
	 * this object is not on the screen.
	 * @see #contains
	 */
	public Rectangle getBounds() {
	    int startOffset = view.getStartOffset();
	    int endOffset = view.getEndOffset();
	    Rectangle bounds;
	    TextUI ui = editor.getUI();
	    try {
		Shape lBounds = ui.modelToView(editor, startOffset,
					       Position.Bias.Forward);
		Shape rBounds = ui.modelToView(editor, endOffset,
					       Position.Bias.Forward);
		bounds = lBounds.getBounds();
		bounds.add((rBounds instanceof Rectangle) ?
			   (Rectangle)rBounds : rBounds.getBounds());
		return bounds;
	    } catch (BadLocationException ble) {
		return null;
	    }
	}
	
	/** 
	 * Sets the bounds of this object in the form of a Rectangle object. 
	 * The bounds specify this object's width, height, and location
	 * relative to its parent.
	 *	
	 * @param r rectangle indicating this component's bounds
	 * @see #getBounds
	 */
	public void setBounds(Rectangle r) {
	    // should be done by view
	}
	
	/** 
	 * Returns the size of this object in the form of a Dimension object. 
	 * The height field of the Dimension object contains this object's
	 * height, and the width field of the Dimension object contains this 
	 * object's width. 
	 *
	 * @return A Dimension object that indicates the size of this component; 
	 * null if this object is not on the screen
	 * @see #setSize
	 */
	public Dimension getSize() {
	    Rectangle r = getBounds();
	    if (r != null) {
		return new Dimension(r.width, r.height);
	    } else {
		return null;
	    }
	}
	
	/** 
	 * Resizes this object so that it has width and height. 
	 *	
	 * @param d The dimension specifying the new size of the object. 
	 * @see #getSize
	 */
	public void setSize(Dimension d) {
	    // should be done by view
	}
	
	/**
	 * Returns the Accessible child, if one exists, contained at the local 
	 * coordinate Point.
	 *
	 * @param p The point relative to the coordinate system of this object.
	 * @return the Accessible, if it exists, at the specified location; 
	 * otherwise null
	 */
	public Accessible getAccessibleAt (Point p) {
	    for (int i = 0; i < children.size(); i++) {
		AccessibleContext ac = (AccessibleContext)children.elementAt(i);
		Rectangle bounds = null;
		TextUI ui = editor.getUI();
		int start = ((AccessibleHTMLContext)ac).getStartOffset();
		int end = ((AccessibleHTMLContext)ac).getEndOffset();
		try {
		    Shape lBounds = ui.modelToView(editor, start,
						   Position.Bias.Forward);
		    Shape rBounds = ui.modelToView(editor, end,
						   Position.Bias.Forward);
		    bounds = lBounds.getBounds();
		    bounds.add((rBounds instanceof Rectangle) ?
			       (Rectangle)rBounds : rBounds.getBounds());
		    if (bounds.contains(p)) {
			return (Accessible)ac;
		    }
		} catch (BadLocationException e) {
		}
	    }
	    return null;
	}

	/**
	 * Returns whether this object can accept focus or not.   Objects that 
	 * can accept focus will also have the AccessibleState.FOCUSABLE state 
	 * set in their AccessibleStateSets.
	 *
	 * @return true if object can accept focus; otherwise false
	 * @see AccessibleContext#getAccessibleStateSet
	 * @see AccessibleState#FOCUSABLE
	 * @see AccessibleState#FOCUSED
	 * @see AccessibleStateSet
	 */
	public boolean isFocusTraversable() {
	    return view.getContainer().isFocusTraversable();
	}
	
	/**
	 * Requests focus for this object.  If this object cannot accept focus,
	 * nothing will happen.  Otherwise, the object will attempt to take
	 * focus.
	 * @see #isFocusTraversable
	 */
	public void requestFocus() {
	    view.getContainer().requestFocus();
	}
	
	/**
	 * Adds the specified focus listener to receive focus events from this 
	 * component. 
	 *
	 * @param l the focus listener
	 * @see #removeFocusListener
	 */
	public void addFocusListener(FocusListener l) {
	    view.getContainer().addFocusListener(l);
	}
	
	/**
	 * Removes the specified focus listener so it no longer receives focus 
	 * events from this component.
	 *
	 * @param l the focus listener
	 * @see #addFocusListener
	 */
	public void removeFocusListener(FocusListener l) {
	    view.getContainer().removeFocusListener(l);
	}
    }

    /*
     * Implements AccessibleIcon
     */
    private class AccessibleHTMLImage extends AccessibleHTMLContext
        implements Accessible, AccessibleIcon, ImageObserver {

	/*
	 * constructor
	 */
	protected AccessibleHTMLImage(ImageView view, Accessible parent,
				      int indexInParent) {
	    super(view, parent, indexInParent);
	}

	// ----- AccessibleContext -----

	/*
	 * returns the AccessibleContext for this Accessible
	 */
	public AccessibleContext getAccessibleContext() {
	    return this;
	}

	/**
	 * Gets the role of this object.  The role of the object is the generic
	 * purpose or use of the class of this object.  For example, the role
	 * of a push button is AccessibleRole.PUSH_BUTTON.  The roles in 
	 * AccessibleRole are provided so component developers can pick from
	 * a set of predefined roles.  This enables assistive technologies to
	 * provide a consistent interface to various tweaked subclasses of 
	 * components (e.g., use AccessibleRole.PUSH_BUTTON for all components
	 * that act like a push button) as well as distinguish between sublasses
	 * that behave differently (e.g., AccessibleRole.CHECK_BOX for check boxes
	 * and AccessibleRole.RADIO_BUTTON for radio buttons).
	 * <p>Note that the AccessibleRole class is also extensible, so 
	 * custom component developers can define their own AccessibleRole's
	 * if the set of predefined roles is inadequate.
	 *
	 * @return an instance of AccessibleRole describing the role of the object
	 * @see AccessibleRole
	 */
	public AccessibleRole getAccessibleRole() {
	    return AccessibleRole.ICON;
	}
	
	/**
	 * Gets the 0-based index of this object in its accessible parent.
	 *
	 * @return the 0-based index of this object in its parent; -1 if this 
	 * object does not have an accessible parent.
	 *
	 * @see #getAccessibleParent 
	 * @see #getAccessibleChildrenCount
	 * @see #getAccessibleChild
	 */
	public int getAccessibleIndexInParent() {
	    return indexInParent;
	}

	/**
	 * Returns the number of accessible children of the object.
	 *
	 * @return the number of accessible children of the object.
	 */
	public int getAccessibleChildrenCount() {
	    return 0;
	}
	
	/**
	 * Returns the specified Accessible child of the object.  The Accessible
	 * children of an Accessible object are zero-based, so the first child 
	 * of an Accessible child is at index 0, the second child is at index 1,
	 * and so on.
	 *
	 * @param i zero-based index of child
	 * @return the Accessible child of the object
	 * @see #getAccessibleChildrenCount
	 */
	public Accessible getAccessibleChild(int i) {
	    return null;
	}

	// ----- AccessibleIcon -----
	
	public AccessibleIcon [] getAccessibleIcon() {
	    AccessibleIcon [] icons = new AccessibleIcon[1];
	    icons[0] = this;
	    return icons;
	}

	/**
	 * Gets the description of the icon.  This is meant to be a brief
	 * textual description of the object.  For example, it might be
	 * presented to a blind user to give an indication of the purpose
	 * of the icon.
	 *
	 * @return the description of the icon
	 */
	public String getAccessibleIconDescription() {
	    return ((ImageView)view).getAltText();
	}
	
	/**
	 * Sets the description of the icon.  This is meant to be a brief
	 * textual description of the object.  For example, it might be
	 * presented to a blind user to give an indication of the purpose
	 * of the icon.
	 *
	 * @param description the description of the icon
	 */
	public void setAccessibleIconDescription(String description) {
	    // should be handled by the view
	}
	
	/**
	 * Gets the width of the icon
	 *
	 * @return the width of the icon.
	 */
	public int getAccessibleIconWidth() {
	    return ((ImageView)view).getImage().getWidth(this);
	}
	
	/**
	 * Gets the height of the icon
	 *
	 * @return the height of the icon.
	 */
	public int getAccessibleIconHeight() {
	    return ((ImageView)view).getImage().getHeight(this);
	}
	
	/**
	 * ImageObserver callback 
	 */
        public boolean imageUpdate(Image img, int flags, int x, int y,
                                   int newWidth, int newHeight ) {
	    return false;
	}
    }

    /*
     * Implements AccessibleHypertext
     */
    private class AccessibleHTMLAnchorSet extends AccessibleHTMLText
        implements Accessible, AccessibleHypertext {

	private AccessibleHTMLAnchor link;

	/*
	 * constructor
	 */
	protected AccessibleHTMLAnchorSet(InlineView view, Accessible parent,
					  int indexInParent, String href) {
	    super(view, parent, indexInParent);
	    link = new AccessibleHTMLAnchor();
	}

	// ----- AccessibleContext -----

	/*
	 * returns the AccessibleContext for this Accessible
	 */
	public AccessibleContext getAccessibleContext() {
	    return this;
	}

	/**
	 * Gets the role of this object.  The role of the object is the generic
	 * purpose or use of the class of this object.  For example, the role
	 * of a push button is AccessibleRole.PUSH_BUTTON.  The roles in 
	 * AccessibleRole are provided so component developers can pick from
	 * a set of predefined roles.  This enables assistive technologies to
	 * provide a consistent interface to various tweaked subclasses of 
	 * components (e.g., use AccessibleRole.PUSH_BUTTON for all components
	 * that act like a push button) as well as distinguish between sublasses
	 * that behave differently (e.g., AccessibleRole.CHECK_BOX for check boxes
	 * and AccessibleRole.RADIO_BUTTON for radio buttons).
	 * <p>Note that the AccessibleRole class is also extensible, so 
	 * custom component developers can define their own AccessibleRole's
	 * if the set of predefined roles is inadequate.
	 *
	 * @return an instance of AccessibleRole describing the role of the object
	 * @see AccessibleRole
	 */
	public AccessibleRole getAccessibleRole() {
	    return AccessibleRole.HYPERLINK;
	}
	
	/**
	 * Gets the 0-based index of this object in its accessible parent.
	 *
	 * @return the 0-based index of this object in its parent; -1 if this 
	 * object does not have an accessible parent.
	 *
	 * @see #getAccessibleParent 
	 * @see #getAccessibleChildrenCount
	 * @see #getAccessibleChild
	 */
	public int getAccessibleIndexInParent() {
	    return indexInParent;
	}

	/**
	 * Returns the number of accessible children of the object.
	 *
	 * @return the number of accessible children of the object.
	 */
	public int getAccessibleChildrenCount() {
	    return 0;
	}

	/**
	 * Returns the specified Accessible child of the object.  The Accessible
	 * children of an Accessible object are zero-based, so the first child 
	 * of an Accessible child is at index 0, the second child is at index 1,
	 * and so on.
	 *
	 * @param i zero-based index of child
	 * @return the Accessible child of the object
	 * @see #getAccessibleChildrenCount
	 */
	public Accessible getAccessibleChild(int i) {
	    return null;
	}

	// ----- AccessibleHypertext -----

	public AccessibleText getAccessibleText() {
	    return this;
	}

	public AccessibleHypertext getAccessibleHypertext() {
	    return this;
	}
	
	/**
	 * Returns the number of links within this hypertext document.
	 *
	 * @return number of links in this hypertext doc.
	 */
	public int getLinkCount() {
	    return 1;
	}
	
	/**
	 * Returns the nth Link of this Hypertext document.
	 *
	 * @param linkIndex within the links of this Hypertext
	 * @return Link object encapsulating the nth link(s)
	 */
	public AccessibleHyperlink getLink(int linkIndex) {
	    if (linkIndex != 0) {
		return null;
	    } else {
		return link;
	    }
	}
	
	/**
	 * Returns the index into an array of hyperlinks that
	 * is associated with this character index, or -1 if there
	 * is no hyperlink associated with this index.
	 *
	 * @param charIndex index within the text
	 * @return index into the set of hyperlinks for this hypertext doc.
	 */
	public int getLinkIndex(int charIndex) {
	    return 0;
	}

	// ----- AccessibleHyperlink -----
	
	class AccessibleHTMLAnchor extends AccessibleHyperlink {
	    /**
	     * Since the document a link is associated with may have
	     * changed, this method returns whether this Link is valid
	     * anymore (with respect to the document it references).
	     *
	     * @return a flag indicating whether this link is still valid with
	     *         respect to the AccessibleHypertext it belongs to
	     */
	    public boolean isValid() {
		return true;	// needs to be implemented
	    }
	    
	    /**
	     * Returns the number of accessible actions available in this Link
	     * If there are more than one, the first one is NOT considered the
	     * "default" action of this LINK object (e.g. in an HTML imagemap).
	     * In general, links will have only one AccessibleAction in them.
	     *
	     * @return the zero-based number of Actions in this object
	     */
	    public int getAccessibleActionCount() {
		return 1;
	    }
	    
	    /**
	     * Perform the specified Action on the object
	     *
	     * @param i zero-based index of actions
	     * @return true if the the action was performed; else false.
	     * @see #getAccessibleActionCount
	     */
	    public boolean doAccessibleAction(int i) {
		if (i == 0 && isValid() == true) {
		    URL u = (URL) getAccessibleActionObject(i);
		    if (u != null) {
			HyperlinkEvent linkEvent =
			    new HyperlinkEvent(editor, HyperlinkEvent.EventType.ACTIVATED, u);
			editor.fireHyperlinkUpdate(linkEvent);
			return true;
		    }
		}
		return false;  // link invalid or i != 0
	    }
	    
	    /**
	     * Return a String description of this particular
	     * link action.  The string returned is the text
	     * within the document associated with the element
	     * which contains this link.
	     *
	     * @param i zero-based index of the actions
	     * @return a String description of the action
	     * @see #getAccessibleActionCount
	     */
	    public String getAccessibleActionDescription(int i) {
		if (i == 0 && isValid() == true) {
		    Document d = editor.getDocument();
		    if (d != null) {
			try {
			    return d.getText(getStartIndex(),
					     getEndIndex() - getStartIndex());
			} catch (BadLocationException exception) {
			    return null;
			}
		    }
		}
		return null;
	    }
	    
	    /**
	     * Returns a URL object that represents the link.
	     *
	     * @param i zero-based index of the actions
	     * @return an URL representing the HTML link itself
	     * @see #getAccessibleActionCount
	     */
	    public Object getAccessibleActionObject(int i) {
		if (i == 0 && isValid() == true) {
		    AttributeSet as = view.getAttributes();
		    AttributeSet anchor = 
			(AttributeSet) as.getAttribute(HTML.Tag.A);
		    String href = (anchor != null) ?
			(String) anchor.getAttribute(HTML.Attribute.HREF) : null;
		    if (href != null) {
			URL u;
			try {
			    u = new URL(editor.getPage(), href);
			} catch (MalformedURLException m) {
			    u = null;
			}
			return u;
		    }
		}
		return null;  // link invalid or i != 0
	    }
	    
	    /**
	     * Return an object that represents the link anchor,
	     * as appropriate for that link.  E.g. from HTML:
	     *   <a href="http://www.sun.com/access">Accessibility</a>
	     * this method would return a String containing the text:
	     * 'Accessibility'.
	     *
	     * Similarly, from this HTML:
	     *   &lt;a HREF="#top"&gt;&lt;img src="top-hat.gif" alt="top hat"&gt;&lt;/a&gt;
	     * this might return the object ImageIcon("top-hat.gif", "top hat");
	     *
	     * @param i zero-based index of the actions
	     * @return an Object representing the hypertext anchor
	     * @see #getAccessibleActionCount
	     */
	    public Object getAccessibleActionAnchor(int i) {
		return getAccessibleActionDescription(i);
	    }
	    
	    /**
	     * Get the index with the hypertext document at which this
	     * link begins
	     *
	     * @return index of start of link
	     */
	    public int getStartIndex() {
		return view.getStartOffset();
	    }
	    
	    /**
	     * Get the index with the hypertext document at which this
	     * link ends
	     *
	     * @return index of end of link
	     */
	    public int getEndIndex() {
		return view.getEndOffset();
	    }
	}
    }

    /*
     * Implements AccessibleTable
     */
    private class AccessibleHTMLTable extends AccessibleHTMLContext
        implements Accessible, AccessibleTable {

	private View cellViews [][];
	private Accessible cells [][];
	private int rowCount;
	private int columnCount;

	/*
	 * constructor
	 */
	protected AccessibleHTMLTable(TableView view, Accessible parent,
				      int indexInParent) {
	    super(view, parent, indexInParent);

	    // iterate through the individual table cells
	    rowCount = ((TableView)view).getRowCount();
	    columnCount = ((TableView)view).getColumnCount();
	    cellViews = new View[rowCount][columnCount];
	    cells = new Accessible[rowCount][columnCount];

	    int numRows = ((TableView)view).getViewCount();
	    for (int i = 0; i < numRows; i++) {  // rows

		TableView.RowView rowView = (TableView.RowView)view.getView(i);
		for (int j = 0; j < rowView.getColumnCount(); j++) {  // columns

		    TableView.CellView cellView = // cells
			(TableView.CellView)rowView.getView(j);
		    
		    // next two views only contain a single subview
		    // XXX is this correct?
		    ParagraphView paraView = (ParagraphView)cellView.getView(0);
		    View paraViewRow = paraView.getView(0); // ParagraphView.row

		    // XXX only handles text views right now and ignores
		    // the single newline in the view(1)
		    if (paraViewRow.getView(0) instanceof InlineView) {
			InlineView inlineView = (InlineView)paraViewRow.getView(0);
			cellViews[i][j] = inlineView;
			cells[i][j] = new AccessibleHTMLText(inlineView, this,
						i * rowView.getColumnCount() + j);
		    }
		}
	    }
	}

	// ----- AccessibleContext -----

	/*
	 * returns the AccessibleContext for this Accessible
	 */
	public AccessibleContext getAccessibleContext() {
	    return this;
	}

	/**
	 * Gets the role of this object.  The role of the object is the generic
	 * purpose or use of the class of this object.  For example, the role
	 * of a push button is AccessibleRole.PUSH_BUTTON.  The roles in 
	 * AccessibleRole are provided so component developers can pick from
	 * a set of predefined roles.  This enables assistive technologies to
	 * provide a consistent interface to various tweaked subclasses of 
	 * components (e.g., use AccessibleRole.PUSH_BUTTON for all components
	 * that act like a push button) as well as distinguish between sublasses
	 * that behave differently (e.g., AccessibleRole.CHECK_BOX for check boxes
	 * and AccessibleRole.RADIO_BUTTON for radio buttons).
	 * <p>Note that the AccessibleRole class is also extensible, so 
	 * custom component developers can define their own AccessibleRole's
	 * if the set of predefined roles is inadequate.
	 *
	 * @return an instance of AccessibleRole describing the role of the object
	 * @see AccessibleRole
	 */
	public AccessibleRole getAccessibleRole() {
	    return AccessibleRole.TABLE;
	}

	/**
	 * Gets the 0-based index of this object in its accessible parent.
	 *
	 * @return the 0-based index of this object in its parent; -1 if this 
	 * object does not have an accessible parent.
	 *
	 * @see #getAccessibleParent 
	 * @see #getAccessibleChildrenCount
	 * @see #getAccessibleChild
	 */
	public int getAccessibleIndexInParent() {
	    return indexInParent;
	}

	/**
	 * Returns the number of accessible children of the object.
	 *
	 * @return the number of accessible children of the object.
	 */
	public int getAccessibleChildrenCount() {
	    return rowCount * columnCount;
	}
	
	/**
	 * Returns the specified Accessible child of the object.  The Accessible
	 * children of an Accessible object are zero-based, so the first child 
	 * of an Accessible child is at index 0, the second child is at index 1,
	 * and so on.
	 *
	 * @param i zero-based index of child
	 * @return the Accessible child of the object
	 * @see #getAccessibleChildrenCount
	 */
	public Accessible getAccessibleChild(int i) {
	    int r = i / rowCount;
	    int c = i % rowCount;
	    if (r < 0 || r >= rowCount || c < 0 || c >= columnCount) {
		return null;
	    } else {
		return cells[r][c];
	    }
	}

	// ----- AccessibleTable -----
	
	public AccessibleTable getAccessibleTable() {
	    return this;
	}

	/**
	 * Returns the caption for the table.
	 *
	 * @return the caption for the table
	 */
	public Accessible getAccessibleCaption() {
	    return null;	// XXX tbd
	}
	
	/**
	 * Sets the caption for the table.
	 *
	 * @param a the caption for the table
	 */
	public void setAccessibleCaption(Accessible a) {
	}
	
	/**
	 * Returns the summary description of the table.
	 * 
	 * @return the summary description of the table
	 */
	public Accessible getAccessibleSummary() {
	    return null;	// XXX tbd
	}
	
	/**
	 * Sets the summary description of the table
	 *
	 * @param a the summary description of the table
	 */
	public void setAccessibleSummary(Accessible a) {
	}
	
	/**
	 * Returns the number of rows in the table.
	 *
	 * @return the number of rows in the table
	 */
	public int getAccessibleRowCount() {
	    return ((TableView)view).getRowCount();
	}
	
	/**
	 * Returns the number of columns in the table.
	 *
	 * @return the number of columns in the table
	 */
	public int getAccessibleColumnCount() {
	    return ((TableView)view).getColumnCount();
	}
	
	/**
	 * Returns the Accessible at a specified row and column
	 * in the table.
	 *
	 * @param r zero-based row of the table
	 * @param c zero-based column of the table
	 * @return the Accessible at the specified row and column
	 */
	public Accessible getAccessibleAt(int r, int c) {
	    return cells[r][c];
	}
	
	/**
	 * Returns the number of rows occupied by the Accessible at
	 * a specified row and column in the table.
	 *
	 * @return the number of rows occupied by the Accessible at a
	 * given specified (row, column)
	 */
	public int getAccessibleRowExtentAt(int r, int c) {
	    return ((TableView)view).getRowsOccupied(cellViews[r][c]);
	}
	
	/**
	 * Returns the number of columns occupied by the Accessible at
	 * a specified row and column in the table.
	 *
	 * @return the number of columns occupied by the Accessible at a
	 * given specified row and column
	 */
	public int getAccessibleColumnExtentAt(int r, int c) {
	    return ((TableView)view).getColumnsOccupied(cellViews[r][c]);
	}

	/**
	 * Returns the row headers as an AccessibleTable.
	 *
	 * @return an AccessibleTable representing the row
	 * headers
	 */
	public AccessibleTable getAccessibleRowHeader() {
	    return null;	// XXX tbd
	}
	
	/**
	 * Sets the row headers.
	 *
	 * @param table an AccessibleTable representing the
	 * row headers
	 */
	public void setAccessibleRowHeader(AccessibleTable table) {
	}
	
	/**
	 * Returns the column headers as an AccessibleTable.
	 *
	 * @return an AccessibleTable representing the column
	 * headers
	 */
	public AccessibleTable getAccessibleColumnHeader() {
	    return null;	// XXX tbd
	}
	
	/**
	 * Sets the column headers.
	 *
	 * @param table an AccessibleTable representing the
	 * column headers
	 */
	public void setAccessibleColumnHeader(AccessibleTable table) {
	}
	
	/**
	 * Returns the description of the specified row in the table.
	 *
	 * @param r zero-based row of the table
	 * @return the description of the row
	 */
	public Accessible getAccessibleRowDescription(int r) {
	    return null;	// XXX tbd
	}
	
	/**
	 * Sets the description text of the specified row of the table.
	 *
	 * @param r zero-based row of the table
	 * @param a the description of the row
	 */
	public void setAccessibleRowDescription(int r, Accessible a) {
	}
	
	/**
	 * Returns the description text of the specified column in the table.
	 *
	 * @param c zero-based column of the table
	 * @return the text description of the column
	 */
	public Accessible getAccessibleColumnDescription(int c) {
	    return null;	// XXX tbd
	}
	
	/**
	 * Sets the description text of the specified column in the table.
	 *
	 * @param c zero-based column of the table
	 * @param a the text description of the column
	 */
	public void setAccessibleColumnDescription(int c, Accessible a) {
	}
	
	/**
	 * Returns a boolean value indicating whether the accessible at
	 * a specified row and column is selected.
	 *
	 * @param r zero-based row of the table
	 * @param c zero-based column of the table
	 * @return the boolean value true if the accessible at the
	 * row and column is selected. Otherwise, the boolean value 
	 * false
	 */
	public boolean isAccessibleSelected(int r, int c) {
	    return false;	// XXX tbd
	}
	
	/**
	 * Returns a boolean value indicating whether the specified row
	 * is selected.
	 *
	 * @param r zero-based row of the table
	 * @return the boolean value true if the specified row is selected.
	 * Otherwise, false.
	 */
	public boolean isAccessibleRowSelected(int r) {
	    return false;	// XXX tbd
	}
	
	/**
	 * Returns a boolean value indicating whether the specified column
	 * is selected.
	 *
	 * @param r zero-based column of the table
	 * @return the boolean value true if the specified column is selected.
	 * Otherwise, false.
	 */
	public boolean isAccessibleColumnSelected(int c) {
	    return false;	// XXX tbd
	}
	
	/**
	 * Returns the selected rows in a table.
	 *
	 * @return an array of selected rows where each element is a
	 * zero-based row of the table
	 */
	public int [] getSelectedAccessibleRows() {
	    return new int [0];	// XXX tbd
	}
	
	/**
	 * Returns the selected columns in a table.
	 *
	 * @return an array of selected columns where each element is a
	 * zero-based column of the table
	 */
	public int [] getSelectedAccessibleColumns() {
	    return new int [0];	// XXX tbd
	}
    }

    /*
     * implements AccessibleText for the view
     */
    private class AccessibleHTMLText extends AccessibleHTMLContext 
        implements Accessible, AccessibleText {

	private Document model = editor.getDocument();

	/*
	 * constructor
	 */
	protected AccessibleHTMLText(InlineView view, Accessible parent,
				     int indexInParent) {
	    super(view, parent, indexInParent);
	}
	    
	// ----- AccessibleContext -----

	public AccessibleContext getAccessibleContext() {
	    return this;
	}

	/**
	 * Gets the role of this object.  The role of the object is the generic
	 * purpose or use of the class of this object.  For example, the role
	 * of a push button is AccessibleRole.PUSH_BUTTON.  The roles in 
	 * AccessibleRole are provided so component developers can pick from
	 * a set of predefined roles.  This enables assistive technologies to
	 * provide a consistent interface to various tweaked subclasses of 
	 * components (e.g., use AccessibleRole.PUSH_BUTTON for all components
	 * that act like a push button) as well as distinguish between sublasses
	 * that behave differently (e.g., AccessibleRole.CHECK_BOX for check boxes
	 * and AccessibleRole.RADIO_BUTTON for radio buttons).
	 * <p>Note that the AccessibleRole class is also extensible, so 
	 * custom component developers can define their own AccessibleRole's
	 * if the set of predefined roles is inadequate.
	 *
	 * @return an instance of AccessibleRole describing the role of the object
	 * @see AccessibleRole
	 */
	public AccessibleRole getAccessibleRole() {
	    return AccessibleRole.TEXT;
	}

	/**
	 * Gets the 0-based index of this object in its accessible parent.
	 *
	 * @return the 0-based index of this object in its parent; -1 if this 
	 * object does not have an accessible parent.
	 *
	 * @see #getAccessibleParent 
	 * @see #getAccessibleChildrenCount
	 * @see #getAccessibleChild
	 */
	public int getAccessibleIndexInParent() {
	    return indexInParent;
	}

	/**
	 * Returns the number of accessible children of the object.
	 *
	 * @return the number of accessible children of the object.
	 */
	public int getAccessibleChildrenCount() {
	    return 0;
	}
	
	/**
	 * Returns the specified Accessible child of the object.  The Accessible
	 * children of an Accessible object are zero-based, so the first child 
	 * of an Accessible child is at index 0, the second child is at index 1,
	 * and so on.
	 *
	 * @param i zero-based index of child
	 * @return the Accessible child of the object
	 * @see #getAccessibleChildrenCount
	 */
	public Accessible getAccessibleChild(int i) {
	    return null;
	}
	
	/* AccessibleText ---------- */

	public AccessibleText getAccessibleText() {
	    return this;
	}

	/**
	 * Given a point in local coordinates, return the zero-based index
	 * of the character under that Point.  If the point is invalid,
	 * this method returns -1.
	 *
	 * @param p the Point in local coordinates
	 * @return the zero-based index of the character under Point p; if 
	 * Point is invalid returns -1.
	 */
	public int getIndexAtPoint(Point p) {
	    return view.viewToModel(p.x, p.y, getBounds());
	}
	
	/**
	 * Determine the bounding box of the character at the given 
	 * index into the string.  The bounds are returned in local
	 * coordinates.  If the index is invalid an empty rectangle is 
	 * returned.
	 *
	 * @param i the index into the String
	 * @return the screen coordinates of the character's the bounding box,
	 * if index is invalid returns an empty rectangle.
	 */
	public Rectangle getCharacterBounds(int i) {
	    try {
		return editor.getUI().modelToView(editor, i);
	    } catch (BadLocationException e) {
		return null;
	    }
	}
	
	/**
	 * Return the number of characters (valid indicies) 
	 *
	 * @return the number of characters
	 */
	public int getCharCount() {
	    Document d = view.getDocument();
	    if (d != null) {
		return d.getLength();
	    } else {
		return 0;
	    }
	}
	
	/**
	 * Return the zero-based offset of the caret.
	 *
	 * Note: That to the right of the caret will have the same index
	 * value as the offset (the caret is between two characters).
	 * @return the zero-based offset of the caret.
	 */
	public int getCaretPosition() {
	    Container c = view.getContainer();
	    if (c instanceof JTextComponent) {
		return ((JTextComponent)c).getCaretPosition();
	    } else {
		return -1;
	    }
	}
	
       /**
         * IndexedSegment extends Segment adding the offset into the
         * the model the <code>Segment</code> was asked for.
         */
        private class IndexedSegment extends Segment {
            /**
             * Offset into the model that the position represents.
             */
            public int modelOffset;
        }


        public String getAtIndex(int part, int index) {
            return getAtIndex(part, index, 0);
        }


        public String getAfterIndex(int part, int index) {
            return getAtIndex(part, index, 1);
        }


        public String getBeforeIndex(int part, int index) {
            return getAtIndex(part, index, -1);
        }


        /**
         * Gets the word, sentence, or character at <code>index</code>.
         * If <code>direction</code> is non-null this will find the
         * next/previous word/sentence/character.
         */
        private String getAtIndex(int part, int index, int direction) {
            if (model instanceof AbstractDocument) {
                ((AbstractDocument)model).readLock();
            }
            try {
                if (index < 0 || index >= model.getLength()) {
                    return null;
                }
                switch (part) {
                case AccessibleText.CHARACTER:
                    if (index + direction < model.getLength() &&
                        index + direction >= 0) {
                        return model.getText(index + direction, 1);
                    }
                    break;


                case AccessibleText.WORD:
                case AccessibleText.SENTENCE:
                    IndexedSegment seg = getSegmentAt(part, index);
                    if (seg != null) {
                        if (direction != 0) {
                            int next;


                            if (direction < 0) {
                                next = seg.modelOffset - 1;
                            }
                            else {
                                next = seg.modelOffset + direction * seg.count;
                            }
                            if (next >= 0 && next <= model.getLength()) {
                                seg = getSegmentAt(part, next);
                            }
                            else {
                                seg = null;
                            }
                        }
                        if (seg != null) {
                            return new String(seg.array, seg.offset,
                                                  seg.count);
                        }
                    }
                    break;


                default:
                    break;
                }
            } catch (BadLocationException e) {
            } finally {
                if (model instanceof AbstractDocument) {
                    ((AbstractDocument)model).readUnlock();
                }
            }
            return null;
        }


        /*
         * Returns the paragraph element for the specified index.
         */
        private Element getParagraphElement(int index) {
            if (model instanceof PlainDocument ) {
                PlainDocument sdoc = (PlainDocument)model;
                return sdoc.getParagraphElement(index);
            } else if (model instanceof StyledDocument) {
                StyledDocument sdoc = (StyledDocument)model;
                return sdoc.getParagraphElement(index);
            } else {
                Element para = null;
                for (para = model.getDefaultRootElement(); ! para.isLeaf(); ) {
                    int pos = para.getElementIndex(index);
                    para = para.getElement(pos);
                }
                if (para == null) {
                    return null;
                }
                return para.getParentElement();
            }
        }
        
        /*
         * Returns a <code>Segment</code> containing the paragraph text
         * at <code>index</code>, or null if <code>index</code> isn't
         * valid.
         */
        private IndexedSegment getParagraphElementText(int index)
                                  throws BadLocationException {
            Element para = getParagraphElement(index);


            if (para != null) {
                IndexedSegment segment = new IndexedSegment();
                try {
                    int length = para.getEndOffset() - para.getStartOffset();
                    model.getText(para.getStartOffset(), length, segment);
                } catch (BadLocationException e) {
                    return null;
                }
                segment.modelOffset = para.getStartOffset();
                return segment;
            }
            return null;
        }

        /**
         * Returns the Segment at <code>index</code> representing either
         * the paragraph or sentence as identified by <code>part</code>, or
         * null if a valid paragraph/sentence can't be found. The offset
         * will point to the start of the word/sentence in the array, and
         * the modelOffset will point to the location of the word/sentence
         * in the model.
         */
        private IndexedSegment getSegmentAt(int part, int index) throws
                                  BadLocationException {
            IndexedSegment seg = getParagraphElementText(index);
            if (seg == null) {
                return null;
            }
            BreakIterator iterator;
            switch (part) {
            case AccessibleText.WORD:
                iterator = BreakIterator.getWordInstance();
                break;
            case AccessibleText.SENTENCE:
                iterator = BreakIterator.getSentenceInstance();
                break;
            default:
                return null;
            }
            seg.first();
            iterator.setText(seg);
            int end = iterator.following(index - seg.modelOffset + seg.offset);
            if (end == BreakIterator.DONE) {
                return null;
            }
            if (end > seg.offset + seg.count) {
                return null;
            }
            int begin = iterator.previous();
            if (begin == BreakIterator.DONE ||
                         begin >= seg.offset + seg.count) {
                return null;
            }
            seg.modelOffset = seg.modelOffset + begin - seg.offset;
            seg.offset = begin;
            seg.count = end - begin;
            return seg;
        }
 
	/**
	 * Return the AttributeSet for a given character at a given index
	 *
	 * @param i the zero-based index into the text 
	 * @return the AttributeSet of the character
	 */
	public AttributeSet getCharacterAttribute(int i) {
	    Document d = view.getDocument();
	    if (d instanceof StyledDocument) {
		StyledDocument doc = (StyledDocument)d;
		Element elem = doc.getCharacterElement(i);
		if (elem != null) {
		    return elem.getAttributes();
		}
	    }
	    return null;
	}

	/**
	 * Returns the start offset within the selected text.
	 * If there is no selection, but there is
	 * a caret, the start and end offsets will be the same.
	 *
	 * @return the index into the text of the start of the selection
	 */
	public int getSelectionStart() {
	    return -1;		// XXX tbd
	}

	/**
	 * Returns the end offset within the selected text.
	 * If there is no selection, but there is
	 * a caret, the start and end offsets will be the same.
	 *
	 * @return the index into teh text of the end of the selection
	 */
	public int getSelectionEnd() {
	    return -1;		// XXX tbd
	}

	/**
	 * Returns the portion of the text that is selected. 
	 *
	 * @return the String portion of the text that is selected
	 */
	public String getSelectedText() {
	    return null;	// XXX tbd
	}

	/*
	 * Returns the text substring starting at the specified
	 * offset with the specified length.
	 */
	private String getText(int offset, int length) 
	    throws BadLocationException {

	    Document d = view.getDocument();
	    if (d != null && d instanceof StyledDocument) {
		StyledDocument doc = (StyledDocument)d;
		return doc.getText(offset, length);
	    } else {
		return null;
	    }
	}
    }
    // ----- end Accessibility support
}
