/*
 *  $Id: SAXParserFactory.java,v 1.13 2001/03/07 02:31:07 edg Exp $
 *
 * Copyright (c) 1998-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package javax.xml.parsers;

import org.xml.sax.Parser;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;

import java.io.InputStream;
import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.util.Locale;

import java.util.Properties;
import java.io.BufferedReader;
import java.io.InputStreamReader;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

/**
 * Defines a factory API that enables applications to configure and
 * obtain a SAX based parser to parse XML documents.<p>
 * An implementation of the <code>SAXParserFactory</code> class is
 * <em>NOT</em> guaranteed to be thread safe. It is up to the user application 
 * to make sure about the use of the <code>SAXParserFactory</code> from 
 * more than one thread. Alternatively the application can have one instance 
 * of the <code>SAXParserFactory</code> per thread.
 * An application can use the same instance of the factory to obtain one or 
 * more instances of the <code>SAXParser</code> provided the instance
 * of the factory isn't being used in more than one thread at a time.
 *
 * @since JAXP 1.0
 * @version 1.0
 */

public abstract class SAXParserFactory {
    /** One FactoryFinder per factory class */
    private static FactoryFinder factoryFinder = new FactoryFinder(
        // The default property name according to the JAXP spec
        "javax.xml.parsers.SAXParserFactory",
        "org.apache.crimson.jaxp.SAXParserFactoryImpl");

    private boolean validating = false;
    private boolean namespaceAware= false;
    
    protected SAXParserFactory () {
    
    }

    /**
     * Obtain a new instance of a <code>SAXParserFactory</code>. This
     * static method creates a new factory instance
     * This method uses the following ordered lookup procedure to determine
     * the <code>SAXParserFactory</code> implementation class to
     * load:
     * <ul>
     * <li>
     * Use the <code>javax.xml.parsers.SAXParserFactory</code> system
     * property.
     * </li>
     * <li>
     * Use the properties file "lib/jaxp.properties" in the JRE directory.
     * This configuration file is in standard <code>java.util.Properties
     * </code> format and contains the fully qualified name of the
     * implementation class with the key being the system property defined
     * above.
     * </li>
     * <li>
     * Use the Services API (as detailed in the JAR specification), if
     * available, to determine the classname. The Services API will look
     * for a classname in the file
     * <code>META-INF/services/javax.xml.parsers.SAXParserFactory</code>
     * in jars available to the runtime.
     * </li>
     * <li>
     * Platform default <code>SAXParserFactory</code> instance.
     * </li>
     * </ul>
     *
     * Once an application has obtained a reference to a
     * <code>SAXParserFactory</code> it can use the factory to
     * configure and obtain parser instances.
     *
     * @return A new instance of a SAXParserFactory.
     *
     * @exception FactoryConfigurationError if the implementation is
     * not available or cannot be instantiated.
     */

    public static SAXParserFactory newInstance() {
        SAXParserFactory factoryImpl;
        try {
            Class factoryClass = factoryFinder.findFactory();
            factoryImpl = (SAXParserFactory) factoryClass.newInstance();
        } catch (ClassNotFoundException cnfe) {
            throw new FactoryConfigurationError(cnfe);
        } catch (IllegalAccessException iae) {
            throw new FactoryConfigurationError(iae);
        } catch (InstantiationException ie) {
            throw new FactoryConfigurationError(ie);
        }
        return factoryImpl;
    }
    
    /**
     * Creates a new instance of a SAXParser using the currently
     * configured factory parameters.
     *
     * @return A new instance of a SAXParser.
     *
     * @exception ParserConfigurationException if a parser cannot
     * be created which satisfies the requested configuration.
     */
    
    public abstract SAXParser newSAXParser()
        throws ParserConfigurationException, SAXException;

    
    /**
     * Specifies that the parser produced by this code will
     * provide support for XML namespaces. By default the value of this is set
     * to <code>false</code>.
     *
     * @param awareness true if the parser produced by this code will
     *                  provide support for XML namespaces; false otherwise.
     */
    
    public void setNamespaceAware(boolean awareness)
    {
        this.namespaceAware = awareness;
    }

    /**
     * Specifies that the parser produced by this code will
     * validate documents as they are parsed. By default the value of this is
     * set to <code>false</code>.
     *
     * @param validating true if the parser produced by this code will
     *                   validate documents as they are parsed; false otherwise.
     */
    
    public void setValidating(boolean validating)
    {
        this.validating = validating;
    }

    /**
     * Indicates whether or not the factory is configured to produce
     * parsers which are namespace aware.
     *
     * @return true if the factory is configured to produce
     *         parsers which are namespace aware; false otherwise.
     */
    
    public boolean isNamespaceAware() {
        return namespaceAware;
    }

    /**
     * Indicates whether or not the factory is configured to produce
     * parsers which validate the XML content during parse.
     *
     * @return true if the factory is configured to produce parsers which validate
     *         the XML content during parse; false otherwise.
     */
    
    public boolean isValidating() {
        return validating;
    }

    /**
     *
     * Sets the particular feature in the underlying implementation of
     * org.xml.sax.XMLReader.
     * A list of the core features and properties can be found at
     * <a href="http://www.megginson.com/SAX/Java/features.html"> http://www.megginson.com/SAX/Java/features.html </a>
     *
     * @param name The name of the feature to be set.
     * @param value The value of the feature to be set.
     * @exception SAXNotRecognizedException When the underlying XMLReader does
     *            not recognize the property name.
     *
     * @exception SAXNotSupportedException When the underlying XMLReader
     *            recognizes the property name but doesn't support the
     *            property.
     *
     * @see org.xml.sax.XMLReader#setFeature
     */
    public abstract void setFeature(String name, boolean value)
        throws ParserConfigurationException, SAXNotRecognizedException,
                SAXNotSupportedException;

    /**
     *
     * Returns the particular property requested for in the underlying
     * implementation of org.xml.sax.XMLReader.
     *
     * @param name The name of the property to be retrieved.
     * @return Value of the requested property.
     *
     * @exception SAXNotRecognizedException When the underlying XMLReader does
     *            not recognize the property name.
     *
     * @exception SAXNotSupportedException When the underlying XMLReader
     *            recognizes the property name but doesn't support the
     *            property.
     *
     * @see org.xml.sax.XMLReader#getProperty
     */
    public abstract boolean getFeature(String name)
        throws ParserConfigurationException, SAXNotRecognizedException,
                SAXNotSupportedException;

    /**
     * This class is duplicated for each JAXP subpackage so keep it in
     * sync.  It is package private so other factories can also use this
     * class.
     */
    static class FactoryFinder {
        /**
         * The ClassLoader is one per jar file so it can be a static field
         */
        private static ClassLoader classLoader;
        private static boolean classLoaderFound = false;

        private String factoryId;
        private String defaultImplClassName;

        /** Avoid reading all the files when the findFactory
            method is called the second time ( cache the result of
            finding the default impl )
        */
        private Class foundFactory = null;

        /** Temp debug code - this will be removed after we test everything
         */
        private static boolean debug = false;
        static {
            // Use try/catch block to support applets
            try {
                debug = System.getProperty("jaxp.debug") != null;
            } catch (Exception e) {
            }
        }

        /**
         * @param factoryId             Name of the factory to find, same as
         *                              a property name
         * @param defaultImplClassName  Default implementation class name,
         *                              if nothing else is found
         *
         * Note: defaultImplClassName may be removed after services are
         * tested well enough.  Package private so this code can be shared.
         */
        FactoryFinder(String factoryId, String defaultImplClassName) {
            this.factoryId = factoryId;
            this.defaultImplClassName = defaultImplClassName;
        }

        /**
         * Figure out which ClassLoader to use.  For JDK 1.2 and later use
         * the context ClassLoader if possible.  Note: we use reflection
         * instead of catching an exception so that this code will run in
         * non-Sun JVMs such as the Microsoft JVM in IE.
         */           
        private static void findClassLoader() {
            try {
                Method m = Thread.class.getMethod("getContextClassLoader",
                                                  null);
                classLoader =
                    (ClassLoader) m.invoke(Thread.currentThread(), null);
            } catch (NoSuchMethodException e) {
                // Assume that we are running JDK 1.1
                if (debug) {
                    System.err.println("JAXP: assuming JDK 1.1");
                }
                classLoader = SAXParserFactory.class.getClassLoader();
            } catch (IllegalAccessException e) {
                // assert(false)
                throw new RuntimeException("Unexpected IllegalAccessException");
            } catch (InvocationTargetException e) {
                // assert(e.getTargetException() instanceof SecurityException)
                throw (SecurityException) e.getTargetException();
            }
        }

        /**
         * Load a class using the predetermined ClassLoader
         */
        private static Class classForName(String className)
            throws ClassNotFoundException
        {
            if (classLoader == null) {
                return Class.forName(className);
            } else {
                return classLoader.loadClass(className);
            }
        }

        /**
         * Finds the implementation Class object in the specified order.
         * @return Class object of factory, never null
         */
        Class findFactory()
            throws ClassNotFoundException
        {
            if (!classLoaderFound) {
                findClassLoader();
                classLoaderFound = true;
            }

            // Use the system property first
            try {
                String systemProp =
                    System.getProperty( factoryId );
                if( systemProp!=null) {
                    if( debug )
                        System.err.println("JAXP: found system property" +
                                           systemProp);
                    return classForName(systemProp);
                }
            } catch (SecurityException se) {
            }

            // Use factory cache after system property in case app
            // purposely redefines the property
            if (foundFactory != null) {
                return foundFactory;
            }
    
            // try to read from $java.home/lib/xml.properties
            try {
                String javah=System.getProperty( "java.home" );
                String configFile = javah + File.separator +
                    "lib" + File.separator + "jaxp.properties";
                File f=new File( configFile );
                if( f.exists()) {
                    Properties props=new Properties();
                    props.load( new FileInputStream(f));
                    String factoryClassName = props.getProperty(factoryId);
                    if( debug )
                        System.err.println("JAXP: found java.home property " +
                                           factoryClassName);
                    foundFactory = classForName(factoryClassName);
                    return foundFactory;
                }
            } catch(Exception ex ) {
                if( debug ) ex.printStackTrace();
            }

            String serviceId = "META-INF/services/" + factoryId;
            // try to find services in CLASSPATH
            try {
                InputStream is=null;
                if (classLoader == null) {
                    is=ClassLoader.getSystemResourceAsStream( serviceId );
                } else {
                    is=classLoader.getResourceAsStream( serviceId );
                }
        
                if( is!=null ) {
                    if( debug )
                        System.err.println("JAXP: found  " +
                                           serviceId);
                    BufferedReader rd =
                        new BufferedReader(new InputStreamReader(is));
        
                    String factoryClassName = rd.readLine();
                    rd.close();

                    if (factoryClassName != null &&
                        ! "".equals(factoryClassName)) {
                        if( debug )
                            System.err.println("JAXP: loaded from services: " +
                                               factoryClassName);
                        foundFactory = classForName(factoryClassName);
                        return foundFactory;
                    }
                }
            } catch( Exception ex ) {
                if( debug ) ex.printStackTrace();
            }

            if( debug )
                System.err.println("JAXP: loaded from default value: " +
                                   defaultImplClassName);
            foundFactory = classForName(defaultImplClassName);
            return foundFactory;
        }
    }
}
