/**
 * $Id: TransformerFactory.java,v 1.13 2001/03/06 20:49:06 edg Exp $
 *
 * Copyright (c) 2000-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package javax.xml.transform;

import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;

import java.util.Properties;
import java.util.Enumeration;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

/**
 * A TransformerFactory instance can be used to create
 * {@link javax.xml.transform.Transformer} and
 * {@link javax.xml.transform.Templates} objects.
 *
 * <p>The system property that determines which Factory implementation to
 * create is named <code>"javax.xml.transform.TransformerFactory"</code>.
 * This property names a concrete subclass of the
 * <code>TransformerFactory</code> abstract class. If the property is not
 * defined, a platform default is be used.</p>
 *
 * An implementation of the <code>TransformerFactory</code> class is
 * <em>NOT</em> guaranteed to be thread safe. It is up to the user application 
 * to make sure about the use of the <code>TransformerFactory</code> from 
 * more than one thread. Alternatively the application can have one instance 
 * of the <code>TransformerFactory</code> per thread.
 * An application can use the same instance of the factory to obtain one or 
 * more instances of a <code>Transformer</code> or <code>Templates</code> 
 * provided the instance of the factory isn't being used in more than one 
 * thread at a time.
 */
public abstract class TransformerFactory {
    /** One FactoryFinder per factory class */
    private static FactoryFinder factoryFinder = new FactoryFinder(
        // The default property name according to the JAXP spec
        "javax.xml.transform.TransformerFactory",
        "org.apache.xalan.processor.TransformerFactoryImpl");

    /**
     * Default constructor is protected on purpose.
     */
    protected TransformerFactory() {}

    /**
     * Obtain a new instance of a <code>TransformerFactory</code>.
     * This static method creates a new factory instance
     * This method uses the following ordered lookup procedure to determine
     * the <code>TransformerFactory</code> implementation class to
     * load:
     * <ul>
     * <li>
     * Use the <code>javax.xml.transform.TransformerFactory</code> system
     * property.
     * </li>
     * <li>
     * Use the properties file "lib/jaxp.properties" in the JRE directory.
     * This configuration file is in standard <code>java.util.Properties
     * </code> format and contains the fully qualified name of the
     * implementation class with the key being the system property defined
     * above.
     * </li>
     * <li>
     * Use the Services API (as detailed in the JAR specification), if
     * available, to determine the classname. The Services API will look
     * for a classname in the file
     * <code>META-INF/services/javax.xml.transform.TransformerFactory</code>
     * in jars available to the runtime.
     * </li>
     * <li>
     * Platform default <code>TransformerFactory</code> instance.
     * </li>
     * </ul>
     *
     * Once an application has obtained a reference to a <code>
     * TransformerFactory</code> it can use the factory to configure
     * and obtain parser instances.
     *
     * @return new TransformerFactory instance, never null.
     *
     * @throws TransformerFactoryConfigurationError
     * if the implmentation is not available or cannot be instantiated.
     */
    public static TransformerFactory newInstance()
            throws TransformerFactoryConfigurationError {
        TransformerFactory factoryImpl;
        try {
            Class factoryClass = factoryFinder.findFactory();
            factoryImpl = (TransformerFactory) factoryClass.newInstance();
        } catch (ClassNotFoundException cnfe) {
            throw new TransformerFactoryConfigurationError(cnfe);
        } catch (IllegalAccessException iae) {
            throw new TransformerFactoryConfigurationError(iae);
        } catch (InstantiationException ie) {
            throw new TransformerFactoryConfigurationError(ie);
        }
        return factoryImpl;
    }

    /**
     * Process the Source into a Transformer object.  Care must
     * be given not to use this object in multiple threads running concurrently.
     * Different TransformerFactories can be used concurrently by different
     * threads.
     *
     * @param source An object that holds a URI, input stream, etc.
     *
     * @return A Transformer object that may be used to perform a transformation
     * in a single thread, never null.
     *
     * @exception TransformerConfigurationException May throw this during the parse
     *            when it is constructing the Templates object and fails.
     */
    public abstract Transformer newTransformer(Source source)
        throws TransformerConfigurationException;

    /**
     * Create a new Transformer object that performs a copy
     * of the source to the result.
     *
     * @param source An object that holds a URI, input stream, etc.
     *
     * @return A Transformer object that may be used to perform a transformation
     * in a single thread, never null.
     *
     * @exception TransformerConfigurationException May throw this during
     *            the parse when it is constructing the
     *            Templates object and fails.
     */
    public abstract Transformer newTransformer()
        throws TransformerConfigurationException;

    /**
     * Process the Source into a Templates object, which is a
     * a compiled representation of the source. This Templates object
     * may then be used concurrently across multiple threads.  Creating
     * a Templates object allows the TransformerFactory to do detailed
     * performance optimization of transformation instructions, without
     * penalizing runtime transformation.
     *
     * @param source An object that holds a URL, input stream, etc.
     *
     * @return A Templates object capable of being used for transformation purposes,
     * never null.
     *
     * @exception TransformerConfigurationException May throw this during the parse when it
     *            is constructing the Templates object and fails.
     */
    public abstract Templates newTemplates(Source source)
        throws TransformerConfigurationException;

    /**
     * Get the stylesheet specification(s) associated
     * via the xml-stylesheet processing instruction (see
     * http://www.w3.org/TR/xml-stylesheet/) with the document
     * document specified in the source parameter, and that match
     * the given criteria.  Note that it is possible to return several
     * stylesheets, in which case they are applied as if they were
     * a list of imports or cascades in a single stylesheet.
     *
     * @param source The XML source document.
     * @param media The media attribute to be matched.  May be null, in which
     *              case the prefered templates will be used (i.e. alternate = no).
     * @param title The value of the title attribute to match.  May be null.
     * @param charset The value of the charset attribute to match.  May be null.
     *
     * @return A Source object suitable for passing to the TransformerFactory.
     *
     * @throws TransformerConfigurationException.
     */
    public abstract Source getAssociatedStylesheet(
        Source source, String media, String title, String charset)
            throws TransformerConfigurationException;

    /**
     * Set an object that is used by default during the transformation
     * to resolve URIs used in xsl:import, or xsl:include.
     *
     * @param resolver An object that implements the URIResolver interface,
     * or null.
     */
    public abstract void setURIResolver(URIResolver resolver);

    /**
     * Get the object that is used by default during the transformation
     * to resolve URIs used in document(), xsl:import, or xsl:include.
     *
     * @return The URIResolver that was set with setURIResolver.
     */
    public abstract URIResolver getURIResolver();

    //======= CONFIGURATION METHODS =======

    /**
     * Look up the value of a feature.
     *
     * <p>The feature name is any absolute URI.</p>
     * @param name The feature name, which is an absolute URI.
     * @return The current state of the feature (true or false).
     */
    public abstract boolean getFeature(String name);

    /**
     * Allows the user to set specific attributes on the underlying
     * implementation.  An attribute in this context is defined to
     * be an option that the implementation provides.
     *
     * @param name The name of the attribute.
     * @param value The value of the attribute.
     * @throws IllegalArgumentException thrown if the underlying
     * implementation doesn't recognize the attribute.
     */
    public abstract void setAttribute(String name, Object value)
        throws IllegalArgumentException;

    /**
     * Allows the user to retrieve specific attributes on the underlying
     * implementation.
     * @param name The name of the attribute.
     * @return value The value of the attribute.
     * @throws IllegalArgumentException thrown if the underlying
     * implementation doesn't recognize the attribute.
     */
    public abstract Object getAttribute(String name)
        throws IllegalArgumentException;

    /**
     * Set the error event listener for the TransformerFactory, which
     * is used for the processing of transformation instructions,
     * and not for the transformation itself.
     *
     * @param listener The new error listener.
     * @throws IllegalArgumentException if listener is null.
     */
    public abstract void setErrorListener(ErrorListener listener)
        throws IllegalArgumentException;

    /**
     * Get the error event handler for the TransformerFactory.
     *
     * @return The current error handler, which should never be null.
     */
    public abstract ErrorListener getErrorListener();

    /**
     * This class is duplicated for each JAXP subpackage so keep it in
     * sync.  It is package private so other factories can also use this
     * class.
     */
    static class FactoryFinder {
        /**
         * The ClassLoader is one per jar file so it can be a static field
         */
        private static ClassLoader classLoader;
        private static boolean classLoaderFound = false;

        private String factoryId;
        private String defaultImplClassName;

        /** Avoid reading all the files when the findFactory
            method is called the second time ( cache the result of
            finding the default impl )
        */
        private Class foundFactory = null;

        /** Temp debug code - this will be removed after we test everything
         */
        private static boolean debug = false;
        static {
            // Use try/catch block to support applets
            try {
                debug = System.getProperty("jaxp.debug") != null;
            } catch (Exception e) {
            }
        }

        /**
         * @param factoryId             Name of the factory to find, same as
         *                              a property name
         * @param defaultImplClassName  Default implementation class name,
         *                              if nothing else is found
         *
         * Note: defaultImplClassName may be removed after services are
         * tested well enough.  Package private so this code can be shared.
         */
        FactoryFinder(String factoryId, String defaultImplClassName) {
            this.factoryId = factoryId;
            this.defaultImplClassName = defaultImplClassName;
        }

        /**
         * Figure out which ClassLoader to use.  For JDK 1.2 and later use
         * the context ClassLoader if possible.  Note: we use reflection
         * instead of catching an exception so that this code will run in
         * non-Sun JVMs such as the Microsoft JVM in IE.
         */           
        private static void findClassLoader() {
            try {
                Method m = Thread.class.getMethod("getContextClassLoader",
                                                  null);
                classLoader =
                    (ClassLoader) m.invoke(Thread.currentThread(), null);
            } catch (NoSuchMethodException e) {
                // Assume that we are running JDK 1.1
                if (debug) {
                    System.err.println("JAXP: assuming JDK 1.1");
                }
                classLoader = TransformerFactory.class.getClassLoader();
            } catch (IllegalAccessException e) {
                // assert(false)
                throw new RuntimeException("Unexpected IllegalAccessException");
            } catch (InvocationTargetException e) {
                // assert(e.getTargetException() instanceof SecurityException)
                throw (SecurityException) e.getTargetException();
            }
        }

        /**
         * Load a class using the predetermined ClassLoader
         */
        private static Class classForName(String className)
            throws ClassNotFoundException
        {
            if (classLoader == null) {
                return Class.forName(className);
            } else {
                return classLoader.loadClass(className);
            }
        }

        /**
         * Finds the implementation Class object in the specified order.
         * @return Class object of factory, never null
         */
        Class findFactory()
            throws ClassNotFoundException
        {
            if (!classLoaderFound) {
                findClassLoader();
                classLoaderFound = true;
            }

            // Use the system property first
            try {
                String systemProp =
                    System.getProperty( factoryId );
                if( systemProp!=null) {
                    if( debug )
                        System.err.println("JAXP: found system property" +
                                           systemProp);
                    return classForName(systemProp);
                }
            } catch (SecurityException se) {
            }

            // Use factory cache after system property in case app
            // purposely redefines the property
            if (foundFactory != null) {
                return foundFactory;
            }
    
            // try to read from $java.home/lib/xml.properties
            try {
                String javah=System.getProperty( "java.home" );
                String configFile = javah + File.separator +
                    "lib" + File.separator + "jaxp.properties";
                File f=new File( configFile );
                if( f.exists()) {
                    Properties props=new Properties();
                    props.load( new FileInputStream(f));
                    String factoryClassName = props.getProperty(factoryId);
                    if( debug )
                        System.err.println("JAXP: found java.home property " +
                                           factoryClassName);
                    foundFactory = classForName(factoryClassName);
                    return foundFactory;
                }
            } catch(Exception ex ) {
                if( debug ) ex.printStackTrace();
            }

            String serviceId = "META-INF/services/" + factoryId;
            // try to find services in CLASSPATH
            try {
                InputStream is=null;
                if (classLoader == null) {
                    is=ClassLoader.getSystemResourceAsStream( serviceId );
                } else {
                    is=classLoader.getResourceAsStream( serviceId );
                }
        
                if( is!=null ) {
                    if( debug )
                        System.err.println("JAXP: found  " +
                                           serviceId);
                    BufferedReader rd =
                        new BufferedReader(new InputStreamReader(is));
        
                    String factoryClassName = rd.readLine();
                    rd.close();

                    if (factoryClassName != null &&
                        ! "".equals(factoryClassName)) {
                        if( debug )
                            System.err.println("JAXP: loaded from services: " +
                                               factoryClassName);
                        foundFactory = classForName(factoryClassName);
                        return foundFactory;
                    }
                }
            } catch( Exception ex ) {
                if( debug ) ex.printStackTrace();
            }

            if( debug )
                System.err.println("JAXP: loaded from default value: " +
                                   defaultImplClassName);
            foundFactory = classForName(defaultImplClassName);
            return foundFactory;
        }
    }
}
